/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TPTR_GEOSERIES_H
#define O2SCL_TPTR_GEOSERIES_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/tptr_schedule.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Temperature schedule for a geometric series

      The temperature begins at \c start, and is divided by \c ratio,
      until it is smaller than \c end. The ending value is divided by
      \c sqrt(ratio) to avoid finite precision problems for series
      when \c start/end is an integral power of \c ratio.

      The default schedule is \f$T=1/(1.01)^{n}\f$ for 
      \f$n=0,1,2,3,...,463\f$ (until \f$T<0.01\f$) 
      given by \c ustart=1, \c uend=0.01, \c uratio=1.01.
      
  */
  template<class vec_t=ovector_view>
    class tptr_geoseries : public tptr_schedule<vec_t> {
    public:
      
    tptr_geoseries() {
      dstart=1.0;
      end=1.0e-2;
      ratio=1.01;
      last=dstart*ratio;
    }
      
    virtual ~tptr_geoseries() {}

    /// Set the limits for the geometric series
    int set_series(double udstart, double uend, double uratio) {
      if ((udstart>uend && uratio<1.0) || (udstart<uend && uratio>1.0)) {
	set_err_ret("Series will never end in tptr_geoseries::set_series()",
		    gsl_einval);
      }
      dstart=udstart;
      end=uend;
      ratio=uratio;
      last=dstart*ratio;
      return 0;
    }
      
    /// Get the number of temperatures in the series
    int get_npoints() {
      double n=log(dstart/end*sqrt(ratio))/log(ratio);
      return ((int)n)+1;
    }
      
    /// Return the first temperature
    virtual double start(double min, int nv, const vec_t &best, void *vp) {
      last=dstart;
      return last;
    }
      
    /// Return the next temperature
    virtual double next(double min, int nv, const vec_t &best, void *vp) {
      last=last/ratio;
      return last; 
    }

    /// Return true if the last step made the temperature too small
    virtual bool done(double min, int nv, const vec_t &best, void *vp) {
      if (last<end/sqrt(ratio)) return true;
      return false;
    }
      
    /// Return string denoting type ("tptr_geoseries")
    virtual const char *type() { return "tptr_geoseries"; }
      
    protected:
      
#ifndef DOXYGEN_INTERNAL
    
    /// \name parameters for the schedule
    //@{
    double dstart, end, ratio;
    //@}

    /// The last temperature returned
    double last;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif

