/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ARRAY_H
#define O2SCL_ARRAY_H

/** \file array.h
    \brief Various array classes

    This file contains classes and functions for operating with
    C-style 1- or 2-dimensional arrays and pointers to double. For
    more generic operations on generic vector objects (including in
    some cases C-style arrays), see also the file \ref vector.h . This
    file contains the allocation classes
    - array_alloc
    - array_2d_alloc
    - pointer_alloc
    - pointer_2d_alloc
    the classes for the manipulation of arrays in smart_interp
    - array_reverse
    - array_subvector
    - array_subvector_reverse
    - array_const_reverse
    - array_const_subvector
    - array_const_subvector_reverse
    the array equivalent of omatrix_row or umatrix_row (see 
    usage proposed in src/ode/ode_it_solve_ts.cpp)
    - array_row

    For an example of the usage of the array allocation classes,
    see the \ref ex_mroot_sect .

    \note The classes 
    - array_reverse
    - array_subvector
    - array_subvector_reverse
    - array_const_reverse
    - array_const_subvector
    - array_const_subvector_reverse
    can be used with pointers or arrays, but array_alloc and
    pointer_alloc are \e not interchangable.

    \todo Ensure that array_row works, either here or in 
    src/ode/ode_it_solve_ts.cpp

*/
#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <gsl/gsl_ieee_utils.h>
#include <gsl/gsl_sort.h>

#ifndef DOXYGENP
namespace o2scl
{
#endif

  /** 
      \brief A simple class to provide an \c allocate() function
      for arrays
      
      The functions here are blank, as fixed-length arrays are
      automatically allocated and destroyed by the compiler. This
      class is present to provide an analog to \ref pointer_alloc and
      \ref ovector_alloc.
  */
  template<class vec_t> class array_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(vec_t &v, size_t i) {}
    /// Free memory
    void free(vec_t &v) {}
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for 2-dimensional arrays
      
      The functions here are blank, as fixed-length arrays are
      automatically allocated and destroyed by the compiler. This
      class is present to provide an analog to \ref pointer_2d_alloc
      and \ref omatrix_alloc
  */
  template<class mat_t> class array_2d_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(mat_t &v, size_t i, size_t j) {}
    /// Free memory
    void free(mat_t &v, size_t i) {}
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for pointers

      Uses \c new and \c delete.
   */
  template<class base_t> class pointer_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(base_t *&v, size_t i) { v=new base_t[i]; }
    /// Free memory
    void free(base_t *&v) { delete[] v; }
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for pointers

      Uses \c new and \c delete.
  */
  template<class base_t> class pointer_2d_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(base_t **&v, size_t i, size_t j) { 
      v=new base_t *[i]; 
      for(size_t m=0;m<i;m++) v[m]=new base_t[j];
    }
    /// Free memory
    void free(base_t **&v, size_t i) { 
      for(size_t m=0;m<i;m++) delete[] v[m];
      delete[] v; 
    }
  };

  /** 
      \brief A simple class which reverses the order of an array
   */
  template<size_t sz> class array_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_reverse(double *arr) {
      a=arr;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[sz-1-i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[sz-1-i];
    }
    
  };

  /** 
      \brief A simple class which reverses the order of an array
   */
  template<size_t sz> class array_const_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_reverse(const double *arr) {
      a=(double *)arr;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[sz-1-i];
    }
    
  };

  /** 
      \brief A simple subvector class for an array (without error checking)
   */
  class array_subvector {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_subvector(double *arr, size_t offset, size_t n) {
      a=arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[i];
    }
    
  };

  /** 
      \brief Column of a 2d array

      This works because two-dimensional arrays are always
      continguous (as indicated in appendix C of Soustroup's book)
  */
  template<size_t R, size_t C> class array_2d_column {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create an object as the <tt>i</tt>th column of \c mat
    array_2d_column(double mat[R][C], size_t i) {
      a=&(mat[0][i]);
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[R*i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[R*i];
    }
    
  };

  /** 
      \brief Row of a 2d array
  */
  template<class array_2d_t> class array_2d_row {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create an object as the <tt>i</tt>th row of \c mat
    array_2d_row(array_2d_t &mat, size_t i) {
      a=&(mat[i][0]);
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[i];
    }
    
  };

  /** \brief A simple subvector class for a const array 
      (without error checking)
   */
  class array_const_subvector {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_subvector(const double *arr, size_t offset, size_t n) {
      a=(double *)arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[i];
    }
    
  };

  /** 
      \brief Reverse a subvector of an array
   */
  class array_subvector_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;
    
    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_subvector_reverse(double *arr, size_t offset, size_t n) {
      a=arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[len-1-i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[len-1-i];
    }
    
  };

  /** 
      \brief Reverse a subvector of a const array
  */
  class array_const_subvector_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_subvector_reverse(const double *arr, size_t offset, size_t n) {
      a=(double *)arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[len-1-i];
    }
    
  };

  /// Extract a row of a C-style 2d-array
  template <class data_t, class array_2d_t> class array_row {
  protected:
    /// The pointer
    data_t *p;
  public:
    /// View the <tt>i</tt>th row of the 2-d array \c a.
    array_row(array_2d_t &a, size_t i) {
      p=&(a[i][0]);
    }
    /// The element in the <tt>i</tt>th column of the chosen row
    data_t &operator[](size_t i) {
      return p[i];
    }
  };

  /// Create a new C-style 2-dimensional array
  template<class type> type **new_2d_array(size_t nr, size_t nc) {
    type **t;
    t=new type *[nr];
    for(size_t i=0;i<nr;i++) t[i]=new type[nc];
    return t;
  }

  /// Create a new C-style 2-dimensional array
  template<class type> int delete_2d_array(type **t, size_t nr) {
    for(size_t i=0;i<nr;i++) delete[] t[i];
    delete[] t;
    return 0;
  }

#ifndef DOXYGENP
}
#endif

#endif
