/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_COLUMNIFY_H
#define O2SCL_COLUMNIFY_H

/** \file columnify.h
    \brief Functions to create output in columns
*/

#include <iostream>
#include <string>
#include <vector>
#include <o2scl/misc.h>
#include <o2scl/array.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Create nicely formatted columns from a table of strings
    
      This is a brute-force approach of order \f$ \mathrm{ncols} \times
      \mathrm{nrows} \f$.
      
      \todo Move the screenify() functionality from misc.h into this class?
  */
  class columnify {
  
  public:
  
    columnify() {
    }

    /// Align the left-hand sides
    static const int align_left=1;
    /// Align the right-hand sides
    static const int align_right=2;
    /// Center, slightly to the left if spacing is uneven
    static const int align_lmid=3;
    /// Center, slightly to the right if spacing is uneven
    static const int align_rmid=4;
    /// Align with decimal points
    static const int align_dp=5;
    /** \brief Align negative numbers to the left and use a space for 
	positive numbers
    */
    static const int align_lnum=6;
  
    /** \brief Take \c table and create a new object \c ctable with 
	appropriately formatted columns

	The table of strings should be stored in \c table in
	"column-major" order, so that \c table has the interpretation
	of a set of columns to be aligned. Before calling align(), \c
	ctable should be allocated so that at least the first \c nrows
	entries can be assigned, and \c align_spec should contain \c
	ncols entries specifying the style of alignment for each
	column.
    */
    template<class mat_string_t, class vec_string_t, class vec_int_t>
      int align(const mat_string_t &table, size_t ncols, size_t nrows, 
		vec_string_t &ctable, vec_int_t &align_spec) {
      
      // Make space for the size information
      size_t *csizes=new size_t[ncols], *csizes2=new size_t[ncols];
      for(size_t i=0;i<ncols;i++) {
	csizes[i]=0;
	csizes2[i]=0;
      }
  
      // Compute the sizes of all the entries in all of the columns so
      // we know how many spaces to add
      for(size_t i=0;i<ncols;i++) {
	for(size_t j=0;j<nrows;j++) {
	  if (align_spec[i]==align_dp) {
	    size_t loc=table[i][j].find('.');
	    std::string left, right;
	    if (loc!=std::string::npos) {
	      left=table[i][j].substr(0,loc+1);
	      right=table[i][j].substr(loc+1,
				       table[i][j].length()-loc-1);
	    } else {
	      left=table[i][j]+' ';
	      right="";
	    }
	    if (left.length()>csizes[i]) csizes[i]=left.length();
	    if (right.length()>csizes2[i]) csizes2[i]=right.length();
	  } else {
	    if (table[i][j].length()>csizes[i]) csizes[i]=table[i][j].length();
	  }
	}
      }

      // Go through row by row, adding enough spaces to make one string
      // per row
      for(size_t j=0;j<nrows;j++) {

	std::string tmp="";

	for(size_t i=0;i<ncols;i++) {

	  // Handle each alignment case separately
	  if (align_spec[i]==align_right) {

	    for(size_t k=table[i][j].length();k<csizes[i];k++) {
	      tmp+=' ';
	    }
	    tmp+=table[i][j];
	    if (i!=ncols-1) tmp+=' ';

	  } else if (align_spec[i]==align_left) {

	    tmp+=table[i][j];
	    for(size_t k=table[i][j].length();k<csizes[i];k++) {
	      tmp+=' ';
	    }
	    if (i!=ncols-1) tmp+=' ';

	  } else if (align_spec[i]==align_lmid) {

	    size_t le=(csizes[i]-table[i][j].length())/2;
	    size_t ri=csizes[i]-table[i][j].length()-le;
	    for(size_t k=0;k<le;k++) tmp+=' ';
	    tmp+=table[i][j];
	    for(size_t k=0;k<ri;k++) tmp+=' ';
	    if (i!=ncols-1) tmp+=' ';

	  } else if (align_spec[i]==align_rmid) {

	    size_t ri=(csizes[i]-table[i][j].length())/2;
	    size_t le=csizes[i]-table[i][j].length()-ri;
	    for(size_t k=0;k<le;k++) tmp+=' ';
	    tmp+=table[i][j];
	    for(size_t k=0;k<ri;k++) tmp+=' ';
	    if (i!=ncols-1) tmp+=' ';

	  } else if (align_spec[i]==align_dp) {

	    size_t loc=table[i][j].find('.');
	    std::string left, right;
	    if (loc!=std::string::npos) {
	      left=table[i][j].substr(0,loc+1);
	      right=table[i][j].substr(loc+1,
				       table[i][j].length()-loc-1);
	    } else {
	      left=table[i][j]+' ';
	      right="";
	    }

	    for(size_t k=left.length();k<csizes[i];k++) tmp+=' ';
	    tmp+=left;
	    tmp+=right;
	    for(size_t k=right.length();k<csizes2[i];k++) tmp+=' ';
	    if (i!=ncols-1) tmp+=' ';

	  } else if (align_spec[i]==align_lnum) {

	    if (table[i][j].length()==csizes[i]) {
	      tmp+=table[i][j];
	    } else {
	      if (table[i][j][0]>='0' && table[i][j][0]<='9') {
		tmp+=' ';
		tmp+=table[i][j];
		for(size_t k=table[i][j].length();k<csizes[i]-1;k++) {
		  tmp+=' ';
		}
	      } else {
		tmp+=table[i][j];
		for(size_t k=table[i][j].length();k<csizes[i];k++) {
		  tmp+=' ';
		}
	      }
	    }
	    if (i!=ncols-1) tmp+=' ';
	  }
	
	}

	// Add the row to the user-specified array and go to the next row
	ctable[j]=tmp;
      
      }

      // Free the memory assocated with the column widths
      delete[] csizes;
      delete[] csizes2;
  
      return 0;
    }
  
  };

  /** 
      \brief A operator for naive matrix output
      
      The type \c mat_t can be any matrix type which allows 
      individual element access using \c operator()(size_t,size_t).
      
      This outputs all of the matrix elements using output settings
      specified by \c os. The alignment performed by \ref columnify
      using columnify::align_dp, i.e. the numbers are aligned by
      their decimal points. If the numbers have no decimal points,
      then the decimal point is assumed to be to the right of the
      last character in the string represetation of the number.
  */
  template<class mat_t> int matrix_out_paren(std::ostream &os, mat_t &A,
					     size_t nrows, size_t ncols) {
    
    columnify co;
    std::string **stab;
    stab=o2scl::new_2d_array<std::string>(ncols,nrows);
    std::vector<std::string> ctable(nrows);
    std::vector<int> alig(ncols);
    
    for(size_t j=0;j<ncols;j++) {
      alig[j]=columnify::align_dp;
      for(size_t i=0;i<nrows;i++) {
	stab[j][i]=dtos(A(i,j),os);
      }
    }
    co.align(stab,ncols,nrows,ctable,alig);
    for(size_t i=0;i<nrows;i++) {
      os << ctable[i] << std::endl;
    }
    
    o2scl::delete_2d_array(stab,ncols);

    return 0;
  }

  /** 
      \brief A operator for simple complex matrix output
  */
  template<class mat_t> int matrix_cx_out_paren(std::ostream &os, mat_t &A,
						size_t nrows, size_t ncols) {
    
    columnify co;
    std::string **stab;
    stab=o2scl::new_2d_array<std::string>(ncols,nrows);
    std::vector<std::string> ctable(nrows);
    std::vector<int> alig(ncols);
    
    for(size_t j=0;j<ncols;j++) {
      alig[j]=columnify::align_left;
      for(size_t i=0;i<nrows;i++) {
	stab[j][i]=((std::string)"(")+dtos(A(i,j).dat[0],os)+
	  ((std::string)",")+dtos(A(i,j).dat[1],os)+((std::string)")");
      }
    }
    co.align(stab,ncols,nrows,ctable,alig);
    for(size_t i=0;i<nrows;i++) {
      os << ctable[i] << std::endl;
    }
    
    o2scl::delete_2d_array(stab,ncols);

    return 0;
  }

  /** 
      \brief A operator for simple 2-d array output
      
      The type \c mat_t can be any 2d-array type which allows 
      individual element access using \c [size_t][size_t]
      
      This outputs all of the matrix elements using output settings
      specified by \c os. The alignment performed by \ref columnify
      using columnify::align_dp, i.e. the numbers are aligned by
      their decimal points. If the numbers have no decimal points,
      then the decimal point is assumed to be to the right of the
      last character in the string represetation of the number.

      \todo If all of the matrix elements are positive integers 
      and scientific mode is not set, then we can avoid printing
      the extra spaces.
  */
  template<class mat_t> int matrix_out(std::ostream &os, mat_t &A,
				       size_t nrows, size_t ncols) {
    
    columnify co;
    std::string **stab;
    stab=o2scl::new_2d_array<std::string>(ncols,nrows);
    std::vector<std::string> ctable(nrows);
    std::vector<int> alig(ncols);
    
    for(size_t j=0;j<ncols;j++) {
      alig[j]=columnify::align_dp;
      for(size_t i=0;i<nrows;i++) {
	stab[j][i]=dtos(A[i][j],os);
      }
    }
    co.align(stab,ncols,nrows,ctable,alig);
    for(size_t i=0;i<nrows;i++) {
      os << ctable[i] << std::endl;
    }
    
    o2scl::delete_2d_array(stab,ncols);

    return 0;
  }

#ifndef DOXYGENP
}
#endif

#endif
