/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FILE_DETECT_H
#define O2SCL_FILE_DETECT_H

#include <iostream>
#include <fstream>
#include <string>
#include <o2scl/misc.h>
#include <o2scl/err_hnd.h>
#include <o2scl/text_file.h>
#include <o2scl/binary_file.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Read a (possibly compressed) file and automatically detect
      the file format 

      Really nasty hack. This works by copying the file to a temporary
      file in /tmp and then uncompressing it using a call to
      \c system("gunzip /tmp/filename") . When the file is
      closed, the temporary file is removed using 'rm -f'.

      If the filename ends with ".gz" or ".bz2", then input_detect
      will try to uncompress it (using gunzip or bunzip2), otherwise,
      the file will be treated as normal.

      Note that there must be enough disk space in the temporary 
      directory for the uncompressed file or the read will fail.
    
      \future Allow the user to specify the compression commands
      in configure, or at least specify the path to gzip, bzip2, etc.

  */
  class file_detect {
  public:

    file_detect();

    virtual ~file_detect() {};

    /** \brief Open an input file with the given name

	If the filename ends with ".gz" or ".bz2", then the file is
	assumed to be compressed.

	It is important to note that the file is not closed until
	file_detect::close() method is called. 
    */
    in_file_format *open(const char *s);
    
    /** \brief Close an input file */
    virtual int close();

    /** \brief Return \c true  if the opened file was
	originally compressed */
    virtual bool is_compressed() { return compressed; }

    /** \brief Return \c true  if the opened file was
	a binary file */
    virtual bool is_binary() { return binary; }

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The temporary filename
    std::string temp_filename;
    /// The user-supplied filename
    std::string user_filename;
    /// The input file
    in_file_format *iffp;
    /// True if the file was compressed
    bool compressed;
    /// True if the file was a binary file
    bool binary;

#endif
  
  };

}

#endif
