/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_INTERP_H
#define O2SCL_INTERP_H

#include <iostream>
#include <string>

// for gsl_linalg_solve_symm_tridiag for cspline_interp
#include <gsl/gsl_linalg.h>

#include <o2scl/err_hnd.h>
#include <o2scl/collection.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/search_vec.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Base low-level interpolation class

      The descendants of this class are intended to be fast
      interpolation routines for increasing functions, leaving the
      some error handling, user-friendliness, and other more 
      complicated improvements for other classes.

      For any pair of vectors x and y into which you would like to
      interpolate, you need to call allocate() and init() first, and then 
      the interpolation functions, and then free(). If the next pair
      of vectors has the same size, then you need only to call init()
      before the next call to an interpolation function. If the vectors
      do not change, then you may call the interpolation functions in
      succession.

      All of the descendants are based on the GSL interpolation 
      routines and give identical results.

      \future These might work for decreasing functions by just
      replacing calls to search_vec::bsearch_inc() with
      search_vec::bsearch_dec(). If this is the case, then this should
      be rewritten accordingly. (I think I might have removed the
      acceleration)
  */
  template<class vec_t> class base_interp {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The binary search object
    search_vec<vec_t> sv;

    /** 
	\brief An internal function to assist in computing the 
	integral for both the cspline and Akima types
    */
    double integ_eval(double ai, double bi, double ci, double di, double xi, 
		      double a, double b) {
      const double r1 = a - xi;
      const double r2 = b - xi;
      const double r12 = r1 + r2;
      const double bterm = 0.5 * bi * r12;
      const double cterm = (1.0 / 3.0) * ci * (r1 * r1 + r2 * r2 + r1 * r2);
      const double dterm = 0.25 * di * r12 * (r1 * r1 + r2 * r2);
      
      return (b - a) * (ai + bterm + cterm + dterm);
    }

#endif

  public:

    virtual ~base_interp() {}

    /** 
	\brief The minimum size of the vectors to interpolate between

	This needs to be set in the constructor of the children for
	access by the class user
    */
    size_t min_size;

    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) { return 0; }

    /// Free allocated memory
    virtual int free() { return 0; }

    /// Initialize interpolation routine
    virtual int init(const vec_t &x, const vec_t &y, size_t size) {
      return 0;
    }
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &y0) {
      return 0;
    }
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x, const vec_t &y, size_t size,
		      double x0, double &dydx) {
      return 0;
    }
    
    /** \brief Give the value of the second derivative 
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2) {
      return 0;
    }
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x, const vec_t &y, size_t size, 
		      double a, double b, double &result) {
      return 0;
    }
  };
  
  /** 
      \brief Linear interpolation (GSL)

      Linear interpolation requires no calls to allocate(), free() or
      init(), as there is no internal storage required.
  */
  template<class vec_t> class linear_interp : public base_interp<vec_t> {
    
  public:

    linear_interp() {
      this->min_size=2;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      size_t index;

      index=this->sv.bsearch_inc(x,x_array,0,size-1);
      
      /* evaluate */
      x_lo = x_array[index];
      x_hi = x_array[index + 1];
      y_lo = y_array[index];
      y_hi = y_array[index + 1];
      dx = x_hi - x_lo;
      if (dx > 0.0)
	{
	  y = y_lo + (x - x_lo) / dx * (y_hi - y_lo);
	  return gsl_success;
	}
      else
	{
	  y = 0.0;
	  return GSL_EINVAL;
	}
    }
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size,
		      double x, double &dydx) {
      
      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      double dy;
      size_t index;
      
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
      
      /* evaluate */
      x_lo = x_array[index];
      x_hi = x_array[index + 1];
      y_lo = y_array[index];
      y_hi = y_array[index + 1];
      dx = x_hi - x_lo;
      dy = y_hi - y_lo;
      if (dx > 0.0)
	{
	  dydx = dy / dx;;
	  return gsl_success;
	}
      else
	{
	  dydx = 0.0;
	  return GSL_EINVAL;
	}
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2) {
      d2ydx2=0.0;
      return gsl_success;
    }


    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, 
		      double &result) {
      size_t i, index_a, index_b;
      
      index_a=this->sv.bsearch_inc(a,x_array,0,size-1);
      index_b=this->sv.bsearch_inc(b,x_array,0,size-1);
      
      /* endpoints span more than one interval */
      
      result = 0.0;
      
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {
	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double y_hi = y_array[i + 1];
	const double dx = x_hi - x_lo;
	
	if(dx != 0.0) {
	  if (i == index_a || i == index_b)
	    {
	      double x1 = (i == index_a) ? a : x_lo;
	      double x2 = (i == index_b) ? b : x_hi;
	      const double D = (y_hi-y_lo)/dx;
	      result += (x2-x1) * (y_lo + 0.5*D*((x2-x_lo)+(x1-x_lo)));
	    }
	  else
	    {
	      result += 0.5 * dx * (y_lo + y_hi);
	    }
	}
      }
      
      return gsl_success;
    }
  };
  
  /** 
      \brief Cubic spline interpolation (GSL)
  */
  template<class vec_t> class cspline_interp : public base_interp<vec_t> {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for cubic spline interpolation
    //@{
    double *c;
    double *g;
    double *diag;
    double *offdiag;
    //@}

    /// True for periodic boundary conditions
    bool peri;

    /// Compute coefficients for cubic spline interpolation
    void coeff_calc(const double c_array[], double dy, double dx, 
		    size_t index,  double * b, double * c2, double * d) {
      const double c_i = c_array[index];
      const double c_ip1 = c_array[index + 1];
      *b = (dy / dx) - dx * (c_ip1 + 2.0 * c_i) / 3.0;
      *c2 = c_i;
      *d = (c_ip1 - c_i) / (3.0 * dx);
    }

#endif
    
  public:

    /** \brief Create a base interpolation object with natural or
	periodic boundary conditions
    */
    cspline_interp(bool periodic=false) {
      
      peri=periodic;
      if (peri) this->min_size=2;
      else this->min_size=3;
      
    }

    virtual ~cspline_interp() {}
    
    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      c = (double *) malloc (size * sizeof (double));
      if (c == 0) {
	set_err2_ret("Failed to allocate space for c",
		     " in cspline_interp::allocate().",gsl_enomem);
      }

      g = (double *) malloc (size * sizeof (double));
      if (g == 0) {
	std::free(c);
	set_err2_ret("Failed to allocate space for g",
		     " in cspline_interp::allocate().",gsl_enomem);
      }

      diag = (double *) malloc (size * sizeof (double));
      if (diag == 0) {
	std::free(g);
	std::free(c);
	set_err2_ret("Failed to allocate space for diag",
		     " in cspline_interp::allocate().",gsl_enomem);
      }
      
      offdiag = (double *) malloc (size * sizeof (double));
      if (offdiag == 0) {
	std::free(diag);
	std::free(g);
	std::free(c);
	set_err2_ret("Failed to allocate space for offdiag",
		     " in cspline_interp::allocate().",gsl_enomem);
      }
      
      return 0;
    }
    
    /// Initialize interpolation routine
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {

      if (peri) {

	/// Periodic boundary conditions
	 
	size_t i;
	size_t num_points = size;
	/* Engeln-Mullges + Uhlig "n" */
	size_t max_index = num_points - 1;  
	/* linear system is sys_size x sys_size */
	size_t sys_size = max_index;    
	
	if (sys_size == 2) {
	  /* solve 2x2 system */
	  
	  const double h0 = xa[1] - xa[0];
	  const double h1 = xa[2] - xa[1];
	  const double h2 = xa[3] - xa[2];
	  const double A = 2.0*(h0 + h1);
	  const double B = h0 + h1;
	  double gx[2];
	  double det;
	  
	  gx[0] = 3.0 * ((ya[2] - ya[1]) / h1 - (ya[1] - ya[0]) / h0);
	  gx[1] = 3.0 * ((ya[1] - ya[2]) / h2 - (ya[2] - ya[1]) / h1);
	  
	  det = 3.0 * (h0 + h1) * (h0 + h1);
	  c[1] = ( A * gx[0] - B * gx[1])/det;
	  c[2] = (-B * gx[0] + A * gx[1])/det;
	  c[0] =  c[2];
	  
	  return gsl_success;

	} else {
	  
	  for (i = 0; i < sys_size-1; i++) {
	    const double h_i       = xa[i + 1] - xa[i];
	    const double h_ip1     = xa[i + 2] - xa[i + 1];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 - ydiff_i * g_i);
	  }
	  
	  i = sys_size - 1;
	  {
	    const double h_i       = xa[i + 1] - xa[i];
	    const double h_ip1     = xa[1] - xa[0];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[1] - ya[0];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 - ydiff_i * g_i);
	  }
	  
	  {
	    gsl_vector_view g_vec = gsl_vector_view_array( g, sys_size);
	    gsl_vector_view diag_vec = 
	      gsl_vector_view_array( diag,sys_size);
	    gsl_vector_view offdiag_vec = 
	      gsl_vector_view_array( offdiag, sys_size);
	    gsl_vector_view solution_vec = 
	      gsl_vector_view_array (( c) + 1, sys_size);
	    int status = 
	      gsl_linalg_solve_symm_cyc_tridiag(&diag_vec.vector,
						&offdiag_vec.vector,
						&g_vec.vector,
						&solution_vec.vector);
	    c[0] =  c[max_index];
	    
	    return status;
	  }
	}

      } else {

	/// Natural boundary conditions

	size_t i;
	size_t num_points = size;
	size_t max_index = num_points - 1;
	size_t sys_size = max_index - 1;

	c[0] = 0.0;
	c[max_index] = 0.0;

	for (i = 0; i < sys_size; i++)
	  {
	    const double h_i   = xa[i + 1] - xa[i];
	    const double h_ip1 = xa[i + 2] - xa[i + 1];
	    const double ydiff_i   = ya[i + 1] - ya[i];
	    const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
	    const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
	    const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
	    offdiag[i] = h_ip1;
	    diag[i] = 2.0 * (h_ip1 + h_i);
	    g[i] = 3.0 * (ydiff_ip1 * g_ip1 -  ydiff_i * g_i);
	  }

	if (sys_size == 1)
	  {
	    c[1] =  g[0] /  diag[0];
	    return gsl_success;
	  }
	else
	  {
	    gsl_vector_view g_vec = gsl_vector_view_array( g, sys_size);
	    gsl_vector_view diag_vec = gsl_vector_view_array( diag, sys_size);
	    gsl_vector_view offdiag_vec = 
	      gsl_vector_view_array( offdiag, sys_size - 1);
	    gsl_vector_view solution_vec = 
	      gsl_vector_view_array (( c) + 1, sys_size);
	    
	    int status = gsl_linalg_solve_symm_tridiag(&diag_vec.vector, 
						       &offdiag_vec.vector, 
						       &g_vec.vector, 
						       &solution_vec.vector);
	    return status;
	  }
      }
    }
      
    
    /// Free allocated memory
    virtual int free() {
      std::free(c);
      std::free(g);
      std::free(diag);
      std::free(offdiag);

      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      double x_lo, x_hi;
      double dx;
      size_t index;
      
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0)
	{
	  const double y_lo = y_array[index];
	  const double y_hi = y_array[index + 1];
	  const double dy = y_hi - y_lo;
	  double delx = x - x_lo;
	  double b_i, c_i, d_i; 
	  coeff_calc( c, dy, dx, index,  &b_i, &c_i, &d_i);
	  y = y_lo + delx * (b_i + delx * (c_i + delx * d_i));
	  return gsl_success;
	}
      else
	{
	  y = 0.0;
	  return GSL_EINVAL;
	}
    }


    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0)
	{
	  const double y_lo = y_array[index];
	  const double y_hi = y_array[index + 1];
	  const double dy = y_hi - y_lo;
	  double delx = x - x_lo;
	  double b_i, c_i, d_i; 
	  coeff_calc( c, dy, dx, index,  &b_i, &c_i, &d_i);
	  dydx = b_i + delx * (2.0 * c_i + 3.0 * d_i * delx);
	  return gsl_success;
	}
      else
	{
	  dydx = 0.0;
	  return GSL_FAILURE;
	}
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
  
      /* evaluate */
      x_hi = x_array[index + 1];
      x_lo = x_array[index];
      dx = x_hi - x_lo;
      if (dx > 0.0)
	{
	  const double y_lo = y_array[index];
	  const double y_hi = y_array[index + 1];
	  const double dy = y_hi - y_lo;
	  double delx = x - x_lo;
	  double b_i, c_i, d_i;
	  coeff_calc( c, dy, dx, index,  &b_i, &c_i, &d_i);
	  d2ydx2 = 2.0 * c_i + 6.0 * d_i * delx;
	  return gsl_success;
	}
      else
	{
	  d2ydx2 = 0.0;
	  return GSL_FAILURE;
	}
    }

  public:
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, 
		      double &result) {

      size_t i, index_a, index_b;
  
      index_a=this->sv.bsearch_inc(a,x_array,0,size-1);
      index_b=this->sv.bsearch_inc(b,x_array,0,size-1);

      result = 0.0;
  
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {
	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double y_hi = y_array[i + 1];
	const double dx = x_hi - x_lo;
	const double dy = y_hi - y_lo;
	if(dx != 0.0) {
	  double b_i, c_i, d_i; 
	  coeff_calc( c, dy, dx, i,  &b_i, &c_i, &d_i);
      
	  if (i == index_a || i == index_b)
	    {
	      double x1 = (i == index_a) ? a : x_lo;
	      double x2 = (i == index_b) ? b : x_hi;
	      result += this->integ_eval(y_lo, b_i, c_i, d_i, x_lo, x1, x2);
	    }
	  else
	    {
	      result += dx * (y_lo + dx*(0.5*b_i + 
					 dx*(c_i/3.0 + 0.25*d_i*dx)));
	    }
	}
	else {
	  result = 0.0;
	  return GSL_FAILURE;
	}
      }
  
      return gsl_success;
    }

  };
  
  /** 
      \brief Cubic spline interpolation with periodic 
      boundary conditions (GSL)

      This is convenient to allow interpolation objects to
      be supplied as template parameters
  */
  template<class vec_t> class cspline_peri_interp : 
  public cspline_interp<vec_t> {

  public:
  cspline_peri_interp() : cspline_interp<vec_t>(true) {
    }

  };

  /** 
      \brief Akima spline interpolation (GSL)

      \future It appears that the interp() function below searches
      for indices slightly differently than the original GSL
      eval() function. This should be checked, as it might be
      slightly non-optimal in terms of speed (shouldn't matter for
      the accuracy). 
  */
  template<class vec_t> class akima_interp : public base_interp<vec_t> {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for Akima spline interpolation
    //@{
    double *b;
    double *c;
    double *d;
    double *um;
    //@}

    /// True for periodic boundary conditions
    bool peri;

    /// For initializing the interpolation
    void akima_calc(const vec_t &x_array, size_t size, double m[]) {
      size_t i;
      
      for (i = 0; i < (size - 1); i++) {

	const double NE = fabs (m[i + 1] - m[i]) + 
	  fabs (m[i - 1] - m[i - 2]);
	
	if (NE == 0.0) {
	  b[i] = m[i];
	  c[i] = 0.0;
	  d[i] = 0.0;
	} else {
	  const double h_i = x_array[i + 1] - x_array[i];
	  const double NE_next = fabs (m[i + 2] - m[i + 1]) + 
	    fabs (m[i] - m[i - 1]);
	  const double alpha_i = fabs (m[i - 1] - m[i - 2]) / NE;
	  double alpha_ip1;
	  double tL_ip1;
	  if (NE_next == 0.0) {
	    tL_ip1 = m[i];
	  } else {
	    alpha_ip1 = fabs (m[i] - m[i - 1]) / NE_next;
	    tL_ip1 = (1.0 - alpha_ip1) * m[i] + alpha_ip1 * m[i + 1];
	  }
	  b[i] = (1.0 - alpha_i) * m[i - 1] + alpha_i * m[i];
	  c[i] = (3.0 * m[i] - 2.0 * b[i] - tL_ip1) / h_i;
	  d[i] = (b[i] + tL_ip1 - 2.0 * m[i]) / (h_i * h_i);
	}
      }
    }
    
#endif
    
  public:

    /** \brief Create a base interpolation object with or without
	periodic boundary conditions
    */
    akima_interp(bool periodic=false) {
      peri=periodic;
      this->min_size=5;
    }

    virtual ~akima_interp() {}
    
    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      b = (double *) malloc (size * sizeof (double));
      if (b==0) {
	set_err2_ret("Failed to allocate space for c",
		     " in akima_interp::allocate().",gsl_enomem);
      }
      
      c = (double *) malloc (size * sizeof (double));
      if (c==0) {
	std::free(b);
	set_err2_ret("Failed to allocate space for g",
		     " in akima_interp::allocate().",gsl_enomem);
      }
      
      d = (double *) malloc (size * sizeof (double));
      if (d==0) {
	std::free(b);
	std::free(c);
	set_err2_ret("Failed to allocate space for diag",
		     " in akima_interp::allocate().",gsl_enomem);
      }
      
      um = (double *) malloc ((size+4) * sizeof (double));
      if (um==0) {
	std::free(b);
	std::free(c);
	std::free(d);
	set_err2_ret("Failed to allocate space for offdiag",
		     " in akima_interp::allocate().",gsl_enomem);
      }
      
      return 0;
    }
    
    /// Initialize interpolation routine
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {
      
      if (peri) {
	
	/// Periodic boundary conditions

	double *m=um+2;
	size_t i;
	for (i = 0; i <= size - 2; i++)
	  {
	    m[i] = (ya[i + 1] - ya[i]) / (xa[i + 1] - xa[i]);
	  }
	
	/* periodic boundary conditions */
	m[-2] = m[size - 1 - 2];
	m[-1] = m[size - 1 - 1];
	m[size - 1] = m[0];
	m[size] = m[1];
	
	akima_calc (xa,size, m);
	
      } else {

	/// Non-periodic boundary conditions

	double *m=um+2;
	size_t i;
	for (i = 0; i <= size - 2; i++) {
	  m[i] = (ya[i + 1] - ya[i]) / (xa[i + 1] - xa[i]);
	}
	
	/* non-periodic boundary conditions */
	m[-2] = 3.0 * m[0] - 2.0 * m[1];
	m[-1] = 2.0 * m[0] - m[1];
	m[size - 1] = 2.0 * m[size - 2] - m[size - 3];
	m[size] = 3.0 * m[size - 2] - 2.0 * m[size - 3];
	
	akima_calc (xa, size, m);

      }
      return 0;
    }
      
    
    /// Free allocated memory
    virtual int free() {
      std::free(b);
      std::free(c);
      std::free(d);
      std::free(um);

      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      size_t index;
      
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
  
      const double x_lo = x_array[index];
      const double delx = x - x_lo;
      const double bb = b[index];
      const double cc = c[index];
      const double dd = d[index];
      y = y_array[index] + delx * (bb + delx * (cc + dd * delx));
      return gsl_success;
    }


    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      size_t index;
  
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
      double x_lo = x_array[index];
      double delx = x - x_lo;
      double bb = b[index];
      double cc = c[index];
      double dd = d[index];
      dydx = bb + delx * (2.0 * cc + 3.0 * dd * delx);
      return gsl_success;

    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      size_t index;
  
      index=this->sv.bsearch_inc(x,x_array,0,size-1);
  
      const double x_lo = x_array[index];
      const double delx = x - x_lo;
      const double cc = c[index];
      const double dd = d[index];
      d2ydx2 = 2.0 * cc + 6.0 * dd * delx;
      return gsl_success;

    }

    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double aa, double bb, 
		      double &result) {

      size_t i, index_a, index_b;
  
      index_a=this->sv.bsearch_inc(aa,x_array,0,size-1);
      index_b=this->sv.bsearch_inc(bb,x_array,0,size-1);

      result = 0.0;
  
      /* interior intervals */

      for(i=index_a; i<=index_b; i++) {

	const double x_hi = x_array[i + 1];
	const double x_lo = x_array[i];
	const double y_lo = y_array[i];
	const double dx = x_hi - x_lo;

	if(dx != 0.0) {
	  
	  if (i==index_a || i==index_b) {
	    double x1 = (i==index_a) ? aa : x_lo;
	    double x2 = (i==index_b) ? bb : x_hi;
	    result += this->integ_eval(y_lo,b[i],c[i],d[i],x_lo,x1,x2);
	  } else {
	    result+=dx*(y_lo+dx*(0.5*b[i]+dx*(c[i]/3.0+0.25*d[i]*dx)));
	  }

	} else {
	  result = 0.0;
	  return GSL_FAILURE;
	}
      }
  
      return gsl_success;
    }

  };
  
  /** 
      \brief Akima spline interpolation with periodic 
      boundary conditions (GSL)
      
      This is convenient to allow interpolation objects to
      be supplied as template parameters
  */
  template<class vec_t> class akima_peri_interp : 
  public akima_interp<vec_t> {

  public:
  akima_peri_interp() : akima_interp<vec_t>(true) {
    }

  };

  /** 
      \brief Interpolation class
      
      This interpolation class is intended to be sufficiently general
      to handle any vector type. Interpolation of \ref ovector like
      objects is performed with the default template parameters, and
      \ref array_interp is provided for simple interpolation on
      C-style arrays.

      The type of interpolation to be performed can be specified using
      the set_type() function or in the constructor.  The default is
      cubic splines with natural boundary conditions.

      The class automatically handles decreasing arrays by converting
      from an object of type \c vec_t to an object ov type \c rvec_t.

      While \c vec_t may be any vector type which allows indexing via
      \c [], \c rvec_t must be a vector type which allows indexing
      and has a constructor with one of the two forms
      \verbatim
      rvec_t::rvec_t(vec_t &v);
      rvec_t::rvec_t(vec_t v);
      \endverbatim
      so that o2scl_interp can automatically "reverse" a vector if 
      necessary.
      
      It is important that different instances of o2scl_interp_vec
      and o2scl_interp not be given the same interpolation objects,
      as they will clash.
      
  */
  template<class vec_t=ovector_view, class rvec_t=ovector_const_reverse> 
    class o2scl_interp {
      
#ifndef DOXYGEN_INTERNAL
      
  protected:
      
  /// Pointer to base interpolation object
  base_interp<vec_t> *itp;
      
  /// Pointer to base interpolation object for reversed vectors
  base_interp<rvec_t> *ritp;
      
#endif
      
  public:
      
  /// Create with base interpolation objects \c it and \c rit
  o2scl_interp(base_interp<vec_t> &it, base_interp<rvec_t> &rit) {
    itp=&it;
    ritp=&rit;
  }

  /** \brief Create with base interpolation object \c it and use
      \ref def_ritp for reverse interpolation if necessary
  */
  o2scl_interp(base_interp<vec_t> &it) {
    itp=&it;
    ritp=&def_ritp;
  }
    
  /** \brief Create an interpolator using \ref def_itp and 
      \ref def_ritp
  */
  o2scl_interp() {
    itp=&def_itp;
    ritp=&def_ritp;
  }

  virtual ~o2scl_interp() {}

  /// Give the value of the function \f$ y(x=x_0) \f$ .
  virtual double interp(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) 
  {
	
    double ret=0.0;
      
    err_hnd->reset();
      
    if (n<itp->min_size) {
      set_err("Vector size too small in interp().",gsl_edom);
    }
      
    if (x[0]>x[n-1]) {

      // Decreasing case

      rvec_t ax(x);
      rvec_t ay(y);

      int r1=ritp->allocate(n);
      if (r1==0) {
	int r2=ritp->init(ax,ay,n);
	if (r2==0) {
	  ritp->interp(ax,ay,n,x0,ret);
	  ritp->free();
	}
      }

    } else {

      // Increasing case
	  
      int r1=itp->allocate(n);
      if (r1==0) {
	int r2=itp->init(x,y,n);
	if (r2==0) {
	  itp->interp(x,y,n,x0,ret);
	  itp->free();
	}
      }
	  
    }
      
    return ret;
  }		      
    
  /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
  virtual double deriv(const double x0, size_t n, const vec_t &x, 
		       const vec_t &y) 
  {
      
    double ret=0.0;
      
    err_hnd->reset();
      
    if (n<itp->min_size) {
      set_err("Vector size too small in interp().",gsl_edom);
    }
      
    if (x[0]>x[n-1]) {
	  
      rvec_t ax(x);
      rvec_t ay(y);

      int r1=ritp->allocate(n);
      if (r1==0) {
	int r2=ritp->init(ax,ay,n);
	if (r2==0) {
	  ritp->deriv(ax,ay,n,x0,ret);
	  ritp->free();
	}
      }
	  
    } else {

      // Increasing case
	  
      int r1=itp->allocate(n);
      if (r1==0) {
	int r2=itp->init(x,y,n);
	if (r2==0) {
	  itp->deriv(x,y,n,x0,ret);
	  itp->free();
	}
      }
    }
      
    return ret;
  }		      
    
  /** \brief Give the value of the second derivative  
      \f$ y^{\prime \prime}(x=x_0) \f$ .
  */
  virtual double deriv2(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) 
  {
      
    double ret=0.0;
      
    err_hnd->reset();
      
    if (n<itp->min_size) {

      set_err("Vector size too small in deriv2().",gsl_edom);

    }
      
    if (x[0]>x[n-1]) {
	  
      // Decreasing case
	    
      rvec_t ax(x);
      rvec_t ay(y);

      int r1=ritp->allocate(n);
      if (r1==0) {
	int r2=ritp->init(ax,ay,n);
	if (r2==0) {
	  ritp->deriv2(ax,ay,n,x0,ret);
	  ritp->free();
	}
      }

    } else {

      // Increasing case

      int r1=itp->allocate(n);
      if (r1==0) {
	int r2=itp->init(x,y,n);
	if (r2==0) {
	  itp->deriv2(x,y,n,x0,ret);
	  itp->free();
	}
      }

    }
      
    return ret;
  }		      
    
  /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
  virtual double integ(const double x1, const double x2, size_t n, 
		       const vec_t &x, const vec_t &y) 
  {
      
    double ret=0.0;
      
    err_hnd->reset();
      
    if (n<itp->min_size) {

      set_err("Vector size too small in integ().",gsl_edom);

    }
      
    if (x[0]>x[n-1]) {
	  
      // Decreasing case
	    
      rvec_t ax(x);
      rvec_t ay(y);

      int r1=ritp->allocate(n);
      if (r1==0) {
	int r2=ritp->init(ax,ay,n);
	if (r2==0) {
	  ritp->integ(ax,ay,n,x1,x2,ret);
	  ritp->free();
	}
      }

    } else {

      // Increasing case
	    
      int r1=itp->allocate(n);
      if (r1==0) {
	int r2=itp->init(x,y,n);
	if (r2==0) {
	  itp->integ(x,y,n,x1,x2,ret);
	  itp->free();
	}
      }

    }
      
    return ret;
  }		      
    
  /// Set base interpolation object
  int set_type(base_interp<vec_t> &it, base_interp<rvec_t> &rit) {
    itp=&it;
    ritp=&rit;
    return 0;
  }
      
  /** \brief Default base interpolation object (cubic spline with natural 
      boundary conditions)
  */
  cspline_interp<vec_t> def_itp;

  /** \brief Default base interpolation object for reversed vectors
      (cubic spline with natural boundary conditions)
  */
  cspline_interp<rvec_t> def_ritp;
      
  };
  
  /** 
      \brief Interpolation class for pre-specified vector
      
      This interpolation class is intended to be sufficiently general
      to handle any vector type. Interpolation of \ref ovector like
      objects is performed with the default template parameters, and
      \ref array_interp_vec is provided for simple interpolation on
      C-style arrays.

      The class automatically handles decreasing arrays by copying the
      old array to a reversed version. For several interpolations on
      the same data, copying the initial array can be faster than
      overloading operator[].

      The type of interpolation to be performed can be specified 
      using the set_type() function. The default is cubic splines
      with natural boundary conditions. 
      
      It is important that different instances of o2scl_interp_vec
      and o2scl_interp not be given the same interpolation objects,
      as they will clash.

      \todo Need to fix constructor to behave properly if init() fails.
      It should free the memory and set \c ln to zero.
  */
  template<class vec_t=ovector_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class o2scl_interp_vec {
      
#ifndef DOXYGEN_INTERNAL
      
  protected:
      
  /// Memory allocator for objects of type \c alloc_vec_t
  alloc_t ao;

  /// Pointer to base interpolation object
  base_interp<vec_t> *itp;
      
  /// True if the original array was increasing
  bool inc;

  /// Pointer to the user-specified x vector
  const vec_t *lx;

  /// Pointer to the user-specified x vector
  const vec_t *ly;

  /// Reversed version of x
  alloc_vec_t lrx;

  /// Reversed version of y
  alloc_vec_t lry;

  /// Size of user-specified vectors
  size_t ln;

#endif
      
  public:
    
  /// Create with base interpolation object \c it
  o2scl_interp_vec(base_interp<vec_t> &it, 
		   size_t n, const vec_t &x, const vec_t &y) {
    itp=&it;
    ln=0;
	
    if (n<itp->min_size) {
      set_err("Vector size too small in o2scl_interp_vec().",gsl_edom);
    } else {
      size_t iend=n-1;
	  
      if (x[0]>x[iend]) {
	    
	ao.allocate(lrx,n);
	ao.allocate(lry,n);

	for(size_t i=0;i<n;i++) {
	  lrx[iend-i]=x[i];
	  lry[iend-i]=y[i];
	}
	    
	int r1=itp->allocate(n);
	if (r1==0) {
	  itp->init(lrx,lry,n);
	  ln=n;
	  inc=false;
	}
	    
      } else {
	    
	int r1=itp->allocate(n);
	if (r1==0) {
	  itp->init(x,y,n);
	  ln=n;
	  lx=&x;
	  ly=&y;
	  inc=true;
	}
	    
      }

    } 
	
  }
      
  virtual ~o2scl_interp_vec() {
    if (ln>0) {
      itp->free();
    }
    if (inc==false) {
      ao.free(lrx);
      ao.free(lry);
    }
  }
    
  /// Give the value of the function \f$ y(x=x_0) \f$ .
  virtual double interp(const double x0) {
    double ret=0.0;
    if (ln>0) {
      if (inc) itp->interp(*lx,*ly,ln,x0,ret);
      else itp->interp(lrx,lry,ln,x0,ret);
    }
    return ret;
  }		      
    
  /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
  virtual double deriv(const double x0) {
    double ret=0.0;
    if (ln>0) {
      if (inc) itp->deriv(*lx,*ly,ln,x0,ret);
      else itp->deriv(lrx,lry,ln,x0,ret);
    }
    return ret;
  }		      
    
  /** \brief Give the value of the second derivative  
      \f$ y^{\prime \prime}(x=x_0) \f$ .
  */
  virtual double deriv2(const double x0) {
    double ret=0.0;
    if (ln>0) {
      if (inc) itp->deriv2(*lx,*ly,ln,x0,ret);
      else itp->deriv2(lrx,lry,ln,x0,ret);
    }
    return ret;
  }		      
    
  /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
  virtual double integ(const double x1, const double x2) {
    double ret=0.0;
    if (ln>0) {
      if (inc) itp->integ(*lx,*ly,ln,x1,x2,ret);
      else itp->integ(lrx,lry,ln,x1,x2,ret);
    }
    return ret;
  }		      
    
  /// Set base interpolation object
  int set_type(base_interp<vec_t> &it) {
    itp=&it;
    return 0;
  }
      
  /** \brief Default base interpolation object (cubic spline with natural 
      boundary conditions)
  */
  cspline_interp<vec_t> def_itp;
      
  };
  
  /** \brief A specialization of o2scl_interp for C-style double arrays
   */
  template<size_t n> class array_interp :
  public o2scl_interp<double[n],array_const_reverse<n> > {
    
  public:
    
    /// Create with base interpolation objects \c it and \c rit
  array_interp(base_interp<double[n]> &it, 
	       base_interp<array_const_reverse<n> > &rit)
    : o2scl_interp<double[n],array_const_reverse<n> >(it,rit) {
    }
    
    /** \brief Create with base interpolation object \c it and use
	the default base interpolation object for reversed arrays
    */
  array_interp(base_interp<double[n]> &it) : 
    o2scl_interp<double[n],array_const_reverse<n> >(it) {
    }
      
    /** \brief Create an interpolator using the default base
	interpolation objects
    */
  array_interp() : 
    o2scl_interp<double[n],array_const_reverse<n> >() {
    }

  };

  /** 
      \brief A specialization of o2scl_interp_vec for C-style arrays
  */
  template<class arr_t> class array_interp_vec : 
  public o2scl_interp_vec<arr_t,arr_t,array_alloc<arr_t> > {
    
  public:
    
    /// Create with base interpolation object \c it
  array_interp_vec(base_interp<arr_t> &it, 
		   size_t nv, const arr_t &x, const arr_t &y) :
    o2scl_interp_vec<arr_t,arr_t,array_alloc<arr_t> >(it,nv,x,y) {
    }
    
  };
  
#ifndef DOXYGENP
}
#endif

#endif
