/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_VECTOR_H
#define O2SCL_VECTOR_H

/** \file vector.h
    \brief File for generic vector functions

    For overloaded operators involving vectors and matrices, see \ref
    vec_arith.h . For statistics operations not included here, see
    \ref vec_stats.h in the directory \c src/other . For functions and
    classes which are specific to C-style arrays, see \ref array.h .
    Also related are the matrix output functions, \ref matrix_out(),
    \ref matrix_cx_out_paren(), and \ref matrix_out_paren() which are
    defined in \ref columnify.h because they utilize the class \ref
    columnify to format the output.
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <gsl/gsl_ieee_utils.h>
#include <gsl/gsl_sort.h>

#ifndef DOXYGENP
namespace o2scl
{
#endif
  
  /** 
      \brief Naive vector copy

      \note This ordering is reversed from the GSL function
      \c gsl_vector_memcpy . This is to be used with
      \code
      vector_copy(N,source,destination);
      \endcode
      instead of
      \code
      gsl_vector_memcpy(destination,source);
      \endcode
  */
  template<class vec_t, class vec2_t> 
    void vector_copy(size_t N, vec_t &src, vec2_t &dest) {
    for(size_t i=0;i<N;i++) dest[i]=src[i];
  }
  
  /** 
      \brief Naive matrix copy

      \note This ordering is reversed from the GSL function
      \c gsl_matrix_memcpy . This is to be used with
      \code
      matrix_copy(N,source,destination);
      \endcode
      instead of
      \code
      gsl_matrix_memcpy(destination,source);
      \endcode
   */
  template<class mat_t, class mat2_t> 
    void matrix_copy(size_t M, size_t N, mat_t &src, mat2_t &dest) {
    for(size_t i=0;i<M;i++) {
      for(size_t j=0;j<N;j++) {
	dest[i][j]=src[i][j];
      }
    }
  }

  /** 
      \brief GSL complex vector copy

      \future At present this works only with complex types based 
      directly on the GSL complex format. This could be improved.
  */
  template<class vec_t, class vec2_t> 
    void vector_cx_copy_gsl(size_t N, vec_t &src, vec2_t &dest) {
    for(size_t i=0;i<N;i++) {
      dest[i].dat[0]=src[i].dat[0];
      dest[i].dat[1]=src[i].dat[1];
    }
  }
  
  /** 
      \brief GSL complex matrix copy

      \future At present this works only with complex types based 
      directly on the GSL complex format. This could be improved.
  */
  template<class mat_t, class mat2_t> 
    void matrix_cx_copy_gsl(size_t M, size_t N, mat_t &src, mat2_t &dest) {
    for(size_t i=0;i<M;i++) {
      for(size_t j=0;j<N;j++) {
	dest[i][j].dat[0]=src[i][j].dat[0];
	dest[i][j].dat[1]=src[i][j].dat[1];
      }
    }
  }

  /** 
      \brief Output a vector to a stream
      
      No trailing space is output after the last element, and an
      endline is output only if \c endline is set to \c true.  If the
      parameter \c n is zero, this function silently does nothing.

      Note that the \o2 vector classes also have their own
      \c operator<<() defined for them.
  */
  template<class vec_t> 
    int vector_out(std::ostream &os, size_t n, vec_t &v, 
		   bool endline=false) {
    // This next line is important since n-1 is not well-defined if n=0
    if (n==0) return 0;
    for(size_t i=0;i<n-1;i++) os << v[i] << " ";
    os << v[n-1];
    if (endline) os << std::endl;
    return 0;
  }
  
  /** 
      \brief Provide a downheap() function for vector_sort()
  */
  template<class data_t, class vec_t>
    void sort_downheap(vec_t &data, const size_t N, size_t k) {
    
    data_t v=data[k];
    
    while (k<=N/2) {
      size_t j=2*k;
      
      if (j<N && data[j] < data[j+1]) j++;
      if (!(v < data[j])) break;
      data[k]=data[j];
      k=j;
    }
    
    data[k]=v;
  }

  /** 
      \brief Sort a vector

      This is a generic sorting template function. It will work for
      any types \c data_t and \c vec_t for which
      - \c data_t has an <tt>operator=</tt>
      - \c data_t has a less than operator to compare elements
      - <tt>vec_t::operator[]</tt> returns a reference
      to an object of type \c data_t 
      
      In particular, it will work with \ref ovector, \ref uvector, 
      \ref ovector_int, \ref uvector_int (and other related O2scl
      vector classes), the STL template class <tt>std::vector</tt>,
      and arrays and pointers of numeric, character, and string
      objects.

      For example,
      \code
      std::string list[3]={"dog","cat","fox"};
      vector_sort<std::string, std::string[3]>(3,list);
      \endcode
  */
  template<class data_t, class vec_t>
    int vector_sort(const size_t n, vec_t &data) {
    
    size_t N;
    size_t k;
    
    if (n==0) return 0;
    
    N=n-1;
    k=N/2;
    k++;
    do {
      k--;
      sort_downheap<data_t,vec_t>(data,N,k);
    } while (k > 0);
    
    while (N > 0) {
      data_t tmp=data[0];
      data[0]=data[N];
      data[N]=tmp;
      N--;
      sort_downheap<data_t,vec_t>(data,N,0);
    }
    
    return 0;
  }

  /** 
      \brief "Rotate" a vector so that the kth element is now the beginning

      This is a generic template function which will work for
      any types \c data_t and \c vec_t for which
      - \c data_t has an <tt>operator=</tt>
      - <tt>vec_t::operator[]</tt> returns a reference
      to an object of type \c data_t 
  */
  template<class data_t, class vec_t>
    int vector_rotate(const size_t n, vec_t &data, size_t k) {

    data_t *tmp=new data_t[n];
    for(size_t i=0;i<n;i++) {
      tmp[i]=data[(i+k)%n];
    }
    for(size_t i=0;i<n;i++) {
      data[i]=tmp[i];
    }
    delete[] tmp;
    
    return 0;
  }

  /// Compute the maximum of the first \c n elements of a vector
  template<class vec_t>
    double vector_max(const size_t n, vec_t &data) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_max().",gsl_efailed);
    }
    double max=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]>max) {
	max=data[i];
      }
    }
    return max;
  }

  /// Compute the minimum of the first \c n elements of a vector
  template<class vec_t>
    double vector_min(const size_t n, vec_t &data) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_min().",gsl_efailed);
    }
    double min=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]<min) {
	min=data[i];
      }
    }
    return min;
  }

  /// Compute the minimum and maximum of the first \c n elements of a vector
  template<class vec_t>
    int vector_minmax(const size_t n, vec_t &data, double &min, double &max) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_min().",gsl_efailed);
    }
    min=data[0];
    max=min;
    for(size_t i=1;i<n;i++) {
      if (data[i]<min) {
	min=data[i];
      }
      if (data[i]>max) {
	max=data[i];
      }
    }
    return 0;
  }

  /// Compute the maximum of the first \c n elements of a vector
  template<class vec_t>
    size_t vector_max_index(const size_t n, vec_t &data, double &max) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_max().",gsl_efailed);
    }
    size_t ix=0;
    max=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]>max) {
	max=data[i];
	ix=i;
      }
    }
    return ix;
  }

  /// Compute the minimum of the first \c n elements of a vector
  template<class vec_t>
    int vector_min_index(const size_t n, vec_t &data, double &min) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_min().",gsl_efailed);
    }
    size_t ix=0;
    min=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]<min) {
	min=data[i];
	ix=i;
      }
    }
    return ix;
  }

  /// Compute the minimum and maximum of the first \c n elements of a vector
  template<class vec_t>
    int vector_minmax_index(const size_t n, vec_t &data, double &min, 
			    size_t &ix, double &max, size_t &ix2) {
    
    if (n==0) {
      set_err_ret("Sent size=0 to vector_min().",gsl_efailed);
    }
    ix=0;
    ix2=0;
    min=data[0];
    max=min;
    for(size_t i=1;i<n;i++) {
      if (data[i]<min) {
	min=data[i];
	ix=i;
      }
      if (data[i]>max) {
	max=data[i];
	ix2=i;
      }
    }
    return 0;
  }

  /** 
      \brief Compute the sum of the first \c n elements of a vector

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
  */
  template<class vec_t>
    double vector_sum(const size_t n, vec_t &data) {
    
    double sum=0;
    for(size_t i=0;i<n;i++) {
      sum+=data[i];
    }
    return sum;
  }

  /** \brief Reverse a vector
   */
  template<class data_t, class vec_t>
    int vector_reverse(const size_t n, vec_t &data) {
    data_t tmp;

    for(size_t i=0;i<n/2;i++) {
      tmp=data[n-1-i];
      data[n-1-i]=data[i];
      data[i]=tmp;
    }
    return 0;
  }

#ifndef DOXYGENP
}
#endif

#endif
