/*
  -------------------------------------------------------------------

  Copyright (C) 2006, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/cfl6_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_arith;
using namespace o2scl_const;

cfl6_eos::cfl6_eos() {
  KD=0.0;
  
  iprop6.allocate(mat_size,mat_size);
  eivec6.allocate(mat_size,mat_size);
  dipdgapu.allocate(mat_size,mat_size);
  dipdgapd.allocate(mat_size,mat_size);
  dipdgaps.allocate(mat_size,mat_size);
  dipdqqu.allocate(mat_size,mat_size);
  dipdqqd.allocate(mat_size,mat_size);
  dipdqqs.allocate(mat_size,mat_size);
  eval6.allocate(mat_size);
  w6=gsl_eigen_hermv_alloc(mat_size);
  
  kdlimit=1.0e-6;
}

cfl6_eos::~cfl6_eos() {
  iprop6.free();
  eivec6.free();
  dipdgapu.free();
  dipdgapd.free();
  dipdgaps.free();
  dipdqqu.free();
  dipdqqd.free();
  dipdqqs.free();
  eval6.free();
  gsl_eigen_hermv_free(w6);

}

int cfl6_eos::set_masses() {
  if (fixed_mass) {
    up->ms=up->m+KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m+KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m+KD/4.0/GD/GD*strange->del*strange->del;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq+
      KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq+
      KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq+
      KD/4.0/GD/GD*strange->del*strange->del;
  }
  if (up->ms<up->m) up->ms=up->m;
  if (down->ms<down->m) down->ms=down->m;
  if (strange->ms<strange->m) strange->ms=strange->m;
  return 0;
}

int cfl6_eos::calc_eq_temp_p(quark &u, quark &d, quark &s,
			      double &qq1, double &qq2, double &qq3, 
			      double &gap1, double &gap2, double &gap3, 
			      double mu3, double mu8, 
			      double &n3, double &n8, 
			      thermo &qb, const double ltemper) {
  
  if (fabs(KD)<kdlimit) {
    return cfl_njl_eos::calc_eq_temp_p(u,d,s,qq1,qq2,qq3,gap1,gap2,gap3,
					mu3,mu8,n3,n8,qb,ltemper);
  }
  
  up=&u;
  down=&d;
  strange=&s;

  if (fromqq==false) {
    set_err("cfl6_eos does not work with fromqq=false.",1);
    return 1;
  }

  set_masses();
  
  // static variables for communication with tpot
  temper=ltemper;
  smu3=mu3;
  smu8=mu8;

  // Compute the integrals

  ovector res(13);
  double err2;
  integ_err(0.0,L,13,res,err2);
  qb.pr=-res[0];
  qb.en=res[1];
  u.n=-res[2];
  d.n=-res[3];
  s.n=-res[4];
  qq1=res[5];
  qq2=res[6];
  qq3=res[7];
  gap1=res[8];
  gap2=res[9];
  gap3=res[10];
  n3=res[11];
  n8=res[12];

  // Add the remaining terms and their corresponding contributions
  // to the derivatives

  qb.pr-=B0;
  
  if (!fixed_mass) {
    qb.pr-=u.qq*u.qq*2.0*G;
    qb.pr-=d.qq*d.qq*2.0*G;
    qb.pr-=s.qq*s.qq*2.0*G;
    qb.pr+=u.qq*d.qq*s.qq*4.0*K;
    qb.pr-=KD/GD/GD/2.0*(u.del*u.del*u.qq+d.del*d.del*d.qq+
			 s.del*s.del*s.qq);
  }
  qb.pr-=u.del*u.del/4.0/GD;
  qb.pr-=d.del*d.del/4.0/GD;
  qb.pr-=s.del*s.del/4.0/GD;

  qq1+=4.0*G*u.qq-4.0*K*d.qq*s.qq+KD/GD/GD/2.0*u.del*u.del;
  qq2+=4.0*G*d.qq-4.0*K*u.qq*s.qq+KD/GD/GD/2.0*d.del*d.del;
  qq3+=4.0*G*s.qq-4.0*K*u.qq*d.qq+KD/GD/GD/2.0*s.del*s.del;
  
  gap1+=u.del/2.0/GD+KD/GD/GD*u.del*u.qq;
  gap2+=d.del/2.0/GD+KD/GD/GD*d.del*d.qq;
  gap3+=s.del/2.0/GD+KD/GD/GD*s.del*s.qq;

  // Compute the energy densities 
  
  u.ed=-u.pr+u.n*u.mu;
  d.ed=-d.pr+d.n*d.mu;
  s.ed=-s.pr+s.n*s.mu;
  qb.ed=qb.en*temper-qb.pr+u.mu*u.n+d.mu*d.n+s.mu*s.n;

  return 0;
}

int cfl6_eos::integrands(double p, double res[]) {
  
  if (fabs(KD)<kdlimit) {
    return cfl_njl_eos::integrands(p,res);
  }

  if (integ_test) {
    res[0]=sin(p);
    res[1]=sin(p);
    res[2]=cos(p);
    return 0;
  }

  int k;
  double egv[36];
  double dedmuu[36], dedmud[36], dedmus[36];
  double dedqqu[36], dedqqd[36], dedqqs[36], deds[36];
  double dedd[36], dedu[36], dedmu3[36], dedmu8[36];
  
  eigenvalues6(p,smu3,smu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,dedqqs,
	       dedu,dedd,deds,dedmu3,dedmu8);
  
  res[0]=0.0;
  res[1]=0.0;
  res[2]=0.0;
  res[3]=0.0;
  res[4]=0.0;
  res[5]=0.0;
  res[6]=0.0;
  res[7]=0.0;
  res[8]=0.0;
  res[9]=0.0;
  res[10]=0.0;
  res[11]=0.0;
  res[12]=0.0;
  for (k = 0; k < 36; k++) {
    if (zerot) {
      res[0]+=fabs(egv[k]/2.0);
      if (egv[k]>0.0) {
	res[2]+=dedmuu[k]/2.0;
	res[3]+=dedmud[k]/2.0;
	res[4]+=dedmus[k]/2.0;
	res[5]+=dedqqu[k]/2.0;
	res[6]+=dedqqd[k]/2.0;
	res[7]+=dedqqs[k]/2.0;
	res[8]+=dedu[k]/2.0;
	res[9]+=dedd[k]/2.0;
	res[10]+=deds[k]/2.0;
	res[11]+=dedmu3[k]/2.0;
	res[12]+=dedmu8[k]/2.0;
      } else {
	res[2]-=dedmuu[k]/2.0;
	res[3]-=dedmud[k]/2.0;
	res[4]-=dedmus[k]/2.0;
	res[5]-=dedqqu[k]/2.0;
	res[6]-=dedqqd[k]/2.0;
	res[7]-=dedqqs[k]/2.0;
	res[8]-=dedu[k]/2.0;
	res[9]-=dedd[k]/2.0;
	res[10]-=deds[k]/2.0;
	res[11]-=dedmu3[k]/2.0;
	res[12]-=dedmu8[k]/2.0;
      }
    } else {
      res[0]+=egv[k]/2.0;
      res[1]+=egv[k];
      res[2]+=dedmuu[k]/2.0;
      res[3]+=dedmud[k]/2.0;
      res[4]+=dedmus[k]/2.0;
      res[5]+=dedqqu[k]/2.0;
      res[6]+=dedqqd[k]/2.0;
      res[7]+=dedqqs[k]/2.0;
      res[8]+=dedu[k]/2.0;
      res[9]+=dedd[k]/2.0;
      res[10]+=deds[k]/2.0;
      res[11]+=dedmu3[k]/2.0;
      res[12]+=dedmu8[k]/2.0;
      if ((temper==0.0 && egv[k]<0.0) || egv[k]/temper<-30.0) {
	res[0]-=egv[k];
	res[2]-=dedmuu[k];
	res[3]-=dedmud[k];
	res[4]-=dedmus[k];
	res[5]-=dedqqu[k];
	res[6]-=dedqqd[k];
	res[7]-=dedqqs[k];
	res[8]-=dedu[k];
	res[9]-=dedd[k];
	res[10]-=deds[k];
	res[11]-=dedmu3[k];
	res[12]-=dedmu8[k];
      } else if (temper!=0.0 && egv[k]/temper<30.0) {
	res[0]+=temper*log(1.0+exp(-egv[k]/temper));
	res[2]-=1.0/(1.0+exp(egv[k]/temper))*dedmuu[k];
	res[3]-=1.0/(1.0+exp(egv[k]/temper))*dedmud[k];
	res[4]-=1.0/(1.0+exp(egv[k]/temper))*dedmus[k];
	res[5]-=1.0/(1.0+exp(egv[k]/temper))*dedqqu[k];
	res[6]-=1.0/(1.0+exp(egv[k]/temper))*dedqqd[k];
	res[7]-=1.0/(1.0+exp(egv[k]/temper))*dedqqs[k];
	res[8]-=1.0/(1.0+exp(egv[k]/temper))*dedu[k];
	res[9]-=1.0/(1.0+exp(egv[k]/temper))*dedd[k];
	res[10]-=1.0/(1.0+exp(egv[k]/temper))*deds[k];
	res[11]-=1.0/(1.0+exp(egv[k]/temper))*dedmu3[k];
	res[12]-=1.0/(1.0+exp(egv[k]/temper))*dedmu8[k];
      }
      if (temper!=0.0 && egv[k]/temper<30.0) {
	res[1]+=log(1.0+exp(-egv[k]/temper))+
	  egv[k]/temper/(1.0+exp(egv[k]/temper));
      }
    }
  }
  
  // Multiply by the integration weights and the 
  // Jacobian (momentum^2)
  res[0]*=-p*p*0.5/pi2;
  res[1]*=p*p*0.5/pi2;
  res[2]*=-p*p*0.5/pi2;
  res[3]*=-p*p*0.5/pi2;
  res[4]*=-p*p*0.5/pi2;
  res[5]*=-p*p*0.5/pi2;
  res[6]*=-p*p*0.5/pi2;
  res[7]*=-p*p*0.5/pi2;
  res[8]*=-p*p*0.5/pi2;
  res[9]*=-p*p*0.5/pi2;
  res[10]*=-p*p*0.5/pi2;
  res[11]*=-p*p*0.5/pi2;
  res[12]*=-p*p*0.5/pi2;

  return 0;
}

int cfl6_eos::test_derivatives(double lmom, double mu3, double mu8,
				test_mgr &t) {
  double egv[36];
  double dedmuu[36], dedmud[36], dedmus[36];
  double dedqqu[36], dedqqd[36], dedqqs[36], deds[36];
  double dedd[36], dedu[36], dedmu3[36], dedmu8[36];
  double h=1.0e-7;
  double d1[36], d2[36], tmpt[36];
  
  set_masses();

  // We have to do a lot of sorting here since the ordering of the
  // eigenvalues is not necessarily the same for subsequent calls to
  // the function eigenvalues6(). The same sorting should probably
  // be added to cfl_njl_eos::test_derivatives().

  // Check muu
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedmuu,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    up->mu+=h;
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],3.0e-4,"muu");
      }
    }
    up->mu-=h;
  }
  // Check mud
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedmud,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    down->mu+=h;
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],3.0e-4,"mud");
      }
    }
    down->mu-=h;
  }
  // Check mus
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedmus,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    strange->mu+=h;
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-3,"mus");
      }
    }
    strange->mu-=h;
  }
  // Check qqu
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedqqu,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    up->qq+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"qqu");
      }
    }
    up->qq-=h;
    set_masses();
  }
  // Check qqd
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedqqd,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    down->qq+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"qqd");
      }
    }
    down->qq-=h;
    set_masses();
  }
  // Check qqs
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedqqs,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    strange->qq+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"qqs");
      }
    }
    strange->qq-=h;
    set_masses();
  }

  // Check del matrices
  omatrix_cx gu1(36,36), gu2(36,36), dgg(36,36);

  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu1);
  matrix_cx_copy_gsl(36,36,dipdgapu,dgg);
  up->del+=h;
  set_masses();
  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu2);
  up->del-=h;
  set_masses();
  for(size_t i=0;i<36;i++) {
    for(size_t j=0;j<36;j++) {
      if (i>j) {
	t.test_rel(dgg[i][j].dat[0],(gu2[i][j].dat[0]-gu1[i][j].dat[0])/h,
		   1.0e-5,"gapud");
	t.test_rel(dgg[i][j].dat[1],(gu2[i][j].dat[1]-gu1[i][j].dat[1])/h,
		   1.0e-5,"gapud");
      }
    }
  }

  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu1);
  matrix_cx_copy_gsl(36,36,dipdgapd,dgg);
  down->del+=h;
  set_masses();
  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu2);
  down->del-=h;
  set_masses();
  for(size_t i=0;i<36;i++) {
    for(size_t j=0;j<36;j++) {
      if (i>j) {
	t.test_rel(dgg[i][j].dat[0],(gu2[i][j].dat[0]-gu1[i][j].dat[0])/h,
		   1.0e-5,"gapud");
	t.test_rel(dgg[i][j].dat[1],(gu2[i][j].dat[1]-gu1[i][j].dat[1])/h,
		   1.0e-5,"gapud");
      }
    }
  }

  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu1);
  matrix_cx_copy_gsl(36,36,dipdgaps,dgg);
  strange->del+=h;
  set_masses();
  make_matrices(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
  matrix_cx_copy_gsl(36,36,iprop6,gu2);
  strange->del-=h;
  set_masses();
  for(size_t i=0;i<36;i++) {
    for(size_t j=0;j<36;j++) {
      if (i>j) {
	t.test_rel(dgg[i][j].dat[0],(gu2[i][j].dat[0]-gu1[i][j].dat[0])/h,
		   1.0e-5,"gapud");
	t.test_rel(dgg[i][j].dat[1],(gu2[i][j].dat[1]-gu1[i][j].dat[1])/h,
		   1.0e-5,"gapud");
      }
    }
  }

  // Check delu
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedu,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    up->del+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"delu");
      }
    }
    up->del-=h;
    set_masses();
  }
  // Check deld
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedd,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    down->del+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"deld");
      }
    }
    down->del-=h;
    set_masses();
  }
  // Check dels
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,deds,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    strange->del+=h;
    set_masses();
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],1.0e-4,"dels");
      }
    }
    strange->del-=h;
    set_masses();
  }
  // Check mu3
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedmu3,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    mu3+=h;
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],5.0e-3,"mu3");
      }
    }
    mu3-=h;
  }
  // Check mu8
  {
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_copy(36,egv,d1);
    vector_copy(36,dedmu8,d2);
    vector_sort<double,double[36]>(36,d1);
    vector_sort<double,double[36]>(36,d2);
    mu8+=h;
    eigenvalues6(lmom,mu3,mu8,egv,dedmuu,dedmud,dedmus,dedqqu,dedqqd,
		 dedqqs,dedu,dedd,deds,dedmu3,dedmu8);
    vector_sort<double,double[36]>(36,egv);
    for(size_t i=0;i<36;i++) {
      tmpt[i]=(egv[i]-d1[i])/h;
    }
    vector_sort<double,double[36]>(36,tmpt);
    for(size_t i=0;i<36;i++) {
      if (fabs(d2[i])>1.0e-4) {
	t.test_rel(d2[i],tmpt[i],4.0e-4,"mu8");
      }
    }
    mu8-=h;
  }

  double res[13];
  double x1,x2,x3,x4,x5,x6,n3,n8;
  // Check derivative wrt muu in integrands() and calc_eq_temp_p()
  {
    integrands(lmom,res);
    double om1=res[0];
    double dmuu=res[2];
    up->mu+=h;
    integrands(lmom,res);
    double om2=res[0];
    up->mu-=h;
    t.test_rel((om2-om1)/h,dmuu,1.0e-6,"dmuu");
    
    calc_eq_temp_p(*up,*down,*strange,x1,x2,x3,x4,x5,x6,mu3,mu8,
		   n3,n8,*eos_thermo,4.0/hc_mev_fm);
    double pr1=eos_thermo->pr;
    double nu=up->n;
    up->mu+=h;
    calc_eq_temp_p(*up,*down,*strange,x1,x2,x3,x4,x5,x6,mu3,mu8,
		   n3,n8,*eos_thermo,4.0/hc_mev_fm);
    double pr2=eos_thermo->pr;
    up->mu-=h;
    t.test_rel((pr2-pr1)/h,nu,1.0e-5,"nup");
  }
    
  return 0;
}

int cfl6_eos::eigenvalues6(double lmom, double mu3, 
			    double mu8, double egv[36],
			    double dedmuu[36], double dedmud[36],
			    double dedmus[36], double dedqqu[36], 
			    double dedqqd[36], double dedqqs[36],
			    double dedu[36], double dedd[36],
			    double deds[36], double dedmu3[36],
			    double dedmu8[36]) {
  
  int k;
  const double mu=up->ms, md=down->ms, ms=strange->ms;
  const double muu=up->mu, mud=down->mu, mus=strange->mu;
  const double du=up->del, dd=down->del, ds=strange->del;
  const double qqu=up->qq, qqd=down->qq, qqs=strange->qq;

  double mur, mug, mub;
  mur=mu3+mu8/sqrt(3.0);
  mug=-mu3+mu8/sqrt(3.0);
  mub=-2.0*mu8/sqrt(3.0);
  
  for(k=0;k<36;k++) {
    egv[k]=0.0;
    dedmuu[k]=0.0;
    dedmud[k]=0.0;
    dedmus[k]=0.0;
    dedqqu[k]=0.0;
    dedqqd[k]=0.0;
    dedqqs[k]=0.0;
    dedu[k]=0.0;
    dedd[k]=0.0;
    deds[k]=0.0;
    dedmu3[k]=0.0;
    dedmu8[k]=0.0;
  }
  
  ovector mmuu(mat_size), mmud(mat_size), mmus(mat_size);
  ovector mmu(mat_size), mmd(mat_size), mms(mat_size);
  ovector mmur(mat_size), mmug(mat_size), mmub(mat_size);
  for(k=0;k<36;k++) {
    mmuu[k]=0.0;
    mmud[k]=0.0;
    mmus[k]=0.0;
    mmu[k]=0.0;
    mmd[k]=0.0;
    mms[k]=0.0;
    mmur[k]=0.0;
    mmug[k]=0.0;
    mmub[k]=0.0;
  }

  gsl_complex zero={{0.0,0.0}};
  iprop6.set_all(zero);
  dipdgapu.set_all(zero);
  dipdgapd.set_all(zero);
  dipdgaps.set_all(zero);
  dipdqqu.set_all(zero);
  dipdqqd.set_all(zero);
  dipdqqs.set_all(zero);
  
  /// Set diagonal elements

  iprop6.set_real(0,0,-mu-muu-mur);
  iprop6.set_real(1,1,mu-muu-mur);
  iprop6.set_real(2,2,-mu+muu+mur);
  iprop6.set_real(3,3,mu+muu+mur);
  mmuu[0]=-1.0;
  mmuu[1]=-1.0;
  mmuu[2]=1.0;
  mmuu[3]=1.0;
  mmur[0]=-1.0;
  mmur[1]=-1.0;
  mmur[2]=1.0;
  mmur[3]=1.0;
  mmu[0]=-1.0;
  mmu[1]=1.0;
  mmu[2]=-1.0;
  mmu[3]=1.0;

  iprop6.set_real(4,4,-mu-muu-mug);
  iprop6.set_real(5,5,mu-muu-mug);
  iprop6.set_real(6,6,-mu+muu+mug);
  iprop6.set_real(7,7,mu+muu+mug);
  mmuu[4]=-1.0;
  mmuu[5]=-1.0;
  mmuu[6]=1.0;
  mmuu[7]=1.0;
  mmug[4]=-1.0;
  mmug[5]=-1.0;
  mmug[6]=1.0;
  mmug[7]=1.0;
  mmu[4]=-1.0;
  mmu[5]=1.0;
  mmu[6]=-1.0;
  mmu[7]=1.0;
  
  iprop6.set_real(8,8,-mu-muu-mub);
  iprop6.set_real(9,9,mu-muu-mub);
  iprop6.set_real(10,10,-mu+muu+mub);
  iprop6.set_real(11,11,mu+muu+mub);
  mmuu[8]=-1.0;
  mmuu[9]=-1.0;
  mmuu[10]=1.0;
  mmuu[11]=1.0;
  mmub[8]=-1.0;
  mmub[9]=-1.0;
  mmub[10]=1.0;
  mmub[11]=1.0;
  mmu[8]=-1.0;
  mmu[9]=1.0;
  mmu[10]=-1.0;
  mmu[11]=1.0;
  
  iprop6.set_real(12,12,-md-mud-mur);
  iprop6.set_real(13,13,md-mud-mur);
  iprop6.set_real(14,14,-md+mud+mur);
  iprop6.set_real(15,15,md+mud+mur);
  mmud[12]=-1.0;
  mmud[13]=-1.0;
  mmud[14]=1.0;
  mmud[15]=1.0;
  mmur[12]=-1.0;
  mmur[13]=-1.0;
  mmur[14]=1.0;
  mmur[15]=1.0;
  mmd[12]=-1.0;
  mmd[13]=1.0;
  mmd[14]=-1.0;
  mmd[15]=1.0;
  
  iprop6.set_real(16,16,-md-mud-mug);
  iprop6.set_real(17,17,md-mud-mug);
  iprop6.set_real(18,18,-md+mud+mug);
  iprop6.set_real(19,19,md+mud+mug);
  mmud[16]=-1.0;
  mmud[17]=-1.0;
  mmud[18]=1.0;
  mmud[19]=1.0;
  mmug[16]=-1.0;
  mmug[17]=-1.0;
  mmug[18]=1.0;
  mmug[19]=1.0;
  mmd[16]=-1.0;
  mmd[17]=1.0;
  mmd[18]=-1.0;
  mmd[19]=1.0;
  
  iprop6.set_real(20,20,-md-mud-mub);
  iprop6.set_real(21,21,md-mud-mub);
  iprop6.set_real(22,22,-md+mud+mub);
  iprop6.set_real(23,23,md+mud+mub);
  mmud[20]=-1.0;
  mmud[21]=-1.0;
  mmud[22]=1.0;
  mmud[23]=1.0;
  mmub[20]=-1.0;
  mmub[21]=-1.0;
  mmub[22]=1.0;
  mmub[23]=1.0;
  mmd[20]=-1.0;
  mmd[21]=1.0;
  mmd[22]=-1.0;
  mmd[23]=1.0;
  
  iprop6.set_real(24,24,-ms-mus-mur);
  iprop6.set_real(25,25,ms-mus-mur);
  iprop6.set_real(26,26,-ms+mus+mur);
  iprop6.set_real(27,27,ms+mus+mur);
  mmus[24]=-1.0;
  mmus[25]=-1.0;
  mmus[26]=1.0;
  mmus[27]=1.0;
  mmur[24]=-1.0;
  mmur[25]=-1.0;
  mmur[26]=1.0;
  mmur[27]=1.0;
  mms[24]=-1.0;
  mms[25]=1.0;
  mms[26]=-1.0;
  mms[27]=1.0;
  
  iprop6.set_real(28,28,-ms-mus-mug);
  iprop6.set_real(29,29,ms-mus-mug);
  iprop6.set_real(30,30,-ms+mus+mug);
  iprop6.set_real(31,31,ms+mus+mug);
  mmus[28]=-1.0;
  mmus[29]=-1.0;
  mmus[30]=1.0;
  mmus[31]=1.0;
  mmug[28]=-1.0;
  mmug[29]=-1.0;
  mmug[30]=1.0;
  mmug[31]=1.0;
  mms[28]=-1.0;
  mms[29]=1.0;
  mms[30]=-1.0;
  mms[31]=1.0;
  
  iprop6.set_real(32,32,-ms-mus-mub);
  iprop6.set_real(33,33,ms-mus-mub);
  iprop6.set_real(34,34,-ms+mus+mub);
  iprop6.set_real(35,35,ms+mus+mub);
  mmus[32]=-1.0;
  mmus[33]=-1.0;
  mmus[34]=1.0;
  mmus[35]=1.0;
  mmub[32]=-1.0;
  mmub[33]=-1.0;
  mmub[34]=1.0;
  mmub[35]=1.0;
  mms[32]=-1.0;
  mms[33]=1.0;
  mms[34]=-1.0;
  mms[35]=1.0;
  
  // Set momentum
  
  for(int kk=0;kk<35;kk+=2) iprop6.set_real(kk+1,kk,-lmom);
  
  // Set gaps

  if (fixed_mass) {

    // Set gap terms

    iprop6.set_imag(28,23,du);
    iprop6.set_imag(30,21,du);
    iprop6.set_imag(33,18,du);
    iprop6.set_imag(35,16,du);

    iprop6.set_imag(29,22,-du);
    iprop6.set_imag(31,20,-du);
    iprop6.set_imag(32,19,-du);
    iprop6.set_imag(34,17,-du);
    
    iprop6.set_imag(24,11,dd);
    iprop6.set_imag(26,9,dd);
    iprop6.set_imag(33,2,dd);
    iprop6.set_imag(35,0,dd);

    iprop6.set_imag(25,10,-dd);
    iprop6.set_imag(27,8,-dd);
    iprop6.set_imag(32,3,-dd);
    iprop6.set_imag(34,1,-dd);
    
    iprop6.set_imag(12,7,ds);
    iprop6.set_imag(14,5,ds);
    iprop6.set_imag(17,2,ds);
    iprop6.set_imag(19,0,ds);

    iprop6.set_imag(13,6,-ds);
    iprop6.set_imag(15,4,-ds);
    iprop6.set_imag(16,3,-ds);
    iprop6.set_imag(18,1,-ds);

    // Set lower-left derivatives of gap terms

    dipdgapu.set_imag(28,23,1.0);
    dipdgapu.set_imag(30,21,1.0);
    dipdgapu.set_imag(33,18,1.0);
    dipdgapu.set_imag(35,16,1.0);

    dipdgapu.set_imag(29,22,-1.0);
    dipdgapu.set_imag(31,20,-1.0);
    dipdgapu.set_imag(32,19,-1.0);
    dipdgapu.set_imag(34,17,-1.0);
    
    dipdgapd.set_imag(24,11,1.0);
    dipdgapd.set_imag(26,9,1.0);
    dipdgapd.set_imag(33,2,1.0);
    dipdgapd.set_imag(35,0,1.0);

    dipdgapd.set_imag(25,10,-1.0);
    dipdgapd.set_imag(27,8,-1.0);
    dipdgapd.set_imag(32,3,-1.0);
    dipdgapd.set_imag(34,1,-1.0);
    
    dipdgaps.set_imag(12,7,1.0);
    dipdgaps.set_imag(14,5,1.0);
    dipdgaps.set_imag(17,2,1.0);
    dipdgaps.set_imag(19,0,1.0);

    dipdgaps.set_imag(13,6,-1.0);
    dipdgaps.set_imag(15,4,-1.0);
    dipdgaps.set_imag(16,3,-1.0);
    dipdgaps.set_imag(18,1,-1.0);

    // Set upper-right derivatives of gap terms

    dipdgapu.set_imag(23,28,-1.0);
    dipdgapu.set_imag(21,30,-1.0);
    dipdgapu.set_imag(18,33,-1.0);
    dipdgapu.set_imag(16,35,-1.0);
    
    dipdgapu.set_imag(22,29,1.0);
    dipdgapu.set_imag(20,31,1.0);
    dipdgapu.set_imag(19,32,1.0);
    dipdgapu.set_imag(17,34,1.0);
    
    dipdgapd.set_imag(11,24,-1.0);
    dipdgapd.set_imag(9,26,-1.0);
    dipdgapd.set_imag(2,33,-1.0);
    dipdgapd.set_imag(0,35,-1.0);
    
    dipdgapd.set_imag(10,25,1.0);
    dipdgapd.set_imag(8,27,1.0);
    dipdgapd.set_imag(3,32,1.0);
    dipdgapd.set_imag(1,34,1.0);
    
    dipdgaps.set_imag(7,12,-1.0);
    dipdgaps.set_imag(5,14,-1.0);
    dipdgaps.set_imag(2,17,-1.0);
    dipdgaps.set_imag(0,19,-1.0);
    
    dipdgaps.set_imag(6,13,1.0);
    dipdgaps.set_imag(4,15,1.0);
    dipdgaps.set_imag(3,16,1.0);
    dipdgaps.set_imag(1,18,1.0);
    
  } else {

    // set gap terms
    double betau=1.0+KD/GD*qqu/3.0;
    double betad=1.0+KD/GD*qqd/3.0;
    double betas=1.0+KD/GD*qqs/3.0;
    double gammau=du*KD/GD/3.0;
    double gammad=dd*KD/GD/3.0;
    double gammas=ds*KD/GD/3.0;
    
    iprop6.set_imag(28,23,du*betau);
    iprop6.set_imag(30,21,du*betau);
    iprop6.set_imag(33,18,du*betau);
    iprop6.set_imag(35,16,du*betau);

    iprop6.set_imag(29,22,-du*betau);
    iprop6.set_imag(31,20,-du*betau);
    iprop6.set_imag(32,19,-du*betau);
    iprop6.set_imag(34,17,-du*betau);
    
    iprop6.set_imag(24,11,dd*betad);
    iprop6.set_imag(26,9,dd*betad);
    iprop6.set_imag(33,2,dd*betad);
    iprop6.set_imag(35,0,dd*betad);

    iprop6.set_imag(25,10,-dd*betad);
    iprop6.set_imag(27,8,-dd*betad);
    iprop6.set_imag(32,3,-dd*betad);
    iprop6.set_imag(34,1,-dd*betad);
    
    iprop6.set_imag(12,7,ds*betas);
    iprop6.set_imag(14,5,ds*betas);
    iprop6.set_imag(17,2,ds*betas);
    iprop6.set_imag(19,0,ds*betas);
    
    iprop6.set_imag(13,6,-ds*betas);
    iprop6.set_imag(15,4,-ds*betas);
    iprop6.set_imag(16,3,-ds*betas);
    iprop6.set_imag(18,1,-ds*betas);

    // Set lower-left derivatives of gap terms

    dipdgapu.set_imag(28,23,betau);
    dipdgapu.set_imag(30,21,betau);
    dipdgapu.set_imag(33,18,betau);
    dipdgapu.set_imag(35,16,betau);

    dipdgapu.set_imag(29,22,-betau);
    dipdgapu.set_imag(31,20,-betau);
    dipdgapu.set_imag(32,19,-betau);
    dipdgapu.set_imag(34,17,-betau);
    
    dipdgapd.set_imag(24,11,betad);
    dipdgapd.set_imag(26,9,betad);
    dipdgapd.set_imag(33,2,betad);
    dipdgapd.set_imag(35,0,betad);

    dipdgapd.set_imag(25,10,-betad);
    dipdgapd.set_imag(27,8,-betad);
    dipdgapd.set_imag(32,3,-betad);
    dipdgapd.set_imag(34,1,-betad);
    
    dipdgaps.set_imag(12,7,betas);
    dipdgaps.set_imag(14,5,betas);
    dipdgaps.set_imag(17,2,betas);
    dipdgaps.set_imag(19,0,betas);
    
    dipdgaps.set_imag(13,6,-betas);
    dipdgaps.set_imag(15,4,-betas);
    dipdgaps.set_imag(16,3,-betas);
    dipdgaps.set_imag(18,1,-betas);

    dipdqqu.set_imag(28,23,gammau);
    dipdqqu.set_imag(30,21,gammau);
    dipdqqu.set_imag(33,18,gammau);
    dipdqqu.set_imag(35,16,gammau);

    dipdqqu.set_imag(29,22,-gammau);
    dipdqqu.set_imag(31,20,-gammau);
    dipdqqu.set_imag(32,19,-gammau);
    dipdqqu.set_imag(34,17,-gammau);
    
    dipdqqd.set_imag(24,11,gammad);
    dipdqqd.set_imag(26,9,gammad);
    dipdqqd.set_imag(33,2,gammad);
    dipdqqd.set_imag(35,0,gammad);

    dipdqqd.set_imag(25,10,-gammad);
    dipdqqd.set_imag(27,8,-gammad);
    dipdqqd.set_imag(32,3,-gammad);
    dipdqqd.set_imag(34,1,-gammad);
    
    dipdqqs.set_imag(12,7,gammas);
    dipdqqs.set_imag(14,5,gammas);
    dipdqqs.set_imag(17,2,gammas);
    dipdqqs.set_imag(19,0,gammas);
    
    dipdqqs.set_imag(13,6,-gammas);
    dipdqqs.set_imag(15,4,-gammas);
    dipdqqs.set_imag(16,3,-gammas);
    dipdqqs.set_imag(18,1,-gammas);

    // Set upper-right derivatives of gap terms

    dipdgapu.set_imag(23,28,-betau);
    dipdgapu.set_imag(21,30,-betau);
    dipdgapu.set_imag(18,33,-betau);
    dipdgapu.set_imag(16,35,-betau);
    
    dipdgapu.set_imag(22,29,betau);
    dipdgapu.set_imag(20,31,betau);
    dipdgapu.set_imag(19,32,betau);
    dipdgapu.set_imag(17,34,betau);
    
    dipdgapd.set_imag(11,24,-betad);
    dipdgapd.set_imag(9,26,-betad);
    dipdgapd.set_imag(2,33,-betad);
    dipdgapd.set_imag(0,35,-betad);
    
    dipdgapd.set_imag(10,25,betad);
    dipdgapd.set_imag(8,27,betad);
    dipdgapd.set_imag(3,32,betad);
    dipdgapd.set_imag(1,34,betad);
    
    dipdgaps.set_imag(7,12,-betas);
    dipdgaps.set_imag(5,14,-betas);
    dipdgaps.set_imag(2,17,-betas);
    dipdgaps.set_imag(0,19,-betas);
    
    dipdgaps.set_imag(6,13,betas);
    dipdgaps.set_imag(4,15,betas);
    dipdgaps.set_imag(3,16,betas);
    dipdgaps.set_imag(1,18,betas);

    dipdqqu.set_imag(23,28,-gammau);
    dipdqqu.set_imag(21,30,-gammau);
    dipdqqu.set_imag(18,33,-gammau);
    dipdqqu.set_imag(16,35,-gammau);
    
    dipdqqu.set_imag(22,29,gammau);
    dipdqqu.set_imag(20,31,gammau);
    dipdqqu.set_imag(19,32,gammau);
    dipdqqu.set_imag(17,34,gammau);
    
    dipdqqd.set_imag(11,24,-gammad);
    dipdqqd.set_imag(9,26,-gammad);
    dipdqqd.set_imag(2,33,-gammad);
    dipdqqd.set_imag(0,35,-gammad);
    
    dipdqqd.set_imag(10,25,gammad);
    dipdqqd.set_imag(8,27,gammad);
    dipdqqd.set_imag(3,32,gammad);
    dipdqqd.set_imag(1,34,gammad);
    
    dipdqqs.set_imag(7,12,-gammas);
    dipdqqs.set_imag(5,14,-gammas);
    dipdqqs.set_imag(2,17,-gammas);
    dipdqqs.set_imag(0,19,-gammas);
    
    dipdqqs.set_imag(6,13,gammas);
    dipdqqs.set_imag(4,15,gammas);
    dipdqqs.set_imag(3,16,gammas);
    dipdqqs.set_imag(1,18,gammas);
  }
  
  double acons=KD/4.0/GD/GD;
  double ua=du*acons;
  double da=dd*acons;
  double sa=ds*acons;
  double uda=du*dd*acons;
  double usa=du*ds*acons;
  double dsa=dd*ds*acons;

  // Set off-diagonal mass-like terms

  iprop6.set_real(12,0,uda);
  iprop6.set_real(15,3,uda);
  iprop6.set_real(16,4,uda);
  iprop6.set_real(19,7,uda);
  iprop6.set_real(20,8,uda);
  iprop6.set_real(23,11,uda);
  
  iprop6.set_real(13,1,-uda);
  iprop6.set_real(14,2,-uda);
  iprop6.set_real(17,5,-uda);
  iprop6.set_real(18,6,-uda);
  iprop6.set_real(21,9,-uda);
  iprop6.set_real(22,10,-uda);
  
  iprop6.set_real(24,0,usa);
  iprop6.set_real(27,3,usa);
  iprop6.set_real(28,4,usa);
  iprop6.set_real(31,7,usa);
  iprop6.set_real(32,8,usa);
  iprop6.set_real(35,11,usa);
  
  iprop6.set_real(25,1,-usa);
  iprop6.set_real(26,2,-usa);
  iprop6.set_real(29,5,-usa);
  iprop6.set_real(30,6,-usa);
  iprop6.set_real(33,9,-usa);
  iprop6.set_real(34,10,-usa);
  
  iprop6.set_real(24,12,dsa);
  iprop6.set_real(27,15,dsa);
  iprop6.set_real(28,16,dsa);
  iprop6.set_real(31,19,dsa);
  iprop6.set_real(32,20,dsa);
  iprop6.set_real(35,23,dsa);
  
  iprop6.set_real(25,13,-dsa);
  iprop6.set_real(26,14,-dsa);
  iprop6.set_real(29,17,-dsa);
  iprop6.set_real(30,18,-dsa);
  iprop6.set_real(33,21,-dsa);
  iprop6.set_real(34,22,-dsa);

  // Set lower-left derivatives of off-diagonal mass-like terms

  dipdgapu.set_real(12,0,da);
  dipdgapu.set_real(15,3,da);
  dipdgapu.set_real(16,4,da);
  dipdgapu.set_real(19,7,da);
  dipdgapu.set_real(20,8,da);
  dipdgapu.set_real(23,11,da);
  
  dipdgapd.set_real(12,0,ua);
  dipdgapd.set_real(15,3,ua);
  dipdgapd.set_real(16,4,ua);
  dipdgapd.set_real(19,7,ua);
  dipdgapd.set_real(20,8,ua);
  dipdgapd.set_real(23,11,ua);
  
  dipdgapu.set_real(13,1,-da);
  dipdgapu.set_real(14,2,-da);
  dipdgapu.set_real(17,5,-da);
  dipdgapu.set_real(18,6,-da);
  dipdgapu.set_real(21,9,-da);
  dipdgapu.set_real(22,10,-da);
  
  dipdgapd.set_real(13,1,-ua);
  dipdgapd.set_real(14,2,-ua);
  dipdgapd.set_real(17,5,-ua);
  dipdgapd.set_real(18,6,-ua);
  dipdgapd.set_real(21,9,-ua);
  dipdgapd.set_real(22,10,-ua);
  
  dipdgapu.set_real(24,0,sa);
  dipdgapu.set_real(27,3,sa);
  dipdgapu.set_real(28,4,sa);
  dipdgapu.set_real(31,7,sa);
  dipdgapu.set_real(32,8,sa);
  dipdgapu.set_real(35,11,sa);
  
  dipdgaps.set_real(24,0,ua);
  dipdgaps.set_real(27,3,ua);
  dipdgaps.set_real(28,4,ua);
  dipdgaps.set_real(31,7,ua);
  dipdgaps.set_real(32,8,ua);
  dipdgaps.set_real(35,11,ua);
  
  dipdgapu.set_real(25,1,-sa);
  dipdgapu.set_real(26,2,-sa);
  dipdgapu.set_real(29,5,-sa);
  dipdgapu.set_real(30,6,-sa);
  dipdgapu.set_real(33,9,-sa);
  dipdgapu.set_real(34,10,-sa);
  
  dipdgaps.set_real(25,1,-ua);
  dipdgaps.set_real(26,2,-ua);
  dipdgaps.set_real(29,5,-ua);
  dipdgaps.set_real(30,6,-ua);
  dipdgaps.set_real(33,9,-ua);
  dipdgaps.set_real(34,10,-ua);
  
  dipdgapd.set_real(24,12,sa);
  dipdgapd.set_real(27,15,sa);
  dipdgapd.set_real(28,16,sa);
  dipdgapd.set_real(31,19,sa);
  dipdgapd.set_real(32,20,sa);
  dipdgapd.set_real(35,23,sa);
  
  dipdgaps.set_real(24,12,da);
  dipdgaps.set_real(27,15,da);
  dipdgaps.set_real(28,16,da);
  dipdgaps.set_real(31,19,da);
  dipdgaps.set_real(32,20,da);
  dipdgaps.set_real(35,23,da);
  
  dipdgapd.set_real(25,13,-sa);
  dipdgapd.set_real(26,14,-sa);
  dipdgapd.set_real(29,17,-sa);
  dipdgapd.set_real(30,18,-sa);
  dipdgapd.set_real(33,21,-sa);
  dipdgapd.set_real(34,22,-sa);
  
  dipdgaps.set_real(25,13,-da);
  dipdgaps.set_real(26,14,-da);
  dipdgaps.set_real(29,17,-da);
  dipdgaps.set_real(30,18,-da);
  dipdgaps.set_real(33,21,-da);
  dipdgaps.set_real(34,22,-da);
  
  // Set upper-right derivatives of off-diagonal mass-like terms

  dipdgapu.set_real(0,12,da);
  dipdgapu.set_real(3,15,da);
  dipdgapu.set_real(4,16,da);
  dipdgapu.set_real(7,19,da);
  dipdgapu.set_real(8,20,da);
  dipdgapu.set_real(11,23,da);
  
  dipdgapd.set_real(0,12,ua);
  dipdgapd.set_real(3,15,ua);
  dipdgapd.set_real(4,16,ua);
  dipdgapd.set_real(7,19,ua);
  dipdgapd.set_real(8,20,ua);
  dipdgapd.set_real(11,23,ua);
  
  dipdgapu.set_real(1,13,-da);
  dipdgapu.set_real(2,14,-da);
  dipdgapu.set_real(5,17,-da);
  dipdgapu.set_real(6,18,-da);
  dipdgapu.set_real(9,21,-da);
  dipdgapu.set_real(10,22,-da);
  
  dipdgapd.set_real(1,13,-ua);
  dipdgapd.set_real(2,14,-ua);
  dipdgapd.set_real(5,17,-ua);
  dipdgapd.set_real(6,18,-ua);
  dipdgapd.set_real(9,21,-ua);
  dipdgapd.set_real(10,22,-ua);
  
  dipdgapu.set_real(0,24,sa);
  dipdgapu.set_real(3,27,sa);
  dipdgapu.set_real(4,28,sa);
  dipdgapu.set_real(7,31,sa);
  dipdgapu.set_real(8,32,sa);
  dipdgapu.set_real(11,35,sa);
  
  dipdgaps.set_real(0,24,ua);
  dipdgaps.set_real(3,27,ua);
  dipdgaps.set_real(4,28,ua);
  dipdgaps.set_real(7,31,ua);
  dipdgaps.set_real(8,32,ua);
  dipdgaps.set_real(11,35,ua);
  
  dipdgapu.set_real(1,25,-sa);
  dipdgapu.set_real(2,26,-sa);
  dipdgapu.set_real(5,29,-sa);
  dipdgapu.set_real(6,30,-sa);
  dipdgapu.set_real(9,33,-sa);
  dipdgapu.set_real(10,34,-sa);
  
  dipdgaps.set_real(1,25,-ua);
  dipdgaps.set_real(2,26,-ua);
  dipdgaps.set_real(5,29,-ua);
  dipdgaps.set_real(6,30,-ua);
  dipdgaps.set_real(9,33,-ua);
  dipdgaps.set_real(10,34,-ua);
  
  dipdgapd.set_real(12,24,sa);
  dipdgapd.set_real(15,27,sa);
  dipdgapd.set_real(16,28,sa);
  dipdgapd.set_real(19,31,sa);
  dipdgapd.set_real(20,32,sa);
  dipdgapd.set_real(23,35,sa);
  
  dipdgaps.set_real(12,24,da);
  dipdgaps.set_real(15,27,da);
  dipdgaps.set_real(16,28,da);
  dipdgaps.set_real(19,31,da);
  dipdgaps.set_real(20,32,da);
  dipdgaps.set_real(23,35,da);
  
  dipdgapd.set_real(13,25,-sa);
  dipdgapd.set_real(14,26,-sa);
  dipdgapd.set_real(17,29,-sa);
  dipdgapd.set_real(18,30,-sa);
  dipdgapd.set_real(21,33,-sa);
  dipdgapd.set_real(22,34,-sa);
  
  dipdgaps.set_real(13,25,-da);
  dipdgaps.set_real(14,26,-da);
  dipdgaps.set_real(17,29,-da);
  dipdgaps.set_real(18,30,-da);
  dipdgaps.set_real(21,33,-da);
  dipdgaps.set_real(22,34,-da);
  
  gsl_eigen_hermv((gsl_matrix_complex *)(&iprop6),(gsl_vector *)(&eval6),
		  (gsl_matrix_complex *)(&eivec6),w6);
  
  double dmudqqu=-4.0*G;
  double dmudqqd=2.0*K*strange->qq;
  double dmudqqs=2.0*K*down->qq;
  double dmddqqu=2.0*K*strange->qq;
  double dmddqqd=-4.0*G;
  double dmddqqs=2.0*K*up->qq;
  double dmsdqqu=2.0*K*down->qq;
  double dmsdqqd=2.0*K*up->qq;
  double dmsdqqs=-4.0*G;
  double dmudgu=KD/2.0/GD/GD*up->del;
  double dmddgd=KD/2.0/GD/GD*down->del;
  double dmsdgs=KD/2.0/GD/GD*strange->del;
  
  for(k=0;k<36;k++) {
    egv[k]=eval6[k];
    
    // Select the eigenvector and compute its conjugate
    omatrix_cx_col one_eivec(eivec6,k);
    ovector_cx econj=one_eivec;
    econj.conjugate();
    
    // Compute v+ M v
    gsl_complex tmp;
    gsl_complex dedmu, dedmd, dedms;
    
    // Contribution from explicit gap terms
    tmp=dot(econj,dipdgapu*one_eivec);
    dedu[k]=tmp.dat[0];
    tmp=dot(econj,dipdgapd*one_eivec);
    dedd[k]=tmp.dat[0];
    tmp=dot(econj,dipdgaps*one_eivec);
    deds[k]=tmp.dat[0];
    
    // Contribution from explicit <qq> entries
    tmp=dot(econj,dipdqqu*one_eivec);
    dedqqu[k]=tmp.dat[0];
    tmp=dot(econj,dipdqqd*one_eivec);
    dedqqd[k]=tmp.dat[0];
    tmp=dot(econj,dipdqqs*one_eivec);
    dedqqs[k]=tmp.dat[0];
    
    // mass derivatives
    dedmu=dot(econj,pair_prod(mmu,one_eivec));
    dedmd=dot(econj,pair_prod(mmd,one_eivec));
    dedms=dot(econj,pair_prod(mms,one_eivec));

    // Contribution from mass terms to quark condensate derivatives
    dedqqu[k]+=dedmu.dat[0]*dmudqqu+
      dedmd.dat[0]*dmddqqu+dedms.dat[0]*dmsdqqu;
    dedqqd[k]+=dedmu.dat[0]*dmudqqd+
      dedmd.dat[0]*dmddqqd+dedms.dat[0]*dmsdqqd;
    dedqqs[k]+=dedmu.dat[0]*dmudqqs+
      dedmd.dat[0]*dmddqqs+dedms.dat[0]*dmsdqqs;

    // Contribution from mass terms to gap derivatives
    dedu[k]+=dedmu.dat[0]*dmudgu;
    dedd[k]+=dedmd.dat[0]*dmddgd;
    deds[k]+=dedms.dat[0]*dmsdgs;

    // Flavor chemical potential derivatives
    tmp=dot(econj,pair_prod(mmuu,one_eivec));
    dedmuu[k]=tmp.dat[0];
    tmp=dot(econj,pair_prod(mmud,one_eivec));
    dedmud[k]=tmp.dat[0];
    tmp=dot(econj,pair_prod(mmus,one_eivec));
    dedmus[k]=tmp.dat[0];
    
    // Color chemical potential derivatives
    tmp=dot(econj,pair_prod(mmur,one_eivec));
    double dedmur=tmp.dat[0];
    tmp=dot(econj,pair_prod(mmug,one_eivec));
    double dedmug=tmp.dat[0];
    tmp=dot(econj,pair_prod(mmub,one_eivec));
    double dedmub=tmp.dat[0];

    dedmu3[k]=dedmur-dedmug;
    dedmu8[k]=(dedmur+dedmug-2.0*dedmub)/sqrt(3.0);
  } 
  
  return 0;
}

int cfl6_eos::make_matrices(double lmom, double mu3, 
			     double mu8, double egv[36],
			     double dedmuu[36], double dedmud[36],
			     double dedmus[36], double dedqqu[36], 
			     double dedqqd[36], double dedqqs[36],
			     double dedu[36], double dedd[36],
			     double deds[36], double dedmu3[36],
			     double dedmu8[36]) {
  
  int k;
  const double mu=up->ms, md=down->ms, ms=strange->ms;
  const double muu=up->mu, mud=down->mu, mus=strange->mu;
  const double du=up->del, dd=down->del, ds=strange->del;
  const double qqu=up->qq, qqd=down->qq, qqs=strange->qq;

  double mur, mug, mub;
  mur=mu3+mu8/sqrt(3.0);
  mug=-mu3+mu8/sqrt(3.0);
  mub=-2.0*mu8/sqrt(3.0);
  
  for(k=0;k<36;k++) {
    egv[k]=0.0;
    dedmuu[k]=0.0;
    dedmud[k]=0.0;
    dedmus[k]=0.0;
    dedqqu[k]=0.0;
    dedqqd[k]=0.0;
    dedqqs[k]=0.0;
    dedu[k]=0.0;
    dedd[k]=0.0;
    deds[k]=0.0;
    dedmu3[k]=0.0;
    dedmu8[k]=0.0;
  }
  
  ovector mmuu(mat_size), mmud(mat_size), mmus(mat_size);
  ovector mmu(mat_size), mmd(mat_size), mms(mat_size);
  ovector mmur(mat_size), mmug(mat_size), mmub(mat_size);
  for(k=0;k<36;k++) {
    mmuu[k]=0.0;
    mmud[k]=0.0;
    mmus[k]=0.0;
    mmu[k]=0.0;
    mmd[k]=0.0;
    mms[k]=0.0;
    mmur[k]=0.0;
    mmug[k]=0.0;
    mmub[k]=0.0;
  }

  gsl_complex zero={{0.0,0.0}};
  iprop6.set_all(zero);
  dipdgapu.set_all(zero);
  dipdgapd.set_all(zero);
  dipdgaps.set_all(zero);
  dipdqqu.set_all(zero);
  dipdqqd.set_all(zero);
  dipdqqs.set_all(zero);
  
  /// Set diagonal elements

  iprop6.set_real(0,0,-mu-muu-mur);
  iprop6.set_real(1,1,mu-muu-mur);
  iprop6.set_real(2,2,-mu+muu+mur);
  iprop6.set_real(3,3,mu+muu+mur);
  mmuu[0]=-1.0;
  mmuu[1]=-1.0;
  mmuu[2]=1.0;
  mmuu[3]=1.0;
  mmur[0]=-1.0;
  mmur[1]=-1.0;
  mmur[2]=1.0;
  mmur[3]=1.0;
  mmu[0]=-1.0;
  mmu[1]=1.0;
  mmu[2]=-1.0;
  mmu[3]=1.0;

  iprop6.set_real(4,4,-mu-muu-mug);
  iprop6.set_real(5,5,mu-muu-mug);
  iprop6.set_real(6,6,-mu+muu+mug);
  iprop6.set_real(7,7,mu+muu+mug);
  mmuu[4]=-1.0;
  mmuu[5]=-1.0;
  mmuu[6]=1.0;
  mmuu[7]=1.0;
  mmug[4]=-1.0;
  mmug[5]=-1.0;
  mmug[6]=1.0;
  mmug[7]=1.0;
  mmu[4]=-1.0;
  mmu[5]=1.0;
  mmu[6]=-1.0;
  mmu[7]=1.0;
  
  iprop6.set_real(8,8,-mu-muu-mub);
  iprop6.set_real(9,9,mu-muu-mub);
  iprop6.set_real(10,10,-mu+muu+mub);
  iprop6.set_real(11,11,mu+muu+mub);
  mmuu[8]=-1.0;
  mmuu[9]=-1.0;
  mmuu[10]=1.0;
  mmuu[11]=1.0;
  mmub[8]=-1.0;
  mmub[9]=-1.0;
  mmub[10]=1.0;
  mmub[11]=1.0;
  mmu[8]=-1.0;
  mmu[9]=1.0;
  mmu[10]=-1.0;
  mmu[11]=1.0;
  
  iprop6.set_real(12,12,-md-mud-mur);
  iprop6.set_real(13,13,md-mud-mur);
  iprop6.set_real(14,14,-md+mud+mur);
  iprop6.set_real(15,15,md+mud+mur);
  mmud[12]=-1.0;
  mmud[13]=-1.0;
  mmud[14]=1.0;
  mmud[15]=1.0;
  mmur[12]=-1.0;
  mmur[13]=-1.0;
  mmur[14]=1.0;
  mmur[15]=1.0;
  mmd[12]=-1.0;
  mmd[13]=1.0;
  mmd[14]=-1.0;
  mmd[15]=1.0;
  
  iprop6.set_real(16,16,-md-mud-mug);
  iprop6.set_real(17,17,md-mud-mug);
  iprop6.set_real(18,18,-md+mud+mug);
  iprop6.set_real(19,19,md+mud+mug);
  mmud[16]=-1.0;
  mmud[17]=-1.0;
  mmud[18]=1.0;
  mmud[19]=1.0;
  mmug[16]=-1.0;
  mmug[17]=-1.0;
  mmug[18]=1.0;
  mmug[19]=1.0;
  mmd[16]=-1.0;
  mmd[17]=1.0;
  mmd[18]=-1.0;
  mmd[19]=1.0;
  
  iprop6.set_real(20,20,-md-mud-mub);
  iprop6.set_real(21,21,md-mud-mub);
  iprop6.set_real(22,22,-md+mud+mub);
  iprop6.set_real(23,23,md+mud+mub);
  mmud[20]=-1.0;
  mmud[21]=-1.0;
  mmud[22]=1.0;
  mmud[23]=1.0;
  mmub[20]=-1.0;
  mmub[21]=-1.0;
  mmub[22]=1.0;
  mmub[23]=1.0;
  mmd[20]=-1.0;
  mmd[21]=1.0;
  mmd[22]=-1.0;
  mmd[23]=1.0;
  
  iprop6.set_real(24,24,-ms-mus-mur);
  iprop6.set_real(25,25,ms-mus-mur);
  iprop6.set_real(26,26,-ms+mus+mur);
  iprop6.set_real(27,27,ms+mus+mur);
  mmus[24]=-1.0;
  mmus[25]=-1.0;
  mmus[26]=1.0;
  mmus[27]=1.0;
  mmur[24]=-1.0;
  mmur[25]=-1.0;
  mmur[26]=1.0;
  mmur[27]=1.0;
  mms[24]=-1.0;
  mms[25]=1.0;
  mms[26]=-1.0;
  mms[27]=1.0;
  
  iprop6.set_real(28,28,-ms-mus-mug);
  iprop6.set_real(29,29,ms-mus-mug);
  iprop6.set_real(30,30,-ms+mus+mug);
  iprop6.set_real(31,31,ms+mus+mug);
  mmus[28]=-1.0;
  mmus[29]=-1.0;
  mmus[30]=1.0;
  mmus[31]=1.0;
  mmug[28]=-1.0;
  mmug[29]=-1.0;
  mmug[30]=1.0;
  mmug[31]=1.0;
  mms[28]=-1.0;
  mms[29]=1.0;
  mms[30]=-1.0;
  mms[31]=1.0;
  
  iprop6.set_real(32,32,-ms-mus-mub);
  iprop6.set_real(33,33,ms-mus-mub);
  iprop6.set_real(34,34,-ms+mus+mub);
  iprop6.set_real(35,35,ms+mus+mub);
  mmus[32]=-1.0;
  mmus[33]=-1.0;
  mmus[34]=1.0;
  mmus[35]=1.0;
  mmub[32]=-1.0;
  mmub[33]=-1.0;
  mmub[34]=1.0;
  mmub[35]=1.0;
  mms[32]=-1.0;
  mms[33]=1.0;
  mms[34]=-1.0;
  mms[35]=1.0;
  
  // Set momentum
  
  for(int kk=0;kk<35;kk+=2) iprop6.set_real(kk+1,kk,-lmom);
  
  // Set gaps

  if (fixed_mass) {

    // Set gap terms

    iprop6.set_imag(28,23,du);
    iprop6.set_imag(30,21,du);
    iprop6.set_imag(33,18,du);
    iprop6.set_imag(35,16,du);

    iprop6.set_imag(29,22,-du);
    iprop6.set_imag(31,20,-du);
    iprop6.set_imag(32,19,-du);
    iprop6.set_imag(34,17,-du);
    
    iprop6.set_imag(24,11,dd);
    iprop6.set_imag(26,9,dd);
    iprop6.set_imag(33,2,dd);
    iprop6.set_imag(35,0,dd);

    iprop6.set_imag(25,10,-dd);
    iprop6.set_imag(27,8,-dd);
    iprop6.set_imag(32,3,-dd);
    iprop6.set_imag(34,1,-dd);
    
    iprop6.set_imag(12,7,ds);
    iprop6.set_imag(14,5,ds);
    iprop6.set_imag(17,2,ds);
    iprop6.set_imag(19,0,ds);

    iprop6.set_imag(13,6,-ds);
    iprop6.set_imag(15,4,-ds);
    iprop6.set_imag(16,3,-ds);
    iprop6.set_imag(18,1,-ds);

    // Set lower-left derivatives of gap terms

    dipdgapu.set_imag(28,23,1.0);
    dipdgapu.set_imag(30,21,1.0);
    dipdgapu.set_imag(33,18,1.0);
    dipdgapu.set_imag(35,16,1.0);

    dipdgapu.set_imag(29,22,-1.0);
    dipdgapu.set_imag(31,20,-1.0);
    dipdgapu.set_imag(32,19,-1.0);
    dipdgapu.set_imag(34,17,-1.0);
    
    dipdgapd.set_imag(24,11,1.0);
    dipdgapd.set_imag(26,9,1.0);
    dipdgapd.set_imag(33,2,1.0);
    dipdgapd.set_imag(35,0,1.0);

    dipdgapd.set_imag(25,10,-1.0);
    dipdgapd.set_imag(27,8,-1.0);
    dipdgapd.set_imag(32,3,-1.0);
    dipdgapd.set_imag(34,1,-1.0);
    
    dipdgaps.set_imag(12,7,1.0);
    dipdgaps.set_imag(14,5,1.0);
    dipdgaps.set_imag(17,2,1.0);
    dipdgaps.set_imag(19,0,1.0);

    dipdgaps.set_imag(13,6,-1.0);
    dipdgaps.set_imag(15,4,-1.0);
    dipdgaps.set_imag(16,3,-1.0);
    dipdgaps.set_imag(18,1,-1.0);

    // Set upper-right derivatives of gap terms

    dipdgapu.set_imag(23,28,-1.0);
    dipdgapu.set_imag(21,30,-1.0);
    dipdgapu.set_imag(18,33,-1.0);
    dipdgapu.set_imag(16,35,-1.0);
    
    dipdgapu.set_imag(22,29,1.0);
    dipdgapu.set_imag(20,31,1.0);
    dipdgapu.set_imag(19,32,1.0);
    dipdgapu.set_imag(17,34,1.0);
    
    dipdgapd.set_imag(11,24,-1.0);
    dipdgapd.set_imag(9,26,-1.0);
    dipdgapd.set_imag(2,33,-1.0);
    dipdgapd.set_imag(0,35,-1.0);
    
    dipdgapd.set_imag(10,25,1.0);
    dipdgapd.set_imag(8,27,1.0);
    dipdgapd.set_imag(3,32,1.0);
    dipdgapd.set_imag(1,34,1.0);
    
    dipdgaps.set_imag(7,12,-1.0);
    dipdgaps.set_imag(5,14,-1.0);
    dipdgaps.set_imag(2,17,-1.0);
    dipdgaps.set_imag(0,19,-1.0);
    
    dipdgaps.set_imag(6,13,1.0);
    dipdgaps.set_imag(4,15,1.0);
    dipdgaps.set_imag(3,16,1.0);
    dipdgaps.set_imag(1,18,1.0);
    
  } else {

    // set gap terms
    double betau=1.0+KD/GD*qqu/3.0;
    double betad=1.0+KD/GD*qqd/3.0;
    double betas=1.0+KD/GD*qqs/3.0;
    double gammau=du*KD/GD/3.0;
    double gammad=dd*KD/GD/3.0;
    double gammas=ds*KD/GD/3.0;
    
    iprop6.set_imag(28,23,du*betau);
    iprop6.set_imag(30,21,du*betau);
    iprop6.set_imag(33,18,du*betau);
    iprop6.set_imag(35,16,du*betau);

    iprop6.set_imag(29,22,-du*betau);
    iprop6.set_imag(31,20,-du*betau);
    iprop6.set_imag(32,19,-du*betau);
    iprop6.set_imag(34,17,-du*betau);
    
    iprop6.set_imag(24,11,dd*betad);
    iprop6.set_imag(26,9,dd*betad);
    iprop6.set_imag(33,2,dd*betad);
    iprop6.set_imag(35,0,dd*betad);

    iprop6.set_imag(25,10,-dd*betad);
    iprop6.set_imag(27,8,-dd*betad);
    iprop6.set_imag(32,3,-dd*betad);
    iprop6.set_imag(34,1,-dd*betad);
    
    iprop6.set_imag(12,7,ds*betas);
    iprop6.set_imag(14,5,ds*betas);
    iprop6.set_imag(17,2,ds*betas);
    iprop6.set_imag(19,0,ds*betas);
    
    iprop6.set_imag(13,6,-ds*betas);
    iprop6.set_imag(15,4,-ds*betas);
    iprop6.set_imag(16,3,-ds*betas);
    iprop6.set_imag(18,1,-ds*betas);

    // Set lower-left derivatives of gap terms

    dipdgapu.set_imag(28,23,betau);
    dipdgapu.set_imag(30,21,betau);
    dipdgapu.set_imag(33,18,betau);
    dipdgapu.set_imag(35,16,betau);

    dipdgapu.set_imag(29,22,-betau);
    dipdgapu.set_imag(31,20,-betau);
    dipdgapu.set_imag(32,19,-betau);
    dipdgapu.set_imag(34,17,-betau);
    
    dipdgapd.set_imag(24,11,betad);
    dipdgapd.set_imag(26,9,betad);
    dipdgapd.set_imag(33,2,betad);
    dipdgapd.set_imag(35,0,betad);

    dipdgapd.set_imag(25,10,-betad);
    dipdgapd.set_imag(27,8,-betad);
    dipdgapd.set_imag(32,3,-betad);
    dipdgapd.set_imag(34,1,-betad);
    
    dipdgaps.set_imag(12,7,betas);
    dipdgaps.set_imag(14,5,betas);
    dipdgaps.set_imag(17,2,betas);
    dipdgaps.set_imag(19,0,betas);
    
    dipdgaps.set_imag(13,6,-betas);
    dipdgaps.set_imag(15,4,-betas);
    dipdgaps.set_imag(16,3,-betas);
    dipdgaps.set_imag(18,1,-betas);

    dipdqqu.set_imag(28,23,gammau);
    dipdqqu.set_imag(30,21,gammau);
    dipdqqu.set_imag(33,18,gammau);
    dipdqqu.set_imag(35,16,gammau);

    dipdqqu.set_imag(29,22,-gammau);
    dipdqqu.set_imag(31,20,-gammau);
    dipdqqu.set_imag(32,19,-gammau);
    dipdqqu.set_imag(34,17,-gammau);
    
    dipdqqd.set_imag(24,11,gammad);
    dipdqqd.set_imag(26,9,gammad);
    dipdqqd.set_imag(33,2,gammad);
    dipdqqd.set_imag(35,0,gammad);

    dipdqqd.set_imag(25,10,-gammad);
    dipdqqd.set_imag(27,8,-gammad);
    dipdqqd.set_imag(32,3,-gammad);
    dipdqqd.set_imag(34,1,-gammad);
    
    dipdqqs.set_imag(12,7,gammas);
    dipdqqs.set_imag(14,5,gammas);
    dipdqqs.set_imag(17,2,gammas);
    dipdqqs.set_imag(19,0,gammas);
    
    dipdqqs.set_imag(13,6,-gammas);
    dipdqqs.set_imag(15,4,-gammas);
    dipdqqs.set_imag(16,3,-gammas);
    dipdqqs.set_imag(18,1,-gammas);

    // Set upper-right derivatives of gap terms

    dipdgapu.set_imag(23,28,-betau);
    dipdgapu.set_imag(21,30,-betau);
    dipdgapu.set_imag(18,33,-betau);
    dipdgapu.set_imag(16,35,-betau);
    
    dipdgapu.set_imag(22,29,betau);
    dipdgapu.set_imag(20,31,betau);
    dipdgapu.set_imag(19,32,betau);
    dipdgapu.set_imag(17,34,betau);
    
    dipdgapd.set_imag(11,24,-betad);
    dipdgapd.set_imag(9,26,-betad);
    dipdgapd.set_imag(2,33,-betad);
    dipdgapd.set_imag(0,35,-betad);
    
    dipdgapd.set_imag(10,25,betad);
    dipdgapd.set_imag(8,27,betad);
    dipdgapd.set_imag(3,32,betad);
    dipdgapd.set_imag(1,34,betad);
    
    dipdgaps.set_imag(7,12,-betas);
    dipdgaps.set_imag(5,14,-betas);
    dipdgaps.set_imag(2,17,-betas);
    dipdgaps.set_imag(0,19,-betas);
    
    dipdgaps.set_imag(6,13,betas);
    dipdgaps.set_imag(4,15,betas);
    dipdgaps.set_imag(3,16,betas);
    dipdgaps.set_imag(1,18,betas);

    dipdqqu.set_imag(23,28,-gammau);
    dipdqqu.set_imag(21,30,-gammau);
    dipdqqu.set_imag(18,33,-gammau);
    dipdqqu.set_imag(16,35,-gammau);
    
    dipdqqu.set_imag(22,29,gammau);
    dipdqqu.set_imag(20,31,gammau);
    dipdqqu.set_imag(19,32,gammau);
    dipdqqu.set_imag(17,34,gammau);
    
    dipdqqd.set_imag(11,24,-gammad);
    dipdqqd.set_imag(9,26,-gammad);
    dipdqqd.set_imag(2,33,-gammad);
    dipdqqd.set_imag(0,35,-gammad);
    
    dipdqqd.set_imag(10,25,gammad);
    dipdqqd.set_imag(8,27,gammad);
    dipdqqd.set_imag(3,32,gammad);
    dipdqqd.set_imag(1,34,gammad);
    
    dipdqqs.set_imag(7,12,-gammas);
    dipdqqs.set_imag(5,14,-gammas);
    dipdqqs.set_imag(2,17,-gammas);
    dipdqqs.set_imag(0,19,-gammas);
    
    dipdqqs.set_imag(6,13,gammas);
    dipdqqs.set_imag(4,15,gammas);
    dipdqqs.set_imag(3,16,gammas);
    dipdqqs.set_imag(1,18,gammas);
  }
  
  double acons=KD/4.0/GD/GD;
  double ua=du*acons;
  double da=dd*acons;
  double sa=ds*acons;
  double uda=du*dd*acons;
  double usa=du*ds*acons;
  double dsa=dd*ds*acons;

  // Set off-diagonal mass-like terms

  iprop6.set_real(12,0,uda);
  iprop6.set_real(15,3,uda);
  iprop6.set_real(16,4,uda);
  iprop6.set_real(19,7,uda);
  iprop6.set_real(20,8,uda);
  iprop6.set_real(23,11,uda);
  
  iprop6.set_real(13,1,-uda);
  iprop6.set_real(14,2,-uda);
  iprop6.set_real(17,5,-uda);
  iprop6.set_real(18,6,-uda);
  iprop6.set_real(21,9,-uda);
  iprop6.set_real(22,10,-uda);
  
  iprop6.set_real(24,0,usa);
  iprop6.set_real(27,3,usa);
  iprop6.set_real(28,4,usa);
  iprop6.set_real(31,7,usa);
  iprop6.set_real(32,8,usa);
  iprop6.set_real(35,11,usa);
  
  iprop6.set_real(25,1,-usa);
  iprop6.set_real(26,2,-usa);
  iprop6.set_real(29,5,-usa);
  iprop6.set_real(30,6,-usa);
  iprop6.set_real(33,9,-usa);
  iprop6.set_real(34,10,-usa);
  
  iprop6.set_real(24,12,dsa);
  iprop6.set_real(27,15,dsa);
  iprop6.set_real(28,16,dsa);
  iprop6.set_real(31,19,dsa);
  iprop6.set_real(32,20,dsa);
  iprop6.set_real(35,23,dsa);
  
  iprop6.set_real(25,13,-dsa);
  iprop6.set_real(26,14,-dsa);
  iprop6.set_real(29,17,-dsa);
  iprop6.set_real(30,18,-dsa);
  iprop6.set_real(33,21,-dsa);
  iprop6.set_real(34,22,-dsa);

  // Set lower-left derivatives of off-diagonal mass-like terms

  dipdgapu.set_real(12,0,da);
  dipdgapu.set_real(15,3,da);
  dipdgapu.set_real(16,4,da);
  dipdgapu.set_real(19,7,da);
  dipdgapu.set_real(20,8,da);
  dipdgapu.set_real(23,11,da);
  
  dipdgapd.set_real(12,0,ua);
  dipdgapd.set_real(15,3,ua);
  dipdgapd.set_real(16,4,ua);
  dipdgapd.set_real(19,7,ua);
  dipdgapd.set_real(20,8,ua);
  dipdgapd.set_real(23,11,ua);
  
  dipdgapu.set_real(13,1,-da);
  dipdgapu.set_real(14,2,-da);
  dipdgapu.set_real(17,5,-da);
  dipdgapu.set_real(18,6,-da);
  dipdgapu.set_real(21,9,-da);
  dipdgapu.set_real(22,10,-da);
  
  dipdgapd.set_real(13,1,-ua);
  dipdgapd.set_real(14,2,-ua);
  dipdgapd.set_real(17,5,-ua);
  dipdgapd.set_real(18,6,-ua);
  dipdgapd.set_real(21,9,-ua);
  dipdgapd.set_real(22,10,-ua);
  
  dipdgapu.set_real(24,0,sa);
  dipdgapu.set_real(27,3,sa);
  dipdgapu.set_real(28,4,sa);
  dipdgapu.set_real(31,7,sa);
  dipdgapu.set_real(32,8,sa);
  dipdgapu.set_real(35,11,sa);
  
  dipdgaps.set_real(24,0,ua);
  dipdgaps.set_real(27,3,ua);
  dipdgaps.set_real(28,4,ua);
  dipdgaps.set_real(31,7,ua);
  dipdgaps.set_real(32,8,ua);
  dipdgaps.set_real(35,11,ua);
  
  dipdgapu.set_real(25,1,-sa);
  dipdgapu.set_real(26,2,-sa);
  dipdgapu.set_real(29,5,-sa);
  dipdgapu.set_real(30,6,-sa);
  dipdgapu.set_real(33,9,-sa);
  dipdgapu.set_real(34,10,-sa);
  
  dipdgaps.set_real(25,1,-ua);
  dipdgaps.set_real(26,2,-ua);
  dipdgaps.set_real(29,5,-ua);
  dipdgaps.set_real(30,6,-ua);
  dipdgaps.set_real(33,9,-ua);
  dipdgaps.set_real(34,10,-ua);
  
  dipdgapd.set_real(24,12,sa);
  dipdgapd.set_real(27,15,sa);
  dipdgapd.set_real(28,16,sa);
  dipdgapd.set_real(31,19,sa);
  dipdgapd.set_real(32,20,sa);
  dipdgapd.set_real(35,23,sa);
  
  dipdgaps.set_real(24,12,da);
  dipdgaps.set_real(27,15,da);
  dipdgaps.set_real(28,16,da);
  dipdgaps.set_real(31,19,da);
  dipdgaps.set_real(32,20,da);
  dipdgaps.set_real(35,23,da);
  
  dipdgapd.set_real(25,13,-sa);
  dipdgapd.set_real(26,14,-sa);
  dipdgapd.set_real(29,17,-sa);
  dipdgapd.set_real(30,18,-sa);
  dipdgapd.set_real(33,21,-sa);
  dipdgapd.set_real(34,22,-sa);
  
  dipdgaps.set_real(25,13,-da);
  dipdgaps.set_real(26,14,-da);
  dipdgaps.set_real(29,17,-da);
  dipdgaps.set_real(30,18,-da);
  dipdgaps.set_real(33,21,-da);
  dipdgaps.set_real(34,22,-da);
  
  // Set upper-right derivatives of off-diagonal mass-like terms

  dipdgapu.set_real(0,12,da);
  dipdgapu.set_real(3,15,da);
  dipdgapu.set_real(4,16,da);
  dipdgapu.set_real(7,19,da);
  dipdgapu.set_real(8,20,da);
  dipdgapu.set_real(11,23,da);
  
  dipdgapd.set_real(0,12,ua);
  dipdgapd.set_real(3,15,ua);
  dipdgapd.set_real(4,16,ua);
  dipdgapd.set_real(7,19,ua);
  dipdgapd.set_real(8,20,ua);
  dipdgapd.set_real(11,23,ua);
  
  dipdgapu.set_real(1,13,-da);
  dipdgapu.set_real(2,14,-da);
  dipdgapu.set_real(5,17,-da);
  dipdgapu.set_real(6,18,-da);
  dipdgapu.set_real(9,21,-da);
  dipdgapu.set_real(10,22,-da);
  
  dipdgapd.set_real(1,13,-ua);
  dipdgapd.set_real(2,14,-ua);
  dipdgapd.set_real(5,17,-ua);
  dipdgapd.set_real(6,18,-ua);
  dipdgapd.set_real(9,21,-ua);
  dipdgapd.set_real(10,22,-ua);
  
  dipdgapu.set_real(0,24,sa);
  dipdgapu.set_real(3,27,sa);
  dipdgapu.set_real(4,28,sa);
  dipdgapu.set_real(7,31,sa);
  dipdgapu.set_real(8,32,sa);
  dipdgapu.set_real(11,35,sa);
  
  dipdgaps.set_real(0,24,ua);
  dipdgaps.set_real(3,27,ua);
  dipdgaps.set_real(4,28,ua);
  dipdgaps.set_real(7,31,ua);
  dipdgaps.set_real(8,32,ua);
  dipdgaps.set_real(11,35,ua);
  
  dipdgapu.set_real(1,25,-sa);
  dipdgapu.set_real(2,26,-sa);
  dipdgapu.set_real(5,29,-sa);
  dipdgapu.set_real(6,30,-sa);
  dipdgapu.set_real(9,33,-sa);
  dipdgapu.set_real(10,34,-sa);
  
  dipdgaps.set_real(1,25,-ua);
  dipdgaps.set_real(2,26,-ua);
  dipdgaps.set_real(5,29,-ua);
  dipdgaps.set_real(6,30,-ua);
  dipdgaps.set_real(9,33,-ua);
  dipdgaps.set_real(10,34,-ua);
  
  dipdgapd.set_real(12,24,sa);
  dipdgapd.set_real(15,27,sa);
  dipdgapd.set_real(16,28,sa);
  dipdgapd.set_real(19,31,sa);
  dipdgapd.set_real(20,32,sa);
  dipdgapd.set_real(23,35,sa);
  
  dipdgaps.set_real(12,24,da);
  dipdgaps.set_real(15,27,da);
  dipdgaps.set_real(16,28,da);
  dipdgaps.set_real(19,31,da);
  dipdgaps.set_real(20,32,da);
  dipdgaps.set_real(23,35,da);
  
  dipdgapd.set_real(13,25,-sa);
  dipdgapd.set_real(14,26,-sa);
  dipdgapd.set_real(17,29,-sa);
  dipdgapd.set_real(18,30,-sa);
  dipdgapd.set_real(21,33,-sa);
  dipdgapd.set_real(22,34,-sa);
  
  dipdgaps.set_real(13,25,-da);
  dipdgaps.set_real(14,26,-da);
  dipdgaps.set_real(17,29,-da);
  dipdgaps.set_real(18,30,-da);
  dipdgaps.set_real(21,33,-da);
  dipdgaps.set_real(22,34,-da);
  
  return 0;
}
