/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/cfl6_eos_old.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

cfl6_eos_old::cfl6_eos_old() {
  KD=0.0;
  m6.allocate(size,size);
  eval6.allocate(size);
  w6=gsl_eigen_herm_alloc(size);
  
  GSL_REAL(zero)=0.0;
  GSL_IMAG(zero)=0.0;
  
  kdlimit=1.0e-6;
}

cfl6_eos_old::~cfl6_eos_old() {
  m6.free();
  eval6.free();
  gsl_eigen_herm_free(w6);

}

int cfl6_eos_old::calc_eq_temp_p(quark &u, quark &d, quark &s,
			     double &qq1, double &qq2, double &qq3, 
			     double &gap1, double &gap2, double &gap3, 
			     double mu3, double mu8, 
			     double &n3, double &n8, 
			     thermo &qb, const double ltemper) {
  
  if (fabs(KD)<kdlimit) {
    return cfl_njl_eos_old::calc_eq_temp_p(u,d,s,qq1,qq2,qq3,gap1,gap2,gap3,
				    mu3,mu8,n3,n8,qb,ltemper);
  }
  
  up=&u;
  down=&d;
  strange=&s;

  if (fromqq==false) {
    set_err("cfl6_eos_old does not work with fromqq=false.",1);
    return 1;
  }

  if (fixed_mass) {
    up->ms=up->m+KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m+KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m+KD/4.0/GD/GD*strange->del*strange->del;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq+
      KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq+
      KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq+
      KD/4.0/GD/GD*strange->del*strange->del;
  }
  if (up->ms<up->m) up->ms=up->m;
  if (down->ms<down->m) down->ms=down->m;
  if (strange->ms<strange->m) strange->ms=strange->m;
  
  // static variables for communication with tpot
  temper=ltemper;
  smu3=mu3;
  smu8=mu8;

  // In the case that the gaps are too small, go ahead
  // and force them to zero by solving the old gapless equations
  if ((up->del)<gap_limit && (down->del)<gap_limit && 
      (strange->del)<gap_limit) {
    int ret;
    ret=nambujl_eos::calc_eq_temp_p(*up,*down,*strange,
				    qq1,qq2,qq3,qb,temper);
    gap1=up->del;
    gap2=down->del;
    gap3=strange->del;
    return ret;
  }

  // Quark densities
  up->n=-derivfn(&up->mu);
  down->n=-derivfn(&down->mu);
  strange->n=-derivfn(&strange->mu);

  // We add 1 here to cancel the extra factors added to tpot.
  // This seems to help the numerical derivative when n3 and
  // n8 are close to zero.
  n3=-derivfn(&smu3)+1.0;
  n8=-derivfn(&smu8)+1.0;

  if (fixed_mass) {
    qq1=up->qq;
    qq2=down->qq;
    qq3=strange->qq;
  } else {
    qq1=derivfn(&up->qq);
    qq2=derivfn(&down->qq);
    qq3=derivfn(&strange->qq);
  }

  // Gaps
  if (fabs(up->del)>gap_limit) {
    gap1=derivfn(&up->del);
  } else {
    gap1=up->del;
  }
  if (fabs(down->del)>gap_limit) gap2=derivfn(&down->del);
  else gap2=down->del;
  if (fabs(strange->del)>gap_limit) gap3=derivfn(&strange->del);
  else gap3=strange->del;

  // Calculate pressure and entropy after the 
  // calculation of the quark condensates, which 
  // are needed for the final evaluation of the
  // pressure (9/30/04 - I think this comment is moot,
  // since fromqq is now always true)

  double dtemp;
  var=&dtemp;
  void *vpx=0;
  qb.pr=-tpot(temper,vpx);

  funct_mfptr_noerr<cfl6_eos_old,void *> f(this,&cfl6_eos_old::entropy_integrand);
  qb.en=it->integ(f,0.0,L,vpx);

  up->ed=-up->pr+up->n*up->mu;
  down->ed=-down->pr+down->n*down->mu;
  strange->ed=-strange->pr+strange->n*strange->mu;
  qb.ed=qb.en*temper-qb.pr+up->mu*up->n+down->mu*down->n+
    strange->mu*strange->n;

  return 0;

}

double cfl6_eos_old::tpot(double tvar, void *&pa) {
  if (fabs(KD)<kdlimit) return cfl_njl_eos_old::tpot(tvar,pa);

  double tp;

  *var=tvar;

  if (fixed_mass) {
    up->ms=up->m+KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m+KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m+KD/4.0/GD/GD*strange->del*strange->del;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq+
      KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq+
      KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq+
      KD/4.0/GD/GD*strange->del*strange->del;
  }
  if (up->ms<up->m) up->ms=up->m;
  if (down->ms<down->m) down->ms=down->m;
  if (strange->ms<strange->m) strange->ms=strange->m;
  
  void *vpx=0;
  funct_mfptr_noerr<cfl6_eos_old,void *> f(this,&cfl6_eos_old::tpot_integrand);
  tp=it->integ(f,0.0,L,vpx);

  if (fixed_mass) {
    tp+=B0;
  } else {
    tp+=up->qq*up->qq*2.0*G;
    tp+=down->qq*down->qq*2.0*G;
    tp+=strange->qq*strange->qq*2.0*G;
    tp-=up->qq*down->qq*strange->qq*4.0*K-B0;
  }
  
  tp+=up->del*up->del/4.0/GD;
  tp+=down->del*down->del/4.0/GD;
  tp+=strange->del*strange->del/4.0/GD;
  if (fixed_mass==false) {
    tp+=KD/GD/GD/2.0*
      (up->del*up->del*up->qq+
       down->del*down->del*down->qq+
       strange->del*strange->del*strange->qq);
  }
     
  // Add a fudge factor so that derivatives are easier when n3 and n8
  // are close to zero.
  if (var==&smu8) tp+=smu8;
  if (var==&smu3) tp+=smu3;
  
  return tp;
}

double cfl6_eos_old::thd_potential(quark &u, quark &d, quark &s,
			       double mu3, double mu8, 
			       const double ltemper) {
  double tpotx;

  up=&u;
  down=&d;
  strange=&s;

  if (fromqq==false) {
    set_err("cfl6_eos_old does not work with fromqq=false.",1);
    return 1;
  }

  // static variables for communication with tpot
  temper=ltemper;
  smu3=mu3;
  smu8=mu8;

  if (fixed_mass) {
    up->ms=up->m;
    down->ms=down->m;
    strange->ms=strange->m;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq+
      KD/4.0/GD/GD*up->del*up->del;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq+
      KD/4.0/GD/GD*down->del*down->del;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq+
      KD/4.0/GD/GD*strange->del*strange->del;
  }
  if (up->ms<up->m) up->ms=up->m;
  if (down->ms<down->m) down->ms=down->m;
  if (strange->ms<strange->m) strange->ms=strange->m;
  
  void *vpx=0;
  funct_mfptr_noerr<cfl6_eos_old,void *> f(this,&cfl6_eos_old::tpot_integrand);
  tpotx=it->integ(f,0.0,L,vpx);

  if (fixed_mass) {
    tpotx+=B0;
  } else {
    tpotx+=up->qq*up->qq*2.0*G;
    tpotx+=down->qq*down->qq*2.0*G;
    tpotx+=strange->qq*strange->qq*2.0*G;
    tpotx-=up->qq*down->qq*strange->qq*4.0*K-B0;
  }

  tpotx+=up->del*up->del/4.0/GD;
  tpotx+=down->del*down->del/4.0/GD;
  tpotx+=strange->del*strange->del/4.0/GD;
  tpotx+=KD/GD/GD/2.0*
    (up->del*up->del*up->qq+
     down->del*down->del*down->qq+
     strange->del*strange->del*strange->qq);

  return tpotx;

}

int cfl6_eos_old::eigenvalues(double lmom, ovector_view &egv, double mu3, 
			  double mu8) {
  if (fabs(KD)<kdlimit) return cfl_njl_eos_old::eigenvalues(lmom,egv,mu3,mu8);

  ovector_cx qev(36);
  gsl_complex z;
  int k;
  double mur, mug, mub;
  quark *u=up, *d=down, *s=strange;
  const double mu=u->ms, md=d->ms, ms=s->ms;
  const double muu=u->mu, mud=d->mu, mus=s->mu;
  const double du=u->del, dd=d->del, ds=s->del;
  const double qqu=u->qq, qqd=d->qq, qqs=s->qq;

  mur=mu3+mu8/sqrt(3.0);
  mug=-mu3+mu8/sqrt(3.0);
  mub=-2.0*mu8/sqrt(3.0);
  
  m6.set_all(zero);
  
  void *vp=(void *)&m6;
  gsl_matrix_complex *gm6=(gsl_matrix_complex *)vp;
  //  gsl_matrix_complex *gm6=(gsl_matrix_complex *)(&m6);

  /// Set diagonal elements

  GSL_IMAG(z)=0.0;
  
  GSL_REAL(z)=-mu-muu-mur;
  gsl_matrix_complex_set(gm6,0,0,z);
  GSL_REAL(z)=mu-muu-mur;
  gsl_matrix_complex_set(gm6,1,1,z);
  GSL_REAL(z)=-mu+muu+mur;
  gsl_matrix_complex_set(gm6,2,2,z);
  GSL_REAL(z)=mu+muu+mur;
  gsl_matrix_complex_set(gm6,3,3,z);

  GSL_REAL(z)=-mu-muu-mug;
  gsl_matrix_complex_set(gm6,4,4,z);
  GSL_REAL(z)=mu-muu-mug;
  gsl_matrix_complex_set(gm6,5,5,z);
  GSL_REAL(z)=-mu+muu+mug;
  gsl_matrix_complex_set(gm6,6,6,z);
  GSL_REAL(z)=mu+muu+mug;
  gsl_matrix_complex_set(gm6,7,7,z);
  
  GSL_REAL(z)=-mu-muu-mub;
  gsl_matrix_complex_set(gm6,8,8,z);
  GSL_REAL(z)=mu-muu-mub;
  gsl_matrix_complex_set(gm6,9,9,z);
  GSL_REAL(z)=-mu+muu+mub;
  gsl_matrix_complex_set(gm6,10,10,z);
  GSL_REAL(z)=mu+muu+mub;
  gsl_matrix_complex_set(gm6,11,11,z);
  
  GSL_REAL(z)=-md-mud-mur;
  gsl_matrix_complex_set(gm6,12,12,z);
  GSL_REAL(z)=md-mud-mur;
  gsl_matrix_complex_set(gm6,13,13,z);
  GSL_REAL(z)=-md+mud+mur;
  gsl_matrix_complex_set(gm6,14,14,z);
  GSL_REAL(z)=md+mud+mur;
  gsl_matrix_complex_set(gm6,15,15,z);
  
  GSL_REAL(z)=-md-mud-mug;
  gsl_matrix_complex_set(gm6,16,16,z);
  GSL_REAL(z)=md-mud-mug;
  gsl_matrix_complex_set(gm6,17,17,z);
  GSL_REAL(z)=-md+mud+mug;
  gsl_matrix_complex_set(gm6,18,18,z);
  GSL_REAL(z)=md+mud+mug;
  gsl_matrix_complex_set(gm6,19,19,z);
  
  GSL_REAL(z)=-md-mud-mub;
  gsl_matrix_complex_set(gm6,20,20,z);
  GSL_REAL(z)=md-mud-mub;
  gsl_matrix_complex_set(gm6,21,21,z);
  GSL_REAL(z)=-md+mud+mub;
  gsl_matrix_complex_set(gm6,22,22,z);
  GSL_REAL(z)=md+mud+mub;
  gsl_matrix_complex_set(gm6,23,23,z);
  
  GSL_REAL(z)=-ms-mus-mur;
  gsl_matrix_complex_set(gm6,24,24,z);
  GSL_REAL(z)=ms-mus-mur;
  gsl_matrix_complex_set(gm6,25,25,z);
  GSL_REAL(z)=-ms+mus+mur;
  gsl_matrix_complex_set(gm6,26,26,z);
  GSL_REAL(z)=ms+mus+mur;
  gsl_matrix_complex_set(gm6,27,27,z);
  
  GSL_REAL(z)=-ms-mus-mug;
  gsl_matrix_complex_set(gm6,28,28,z);
  GSL_REAL(z)=ms-mus-mug;
  gsl_matrix_complex_set(gm6,29,29,z);
  GSL_REAL(z)=-ms+mus+mug;
  gsl_matrix_complex_set(gm6,30,30,z);
  GSL_REAL(z)=ms+mus+mug;
  gsl_matrix_complex_set(gm6,31,31,z);
  
  GSL_REAL(z)=-ms-mus-mub;
  gsl_matrix_complex_set(gm6,32,32,z);
  GSL_REAL(z)=ms-mus-mub;
  gsl_matrix_complex_set(gm6,33,33,z);
  GSL_REAL(z)=-ms+mus+mub;
  gsl_matrix_complex_set(gm6,34,34,z);
  GSL_REAL(z)=ms+mus+mub;
  gsl_matrix_complex_set(gm6,35,35,z);
  
  // Set momentum

  GSL_REAL(z)=-lmom;
  for(int kk=0;kk<35;kk+=2) gsl_matrix_complex_set(gm6,kk+1,kk,z);

  // Set gaps

  GSL_REAL(z)=0.0;

  if (fixed_mass) {

    GSL_IMAG(z)=du;
    gsl_matrix_complex_set(gm6,28,23,z);
    gsl_matrix_complex_set(gm6,30,21,z);
    gsl_matrix_complex_set(gm6,33,18,z);
    gsl_matrix_complex_set(gm6,35,16,z);
    GSL_IMAG(z)=-du;
    gsl_matrix_complex_set(gm6,29,22,z);
    gsl_matrix_complex_set(gm6,31,20,z);
    gsl_matrix_complex_set(gm6,32,19,z);
    gsl_matrix_complex_set(gm6,34,17,z);
    
    GSL_IMAG(z)=dd;
    gsl_matrix_complex_set(gm6,24,11,z);
    gsl_matrix_complex_set(gm6,26,9,z);
    gsl_matrix_complex_set(gm6,33,2,z);
    gsl_matrix_complex_set(gm6,35,0,z);
    GSL_IMAG(z)=-dd;
    gsl_matrix_complex_set(gm6,25,10,z);
    gsl_matrix_complex_set(gm6,27,8,z);
    gsl_matrix_complex_set(gm6,32,3,z);
    gsl_matrix_complex_set(gm6,34,1,z);
    
    GSL_IMAG(z)=ds;
    gsl_matrix_complex_set(gm6,12,7,z);
    gsl_matrix_complex_set(gm6,14,5,z);
    gsl_matrix_complex_set(gm6,17,2,z);
    gsl_matrix_complex_set(gm6,19,0,z);
    GSL_IMAG(z)=-ds;
    gsl_matrix_complex_set(gm6,13,6,z);
    gsl_matrix_complex_set(gm6,15,4,z);
    gsl_matrix_complex_set(gm6,16,3,z);
    gsl_matrix_complex_set(gm6,18,1,z);

  } else {

    GSL_IMAG(z)=du*(1.0+KD/GD*qqu/3.0);
    gsl_matrix_complex_set(gm6,28,23,z);
    gsl_matrix_complex_set(gm6,30,21,z);
    gsl_matrix_complex_set(gm6,33,18,z);
    gsl_matrix_complex_set(gm6,35,16,z);
    GSL_IMAG(z)=-du*(1.0+KD/GD*qqu/3.0);
    gsl_matrix_complex_set(gm6,29,22,z);
    gsl_matrix_complex_set(gm6,31,20,z);
    gsl_matrix_complex_set(gm6,32,19,z);
    gsl_matrix_complex_set(gm6,34,17,z);
    
    GSL_IMAG(z)=dd*(1.0+KD/GD*qqd/3.0);
    gsl_matrix_complex_set(gm6,24,11,z);
    gsl_matrix_complex_set(gm6,26,9,z);
    gsl_matrix_complex_set(gm6,33,2,z);
    gsl_matrix_complex_set(gm6,35,0,z);
    GSL_IMAG(z)=-dd*(1.0+KD/GD*qqd/3.0);
    gsl_matrix_complex_set(gm6,25,10,z);
    gsl_matrix_complex_set(gm6,27,8,z);
    gsl_matrix_complex_set(gm6,32,3,z);
    gsl_matrix_complex_set(gm6,34,1,z);
    
    GSL_IMAG(z)=ds*(1.0+KD/GD*qqs/3.0);
    gsl_matrix_complex_set(gm6,12,7,z);
    gsl_matrix_complex_set(gm6,14,5,z);
    gsl_matrix_complex_set(gm6,17,2,z);
    gsl_matrix_complex_set(gm6,19,0,z);
    GSL_IMAG(z)=-ds*(1.0+KD/GD*qqs/3.0);
    gsl_matrix_complex_set(gm6,13,6,z);
    gsl_matrix_complex_set(gm6,15,4,z);
    gsl_matrix_complex_set(gm6,16,3,z);
    gsl_matrix_complex_set(gm6,18,1,z);
  }
  
  // Set off-diagonal mass-like terms

  GSL_IMAG(z)=0.0;
  
  double acons=KD/4.0/GD/GD;
  
  GSL_REAL(z)=du*dd*acons;
  gsl_matrix_complex_set(gm6,12,0,z);
  gsl_matrix_complex_set(gm6,15,3,z);
  gsl_matrix_complex_set(gm6,16,4,z);
  gsl_matrix_complex_set(gm6,19,7,z);
  gsl_matrix_complex_set(gm6,20,8,z);
  gsl_matrix_complex_set(gm6,23,11,z);
  
  GSL_REAL(z)=-du*dd*acons;
  gsl_matrix_complex_set(gm6,13,1,z);
  gsl_matrix_complex_set(gm6,14,2,z);
  gsl_matrix_complex_set(gm6,17,5,z);
  gsl_matrix_complex_set(gm6,18,6,z);
  gsl_matrix_complex_set(gm6,21,9,z);
  gsl_matrix_complex_set(gm6,22,10,z);
  
  GSL_REAL(z)=du*ds*acons;
  gsl_matrix_complex_set(gm6,24,0,z);
  gsl_matrix_complex_set(gm6,27,3,z);
  gsl_matrix_complex_set(gm6,28,4,z);
  gsl_matrix_complex_set(gm6,31,7,z);
  gsl_matrix_complex_set(gm6,32,8,z);
  gsl_matrix_complex_set(gm6,35,11,z);
  
  GSL_REAL(z)=-du*ds*acons;
  gsl_matrix_complex_set(gm6,25,1,z);
  gsl_matrix_complex_set(gm6,26,2,z);
  gsl_matrix_complex_set(gm6,29,5,z);
  gsl_matrix_complex_set(gm6,30,6,z);
  gsl_matrix_complex_set(gm6,33,9,z);
  gsl_matrix_complex_set(gm6,34,10,z);
  
  GSL_REAL(z)=dd*ds*acons;
  gsl_matrix_complex_set(gm6,24,12,z);
  gsl_matrix_complex_set(gm6,27,15,z);
  gsl_matrix_complex_set(gm6,28,16,z);
  gsl_matrix_complex_set(gm6,31,19,z);
  gsl_matrix_complex_set(gm6,32,20,z);
  gsl_matrix_complex_set(gm6,35,23,z);
  
  GSL_REAL(z)=-dd*ds*acons;
  gsl_matrix_complex_set(gm6,25,13,z);
  gsl_matrix_complex_set(gm6,26,14,z);
  gsl_matrix_complex_set(gm6,29,17,z);
  gsl_matrix_complex_set(gm6,30,18,z);
  gsl_matrix_complex_set(gm6,33,21,z);
  gsl_matrix_complex_set(gm6,34,22,z);

  vp=(void *)&egv;
  gsl_eigen_herm(gm6,(gsl_vector *)(vp),w6);

  return 0;
}
