/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HAD_GIBBS_NSEOS_H
#define O2SCL_HAD_GIBBS_NSEOS_H

#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/hadronic_eos.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Create a Gibbs phase transition between two hadronic EOSs
      to create neutron star matter

      This computes neutron star matter assuming a Gibbs phase
      transition between two \ref hadronic_eos 's. When possible it
      generally uses hadronic_eos::calc_e() over
      hadronic_eos::calc_p().

      This class is not very sophisticated and the functions calc_e()
      and fixed_chi() will fail if the initial guess is not good
      enough.
  */
  class had_gibbs_nseos {

  public:

    had_gibbs_nseos();
    
    virtual ~had_gibbs_nseos();
    
    /** \brief Compute the properties of neutron star matter at
	
	fixed density
    */
    virtual int calc_e(double nb, fermion &n1, fermion &p1, thermo &th1,
		       fermion &n2, fermion &p2, thermo &th2, double &chi,
		       int type);
    
    /** \brief Compute the properties of neutron star matter in the
	
	mixed phase for fixed chi
    */
    virtual int fixed_chi(double chi, fermion &n1, fermion &p1, thermo &th1,
			  fermion &n2, fermion &p2, thermo &th2, double &nb,
			  int type);
    
    /// Set the Phase 1 hadronic equation of state
    int set_hadronic_eos1(hadronic_eos &he) {
      hep1=&he;
      return 0;
    }

    /// Set the Phase 2 hadronic equation of state
    int set_hadronic_eos2(hadronic_eos &he) {
      hep2=&he;
      return 0;
    }

    /// The default solver
    gsl_mroot_hybrids<void *,mm_funct<void *> > def_solver;

    /** 
	\brief Change the solver
    
	In order to work properly, the solver needs to gracefully
	recover from a function which returns a non-zero value. If a
	\ref gsl_mroot_hybrids object is used, then \ref
	gsl_mroot_hybrids::shrink_step should be set to \c true. This
	is done automatically for the default solver.
     */
    int set_solver(mroot<void *,mm_funct<void *> > &mr) {
      solverp=&mr;
      return 0;
    }

    /** \brief The guess for the phase of the next call to calc_e() 
	(default: \c phase1)
    */
    int phase_guess;

    /** \name Possible values for phase_guess 
     */
    //@{
    static const int phase1=1;
    static const int phase2=2;
    static const int mixed_phase=3;
    //@}

    /** \name Possible values for type
     */
    //@{
    static const int nuclear_matter=1;
    static const int neutron_matter=2;
    static const int nstar_matter=3;
    //@}

#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /** \brief Useful structure for parameters to phase1_eqs(), phase2_eqs(),
	and gibbs_eqs() [protected]
    */
    typedef struct solvepar_s {
      fermion *n1;
      fermion *p1;
      fermion *n2;
      fermion *p2;
      thermo *th1;
      thermo *th2;
      double nb;
    } solvepar;

    /** \brief Useful structure for parameters to 
	had_gibbs_nseos::fixed_chi_eqs() [protected]
     */
    typedef struct chipar_s {
      fermion *n1;
      fermion *p1;
      fermion *n2;
      fermion *p2;
      thermo *th1;
      thermo *th2;
      double chi;
    } chipar;
	
    /// Solve for Phase 1 matter
    int phase1_nstar(size_t nv, const ovector_view &x, ovector_view &y, 
		     void *&pa);
     
    /// Solve for Phase 2 matter
    int phase2_nstar(size_t nv, const ovector_view &x, ovector_view &y, 
		     void *&pa);

    /// Solve for mixed phase matter
    int gibbs_nstar(size_t nv, const ovector_view &x, ovector_view &y, 
		    void *&pa);

    /// Solve for mixed phase matter at fixed chi
    int fixed_chi_nstar(size_t nv, const ovector_view &x, ovector_view &y, 
			void *&pa);
    
    /// Solve for mixed phase matter
    int gibbs_nuc(size_t nv, const ovector_view &x, ovector_view &y, 
		  void *&pa);

    /// Solve for mixed phase matter at fixed chi
    int fixed_chi_nuc(size_t nv, const ovector_view &x, ovector_view &y, 
		      void *&pa);
    
    /// Solve for mixed phase matter
    int gibbs_neut(size_t nv, const ovector_view &x, ovector_view &y, 
		   void *&pa);

    /// Solve for mixed phase matter at fixed chi
    int fixed_chi_neut(size_t nv, const ovector_view &x, ovector_view &y, 
		       void *&pa);
    
    /// The hadronic EOS for Phase 1
    hadronic_eos *hep1;
    
    /// The hadronic EOS for Phase 2
    hadronic_eos *hep2;

    /// The electron
    fermion elec;

    /// The muon
    fermion mu;

    /// The solver
    mroot<void *,mm_funct<void *> > *solverp;

    /// \name Storage for the solution vectors
    //@{
    ovector gx;
    ovector px;
    //@}
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
