/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_COMPOSITE_INTE_H
#define O2SCL_COMPOSITE_INTE_H

#include <iostream>
#include <o2scl/multi_inte.h>
#include <o2scl/inte.h>
#include <o2scl/funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Naive multi-dimensional integration over a hypercube

      Naively combine several one-dimensional integration objects from
      class inte in order to perform a multi-dimensional integration
      over a region defined by constant limits. For more general regions
      of integration, use children of the class \ref gen_inte.

      The integration routines are specified in the function
      set_ptrs().

      The integration routines are called in order of their
      specification in the list <tt>inte **ip</tt>. For the example of
      a two-dimensional integration <tt>ip[0]</tt> is called first
      with limits <tt>a[0]</tt> and <tt>b[0]</tt> and performs the
      integral of the integral given by <tt>ip[1]</tt> of the function
      from <tt>a[1]</tt> to <tt>b[1]</tt>. In other words, the
      integral performed is:
      \f[ 
      \int_{x_0=a_0}^{x_0=b_0} \int_{x_1=a_1}^{x_1=b_1} ...
      \int_{x_{\mathrm{n}-1}=a_{\mathrm{n}-1}}^
      {x_{\mathrm{n}-1}=b_{\mathrm{n}-1}} 
      f(x_0,x_1,...,x_{\mathrm{n}})
      \f]
      
      No error estimate is performed. Error estimation for multiple
      dimension integrals is provided by the Monte Carlo integration
      classes (see \ref mcarlo_inte).

      \todo Convert to using std::vector<inte> for the 1-d integration 
      pointers?
  */
  template<class param_t, class func_t, class vec_t,
    class alloc_vec_t, class alloc_t> class composite_inte : 
  public multi_inte<param_t,func_t,vec_t> {
    
  public:
    
    /** 
	\brief Parameters to send to the 1-d integration functions

	This structure is not intended to be frequently used 
	directly by the class-user, but must be public so 
	that the 1-d integration objects can be specified.
     */
    typedef struct {
      /// The user-specified upper limits
      const vec_t *ax;
      /// The user-specified lower limits
      const vec_t *bx;
      /// The independent variable vector
      alloc_vec_t *cx;
      /// The user-specified function
      func_t *mf;
      /// The user-specified number of dimensions
      int ndim;
      /// The present recursion level
      int idim;
      /// The user-specified parameter
      param_t *vp;
    } od_parms;
      
    composite_inte() {
      ptrs_set=false;
      ndim=0;
      fmn=new funct_mfptr_noerr<composite_inte<param_t,func_t,
	vec_t,alloc_vec_t,alloc_t>,od_parms>
	(this,&composite_inte<param_t,func_t,vec_t,
	 alloc_vec_t,alloc_t>::odfunc);
    }
    
    virtual ~composite_inte() {
      delete fmn;
    }
    
    /** 
	\brief Designate the pointers to the integration routines
	
	This function allows duplicate objects in this list in order
	to allow the user to use only one object for more than one of
	the integrations.
	
	If more 1-d integration routines than necessary are given,
	the extras will be unused.
    */
    int set_ptrs(inte<od_parms,funct<od_parms> > **ip, int n) {
      if (n<1) {
	set_err_ret("Number of dimensions less than 1 in set_ptrs().",
		    gsl_einval);
      }
      iptrs=ip;
      ndim=n;
      ptrs_set=true;
      return 0;
    }
      
    /* \brief Integrate function \c func over the hypercube from 
	\f$ x_i=a_i \f$ to \f$ x_i=b_i \f$ for 
	\f$ 0<i< \f$ ndim-1  

	The function set_ptrs() must be used first to set the 
	one-dimensional routines to use.
    */
    virtual double minteg(func_t &func, size_t n, const vec_t &a, 
			  const vec_t &b, param_t &pa) 
    {
      if (ptrs_set==false || n>ndim) {
	set_err("Too few objects specified with set_ptrs() in minteg().",
		gsl_einval);
	return 0.0;
      }

      alloc_vec_t cx;
      ao.allocate(cx,n);

      od_parms op={&a,&b,&cx,&func,n,0,&pa};
	
      double res=iptrs[0]->integ(*fmn,a[0],b[0],op);
	
      ao.free(cx);

      return res;

    }
    
    /// Return string denoting type ("composite_inte")
    virtual const char *type() { return "composite_inte"; }

  protected:

#ifndef DOXYGEN_INTERNAL

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// The one-dimensional integration function
    double odfunc(double x, od_parms &od) {

      double res;
	
      (*od.cx)[od.idim]=x;
      if (od.idim==od.ndim-1) {
	(*od.mf)(od.ndim,(*od.cx),res,*od.vp);
      } else {
	od_parms op2={od.ax,od.bx,od.cx,od.mf,od.ndim,od.idim+1,od.vp};
	res=iptrs[od.idim]->integ(*fmn,(*od.ax)[od.idim+1],
				  (*od.bx)[od.idim+1],op2);
      }
      return res;
    }
    
    /// This function to send to the integrators
    funct_mfptr_noerr<composite_inte<param_t,func_t,vec_t,
      alloc_vec_t,alloc_t>,od_parms> *fmn;

    /// The number of dimensions
    size_t ndim;

    /// Pointers to the integration objects
    inte<od_parms,funct<od_parms> > **iptrs;

    /// True if the integration objects have been specified
    bool ptrs_set;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif

