/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_INTE_QAG_H
#define O2SCL_GSL_INTE_QAG_H

#include <o2scl/inte.h>
#include <o2scl/gsl_inte_qag_b.h>
#include <o2scl/funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Adaptive integration a function with finite limits of 
      integration (GSL)
      
      \todo Verbose output has been setup for this class, but this
      needs to be done for some of the other GSL-like integrators

      \todo Document workspace size here somehow

      \todo Document use of last_iter
  */
  template<class param_t, class func_t=funct<param_t> > class gsl_inte_qag : 
  public gsl_inte_kronrod<param_t,func_t> {
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// Select the number of integration points
    int lkey;

#endif

  public:

    /// Create an integrator with the specified key
    gsl_inte_qag(int key=1) {
      lkey=1;
      if (key>=1 && key<=6){
	lkey=key;
      }
    }

    virtual ~gsl_inte_qag() {}
      
    /** 
	\brief Set the number of integration points
 
	The possible values for \c key  are:
	- 1: GSL_INTEG_GAUSS15 (default)
	- 2: GSL_INTEG_GAUSS21
	- 3: GSL_INTEG_GAUSS31
	- 4: GSL_INTEG_GAUSS41
	- 5: GSL_INTEG_GAUSS51
	- 6: GSL_INTEG_GAUSS61
 
	If an integer other than 1-6 is given, the default
	(GSL_INTEG_GAUSS15) is assumed, and the error handler is called.
    */
    int set_key(int key) {
      if (key>=1 && key<=6){
	lkey=key;
	return 0;
      }
      lkey=1;
      set_err_ret("Invalid key in gsl_inte_qag::set_key().",gsl_einval);
    }
     
    /** \brief Return the key used (1-6)
     */
    int get_key() { return lkey; }
  
    /** \brief Integrate function \c func from \c a to \c b.
     */
    virtual double integ(func_t &func, double a, double b, param_t &pa) {
      double res, err;
      integ_err(func,a,b,pa,res,err);
      this->interror=err;
      return res;
    }
    
    /** \brief Integrate function \c func from \c a to \c b and place
	the result in \c res and the error in \c err
    */
    virtual int integ_err(func_t &func, double a, double b, 
			  param_t &pa, double &res, double &err2) {
      
      int status;
      gsl_integration_rule * integration_rule = gsl_integration_qk15;

      if (lkey < GSL_INTEG_GAUSS15) {
	lkey = GSL_INTEG_GAUSS15;
      } else if (lkey > GSL_INTEG_GAUSS61) {
	lkey = GSL_INTEG_GAUSS61;
      }

      switch (lkey) {
      case GSL_INTEG_GAUSS15:
	{
	  double fv1[8], fv2[8];
	  status=qag(func,8,o2scl_inte_qag_coeffs::qk15_xgk,
		     o2scl_inte_qag_coeffs::qk15_wg,
		     o2scl_inte_qag_coeffs::qk15_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      case GSL_INTEG_GAUSS21:
	{
	  double fv1[11], fv2[11];
	  status=qag(func,11,o2scl_inte_qag_coeffs::qk21_xgk,
		     o2scl_inte_qag_coeffs::qk21_wg,
		     o2scl_inte_qag_coeffs::qk21_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      case GSL_INTEG_GAUSS31:
	{
	  double fv1[16], fv2[16];
	  status=qag(func,16,o2scl_inte_qag_coeffs::qk31_xgk,
		     o2scl_inte_qag_coeffs::qk31_wg,
		     o2scl_inte_qag_coeffs::qk31_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      case GSL_INTEG_GAUSS41:
	{
	  double fv1[21], fv2[21];
	  status=qag(func,21,o2scl_inte_qag_coeffs::qk41_xgk,
		     o2scl_inte_qag_coeffs::qk41_wg,
		     o2scl_inte_qag_coeffs::qk41_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      case GSL_INTEG_GAUSS51:
	{
	  double fv1[26], fv2[26];
	  status=qag(func,26,o2scl_inte_qag_coeffs::qk51_xgk,
		     o2scl_inte_qag_coeffs::qk51_wg,
		     o2scl_inte_qag_coeffs::qk51_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      case GSL_INTEG_GAUSS61:
	{
	  double fv1[31], fv2[31];
	  status=qag(func,31,o2scl_inte_qag_coeffs::qk61_xgk,
		     o2scl_inte_qag_coeffs::qk61_wg,
		     o2scl_inte_qag_coeffs::qk61_wgk,
		     fv1,fv2,a,b,this->tolx,this->tolf,this->wkspace,
		     &res,&err2,pa);
	}
	break;
      default:
	std::string estr="Incorrect key in gsl_inte_qag::integ_err().";
	set_err_ret(estr.c_str(),gsl_esanity);
      }

      return status;
    
    }

#ifndef DOXYGEN_INTERNAL

  protected:

    /** 
	\brief Perform an adaptive integration given the coefficients,
	and returning \c result
    */
    int qag(func_t &func, const int qn, const double xgk[], 
	    const double wg[], const double wgk[], double fv1[], 
	    double fv2[], const double a, const double b, 
	    const double l_epsabs, const double l_epsrel, const size_t limit, 
	    double *result, double *abserr, param_t &pa)
    {
      double area, errsum;
      double result0, abserr0, resabs0, resasc0;
      double tolerance;
      size_t iteration = 0;
      int roundoff_type1 = 0, roundoff_type2 = 0, error_type = 0;
      
      double round_off;
      
      /* Initialize results */
      
      initialise (this->w, a, b);
      
      *result = 0;
      *abserr = 0;
      
      if (limit > this->w->limit) {

	this->last_iter=0;
	
	std::string estr="Iteration limit exceeds workspace ";
	estr+="in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_einval);
      }
      
      if (l_epsabs <= 0 && 
	  (l_epsrel < 50 * GSL_DBL_EPSILON || l_epsrel < 0.5e-28)) {
	
	this->last_iter=0;
	
	std::string estr="Tolerance cannot be achieved with given ";
	estr+="value of 'tolx' and 'tolf' in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_ebadtol);
      }
	
      /* perform the first integration */
	
      gsl_integration_qk_o2scl(func,qn,xgk,wg,wgk,fv1,fv2,a,b,
			     &result0,&abserr0,&resabs0,&resasc0,pa);
      
      set_initial_result(this->w,result0,abserr0);
      
      /* Test on accuracy */
      
      tolerance = GSL_MAX_DBL(l_epsabs, l_epsrel * fabs (result0));
      
      /* need IEEE rounding here to match original quadpack behavior */
	  
      round_off=gsl_coerce_double(50 * GSL_DBL_EPSILON * resabs0);
	  
      if (abserr0 <= round_off && abserr0 > tolerance) {

	*result = result0;
	*abserr = abserr0;

	// We start with 1 here, because an integration
	// was already performed above
	this->last_iter=1;

	std::string estr="Cannot reach tolerance because of roundoff error ";
	estr+="on first attempt in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_eround);

      } else if ((abserr0 <= tolerance && 
		  abserr0 != resasc0) || abserr0 == 0.0) {
	*result = result0;
	*abserr = abserr0;
	  
	// We start with 1 here, because an integration
	// was already performed above
	this->last_iter=1;

	return gsl_success;

      } else if (limit == 1) {

	*result = result0;
	*abserr = abserr0;
	
	// We start with 1 here, because an integration
	// was already performed above
	this->last_iter=1;

	std::string estr="A maximum of 1 iteration was insufficient ";
	estr+="in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_emaxiter);
      }
      
      area = result0;
      errsum = abserr0;
      
      iteration = 1;
      do {
	double a1, b1, a2, b2;
	double a_i, b_i, r_i, e_i;
	double area1 = 0, area2 = 0, area12 = 0;
	double error1 = 0, error2 = 0, error12 = 0;
	double resasc1, resasc2;
	double resabs1, resabs2;
	  
	/* Bisect the subinterval with the largest error estimate */
	  
	retrieve (this->w, &a_i, &b_i, &r_i, &e_i);
	  
	a1 = a_i;
	b1 = 0.5 * (a_i + b_i);
	a2 = b1;
	b2 = b_i;
	  
	gsl_integration_qk_o2scl(func,qn,xgk,wg,wgk,fv1,fv2,a1,b1,
			       &area1,&error1,&resabs1,&resasc1,pa);
	gsl_integration_qk_o2scl(func,qn,xgk,wg,wgk,fv1,fv2,a2,b2,
			       &area2,&error2,&resabs2,&resasc2,pa);
	  
	area12 = area1 + area2;
	error12 = error1 + error2;
	  
	errsum += (error12 - e_i);
	area += area12 - r_i;
	  
	if (resasc1 != error1 && resasc2 != error2) {
	  double delta = r_i - area12;
	  
	  if (fabs (delta) <= 1.0e-5 * fabs (area12) && 
	      error12 >= 0.99 * e_i) {
	    roundoff_type1++;
	  }
	  if (iteration >= 10 && error12 > e_i) {
	    roundoff_type2++;
	  }
	}
	
	tolerance = GSL_MAX_DBL (l_epsabs, l_epsrel * fabs (area));
	  
	if (errsum > tolerance) {
	  if (roundoff_type1 >= 6 || roundoff_type2 >= 20) {
	    error_type = 2;   /* round off error */
	  }
	  /* set error flag in the case of bad integrand behaviour at
	     a point of the integration range */
	  
	  if (this->subinterval_too_small (a1, a2, b2)) {
	    error_type = 3;
	  }
	}
	
	update (this->w, a1, b1, area1, error1, a2, b2, area2, error2);
	  
	retrieve (this->w, &a_i, &b_i, &r_i, &e_i);
	  
	if (this->verbose>0) {
	  std::cout << "gsl_inte_qag Iter: " << iteration;
	  std::cout.setf(std::ios::showpos);
	  std::cout << " Res: " << area;
	  std::cout.unsetf(std::ios::showpos);
	  std::cout << " Err: " << errsum
		    << " Tol: " << tolerance << std::endl;
	  if (this->verbose>1) {
	    char ch;
	    std::cout << "Press a key and type enter to continue. " ;
	    std::cin >> ch;
	  }
	}
	  
	iteration++;
	
      } while (iteration < limit && !error_type && 
	       errsum > tolerance);
	  
      *result = sum_results (this->w);
      *abserr = errsum;
      
      this->last_iter=iteration;
      
      if (errsum <= tolerance) {
	return gsl_success;
      } else if (error_type == 2) {
	std::string estr="Roundoff error prevents tolerance ";
	estr+="from being achieved in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_eround);
      } else if (error_type == 3) {
	std::string estr="Bad integrand behavior ";
	estr+=" in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_esing);
      } else if (iteration == limit) {
	std::string estr="Maximum number of subdivisions ("+itos(iteration);
	estr+=") reached in gsl_inte_qag::qag().";
	set_err_ret(estr.c_str(),gsl_emaxiter);
      } else {
	std::string estr="Could not integrate function in gsl_inte_qag::";
	estr+="qag() (it may have returned a non-finite result).";
	set_err_ret(estr.c_str(),gsl_efailed);
      }
      
      // No return statement needed since the above if statement
      // always forces a return
    }
    
#endif

  public:
  
    /// Return string denoting type ("gsl_inte_qag")
    const char *type() { return "gsl_inte_qag"; }
  
  };

#ifndef DOXYGENP
}
#endif

#endif
