/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#include <o2scl/gsl_inte_qag_b.h>

using namespace o2scl;

gsl_inte_table::gsl_inte_table() {
  wkspace=1000;
  w=gsl_integration_workspace_alloc(1000);
}
    
gsl_inte_table::~gsl_inte_table() {
  gsl_integration_workspace_free(w);
}
    
int gsl_inte_table::set_wkspace(size_t size) {
  gsl_integration_workspace_free(w);
  wkspace=size;
  w=gsl_integration_workspace_alloc(size);
  return 0;
}
  
void gsl_inte_table::initialise(gsl_integration_workspace * workspace, 
				double a, double b) {
  workspace->size = 0;
  workspace->nrmax = 0;
  workspace->i = 0;
  workspace->alist[0] = a;
  workspace->blist[0] = b;
  workspace->rlist[0] = 0.0;
  workspace->elist[0] = 0.0;
  workspace->order[0] = 0;
  workspace->level[0] = 0;
      
  workspace->maximum_level = 0;
}
  
void gsl_inte_table::set_initial_result(gsl_integration_workspace * workspace,
					double result, double error) {
  workspace->size = 1;
  workspace->rlist[0] = result;
  workspace->elist[0] = error;
}
  
void gsl_inte_table::retrieve(const gsl_integration_workspace * workspace,
			      double * a, double * b, double * r, 
			      double * e) {
  const size_t i = workspace->i;
  double * alist = workspace->alist;
  double * blist = workspace->blist;
  double * rlist = workspace->rlist;
  double * elist = workspace->elist;
      
  *a = alist[i] ;
  *b = blist[i] ;
  *r = rlist[i] ;
  *e = elist[i] ;
}
  
void gsl_inte_table::qpsrt(gsl_integration_workspace * workspace) {

  const size_t last = workspace->size - 1;
  const size_t limit = workspace->limit;
      
  double * elist = workspace->elist;
  size_t * order = workspace->order;
      
  double errmax ;
  double errmin ;
  int i, k, top;
      
  size_t i_nrmax = workspace->nrmax;
  size_t i_maxerr = order[i_nrmax] ;
      
  /* Check whether the list contains more than two error estimates */
      
  if (last < 2)
    {
      order[0] = 0 ;
      order[1] = 1 ;
      workspace->i = i_maxerr ;
      return ;
    }
      
  errmax = elist[i_maxerr] ;
      
  /* This part of the routine is only executed if, due to a difficult
     integrand, subdivision increased the error estimate. In the normal
     case the insert procedure should start after the nrmax-th largest
     error estimate. */
      
  while (i_nrmax > 0 && errmax > elist[order[i_nrmax - 1]])
    {
      order[i_nrmax] = order[i_nrmax - 1] ;
      i_nrmax-- ;
    }
      
  /* Compute the number of elements in the list to be maintained in
     descending order. This number depends on the number of
     subdivisions still allowed. */
  if(last < (limit/2 + 2))
    {
      top = last ;
    }
  else
    {
      top = limit - last + 1;
    }
      
  /* Insert errmax by traversing the list top-down, starting
     comparison from the element elist(order(i_nrmax+1)). */
      
  i = i_nrmax + 1 ;
      
  /* The order of the tests in the following line is important to
     prevent a segmentation fault */
      
  while (i < top && errmax < elist[order[i]])
    {
      order[i-1] = order[i] ;
      i++ ;
    }
      
  order[i-1] = i_maxerr ;
      
  /* Insert errmin by traversing the list bottom-up */
      
  errmin = elist[last] ;
      
  k = top - 1 ;
      
  while (k > i - 2 && errmin >= elist[order[k]])
    {
      order[k+1] = order[k] ;
      k-- ;
    }
      
  order[k+1] = last ;
      
  /* Set i_max and e_max */
      
  i_maxerr = order[i_nrmax] ;
      
  workspace->i = i_maxerr ;
  workspace->nrmax = i_nrmax ;
}

void gsl_inte_table::update(gsl_integration_workspace * workspace,
			    double a1, double b1, double area1, double error1,
			    double a2, double b2, double area2, 
			    double error2) {
  double * alist = workspace->alist ;
  double * blist = workspace->blist ;
  double * rlist = workspace->rlist ;
  double * elist = workspace->elist ;
  size_t * level = workspace->level ;
      
  const size_t i_max = workspace->i ;
  const size_t i_new = workspace->size ;
      
  const size_t new_level = workspace->level[i_max] + 1;
      
  /* append the newly-created intervals to the list */
      
  if (error2 > error1)
    {
      alist[i_max] = a2;        /* blist[maxerr] is already == b2 */
      rlist[i_max] = area2;
      elist[i_max] = error2;
      level[i_max] = new_level;
	  
      alist[i_new] = a1;
      blist[i_new] = b1;
      rlist[i_new] = area1;
      elist[i_new] = error1;
      level[i_new] = new_level;
    }
  else
    {
      blist[i_max] = b1;        /* alist[maxerr] is already == a1 */
      rlist[i_max] = area1;
      elist[i_max] = error1;
      level[i_max] = new_level;
	  
      alist[i_new] = a2;
      blist[i_new] = b2;
      rlist[i_new] = area2;
      elist[i_new] = error2;
      level[i_new] = new_level;
    }
      
  workspace->size++;
      
  if (new_level > workspace->maximum_level)
    {
      workspace->maximum_level = new_level;
    }
  qpsrt (workspace) ;
}

double gsl_inte_table::sum_results
(const gsl_integration_workspace *workspace) {

  const double * const rlist = workspace->rlist ;
  const size_t n = workspace->size;
      
  size_t k;
  double result_sum = 0;
      
  for (k = 0; k < n; k++)
    {
      result_sum += rlist[k];
    }
      
  return result_sum;
}
    
int gsl_inte_table::subinterval_too_small(double a1, double a2, 
					  double b2) {
  const double e = GSL_DBL_EPSILON;
  const double u = GSL_DBL_MIN;
	
  double tmp = (1 + 100 * e) * (fabs (a2) + 1000 * u);
	
  int status = fabs (a1) <= tmp && fabs (b2) <= tmp;
	
  return status;
}
    
void gsl_inte_table::append_interval(gsl_integration_workspace * workspace,
				      double a1, double b1, double area1, 
				      double error1) {

  const size_t i_new = workspace->size ;
  
  workspace->alist[i_new] = a1;
  workspace->blist[i_new] = b1;
  workspace->rlist[i_new] = area1;
  workspace->elist[i_new] = error1;
  workspace->order[i_new] = i_new;
  workspace->level[i_new] = 0;

  workspace->size++;
}
