/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** \file lu_base.h 
    \brief File for LU decomposition and associated solver
*/

#ifdef DOXYGENP
namespace o2scl_linalg {
#endif

  /** 
      \brief Compute the LU decomposition of the matrix \c A

      On output the diagonal and upper triangular part of the input
      matrix A contain the matrix U. The lower triangular part of the
      input matrix (excluding the diagonal) contains L. The diagonal
      elements of L are unity, and are not stored.

      The permutation matrix P is encoded in the permutation p. The
      j-th column of the matrix P is given by the k-th column of the
      identity matrix, where k = p_j the j-th element of the
      permutation vector. The sign of the permutation is given by
      signum. It has the value (-1)^n, where n is the number of
      interchanges in the permutation.
      
      The algorithm used in the decomposition is Gaussian Elimination
      with partial pivoting (Golub & Van Loan, Matrix Computations,
      Algorithm 3.4.1).
  */
  template<class mat_t>
    int LU_decomp(const size_t N, mat_t &A, o2scl::permutation &p, 
		  int &signum) {
    
    size_t i, j, k;
  
    signum=1;
    p.init();
  
    for (j = 0; j < N - 1; j++) {
    
      /* Find maximum in the j-th column */
      double ajj, max = fabs(O2SCL_IX2(A,j,j));
      size_t i_pivot = j;
    
      for (i = j + 1; i < N; i++) {
	double aij = fabs (O2SCL_IX2(A,i,j));
      
	if (aij > max) {
	  max = aij;
	  i_pivot = i;
	}
      }

      if (i_pivot != j) {

	// Swap rows j and i_pivot
	double temp;
	for (k=0;k<N;k++) {
	  temp=O2SCL_IX2(A,j,k);
	  O2SCL_IX2(A,j,k)=O2SCL_IX2(A,i_pivot,k);
	  O2SCL_IX2(A,i_pivot,k)=temp;
	}
	p.swap(j,i_pivot);
	signum=-signum;
      }
    
      ajj = O2SCL_IX2(A,j,j);
      
      if (ajj != 0.0) {
	for (i = j + 1; i < N; i++) {
	  double aij = O2SCL_IX2(A,i,j) / ajj;
	  O2SCL_IX2(A,i,j)=aij;
	  for (k = j + 1; k < N; k++) {
	    double aik = O2SCL_IX2(A,i,k);
	    double ajk = O2SCL_IX2(A,j,k);
	    O2SCL_IX2(A,i,k)=aik - aij * ajk;
	  }
	}
      }
    }
  
    return o2scl::gsl_success;
  }

  /** 
      \brief Solve a linear system after LU decomposition

      This function solve the square system A x = b using the LU
      decomposition of A into (LU, p) given by gsl_linalg_LU_decomp or
      gsl_linalg_complex_LU_decomp.
  */
  template<class mat_t, class vec_t>
    int LU_solve(const size_t N, const mat_t &LU, const o2scl::permutation &p, 
		 const vec_t &b, vec_t &x) {
  
    /* Copy x <- b */
    o2scl::vector_copy(N,b,x);
  
    /* Solve for x */
    LU_svx(N,LU,p,x);
  
    return o2scl::gsl_success;
  }

  /** 
      \brief Solve a linear system after LU decomposition in place
      
      These functions solve the square system A x = b in-place using
      the LU decomposition of A into (LU,p). On input x should contain
      the right-hand side b, which is replaced by the solution on
      output.
  */
  template<class mat_t, class vec_t>
    int LU_svx(const size_t N, const mat_t &LU, 
	       const o2scl::permutation &p, vec_t &x) {
  
    /* Apply permutation to RHS */
    p.apply(x);
  
    /* Solve for c using forward-substitution, L c = P b */
    o2scl_cblas::dtrsv(o2scl_cblas::O2cblasRowMajor,o2scl_cblas::O2cblasLower,
		       o2scl_cblas::O2cblasNoTrans,o2scl_cblas::O2cblasUnit,
		       N,N,LU,x);
  
    /* Perform back-substitution, U x = c */
    o2scl_cblas::dtrsv(o2scl_cblas::O2cblasRowMajor,o2scl_cblas::O2cblasUpper,
		       o2scl_cblas::O2cblasNoTrans,o2scl_cblas::O2cblasNonUnit,
		       N,N,LU,x);
  
    return o2scl::gsl_success;
  }

  /**
     \brief Refine the solution of a linear system
     
     These functions apply an iterative improvement to x, the solution
     of A x = b, using the LU decomposition of A into (LU,p). The
     initial residual r = A x - b is also computed and stored in
     residual.
  */
  template<class mat_t, class vec_t> 
    int LU_refine(const size_t N, const mat_t &A, const mat_t &LU,
		  const o2scl::permutation &p, const vec_t &b, vec_t &x,
		  vec_t &residual) {
  
    /* Compute residual, residual = (A * x  - b) */
    o2scl::vector_copy(N,b,residual);
    o2scl_cblas::dgemv(o2scl_cblas::O2cblasRowMajor,
		       o2scl_cblas::O2cblasNoTrans,
		       N,N,1.0,A,x,-1.0,residual);
  
    /* Find correction, delta = - (A^-1) * residual, and apply it */
  
    LU_svx(N,LU,p,residual);
    o2scl_cblas::daxpy(5,-1.0,residual,x);
  
    return o2scl::gsl_success;
  }

  /** 
      \brief Compute the inverse of a matrix from its LU decomposition

      These functions compute the inverse of a matrix A from its LU
      decomposition (LU,p), storing the result in the matrix inverse.
      The inverse is computed by solving the system A x = b for each
      column of the identity matrix. It is preferable to avoid direct
      use of the inverse whenever possible, as the linear solver
      functions can obtain the same result more efficiently and
      reliably (consult any introductory textbook on numerical linear
      algebra for details).

      \future could rewrite to avoid mat_col_t
  */
  template<class mat_t, class mat_col_t> 
    int LU_invert(const size_t N, const mat_t &LU, 
		  const o2scl::permutation &p, mat_t &inverse) {
    size_t i;
  
    int status=o2scl::gsl_success;
  
    // Set matrix 'inverse' to the identity
    for(i=0;i<N;i++) {
      for(size_t j=0;j<N;j++) {
	if (i==j) O2SCL_IX2(inverse,i,j)=1.0;
	else O2SCL_IX2(inverse,i,j)=0.0;
      }
    }
  
    for (i = 0; i < N; i++) {
      mat_col_t c(inverse,i);
      int status_i=LU_svx(N,LU,p,c);
    
      if (status_i) {
	status = status_i;
      }
    }
  
    return status;
  }

  /**
     \brief Compute the determinant of a matrix from its LU decomposition

     These functions compute the determinant of a matrix A from its LU
     decomposition, LU. The determinant is computed as the product of
     the diagonal elements of U and the sign of the row permutation
     signum.
  */
  template<class mat_t> 
    double LU_det(const size_t N, const mat_t &LU, int signum) {

    size_t i;

    double det=(double)signum;
  
    for (i=0;i<N;i++) {
      det*=O2SCL_IX2(LU,i,i);
    }
  
    return det;
  }

  /**
     \brief Compute the logarithm of the absolute value of the
     determinant of a matrix from its LU decomposition

     These functions compute the logarithm of the absolute value of
     the determinant of a matrix A, \f$ \ln|\det(A)| \f$, from its LU
     decomposition, LU. This function may be useful if the direct
     computation of the determinant would overflow or underflow.
  */
  template<class mat_t> 
    double LU_lndet(const size_t N, const mat_t &LU) {
  
    size_t i;
    double lndet = 0.0;
  
    for (i = 0; i < N; i++) {
      lndet+=log(fabs(O2SCL_IX2(LU,i,i)));
    }
  
    return lndet;
  }

  /** 
      \brief Compute the sign of the 
      determinant of a matrix from its LU decomposition

      These functions compute the sign or phase factor of the
      determinant of a matrix A, \f$ \det(A)/|\det(A)| \f$, from its
      LU decomposition, LU.
  */
  template<class mat_t> 
    int LU_sgndet(const size_t N, const mat_t &LU, int signum) {

    size_t i;
    int s=signum;

    for (i=0;i<N;i++) {
      double u=O2SCL_IX2(LU,i,i);
      if (u<0) {
	s*=-1;
      } else if (u == 0) {
	s=0;
	break;
      }
    }

    return s;
  }

#ifdef DOXYGENP
}
#endif
