/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MISER_H
#define O2SCL_GSL_MISER_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/mcarlo_inte.h>
#include <o2scl/gsl_rnga.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_monte.h>
#include <gsl/gsl_machine.h>
#include <gsl/gsl_monte_miser.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Multidimensional integration using Miser Miser Carlo (GSL)

      \todo Document the fact that min_calls and
      min_calls_per_bisection need to be set beforehand
  */
  template<class param_t, class func_t=multi_funct<param_t>, 
    class rng_t=gsl_rnga, class vec_t=ovector_view, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc> class gsl_miser : 
  public mcarlo_inte<param_t,func_t,rng_t,vec_t> {

    public:

    /** 
	\brief Introduce random variation into bisection (default 0.0)

	From GSL documentation:
	\verbatim
	This parameter introduces a random fractional variation of size
	DITHER into each bisection, which can be used to break the
	symmetry of integrands which are concentrated near the exact
	center of the hypercubic integration region.  The default value of
	dither is zero, so no variation is introduced. If needed, a
	typical value of DITHER is 0.1.
	\endverbatim
    */
    double dither;

    /** 
	\brief Specify fraction of function calls for estimating variance
	
	From GSL documentation:
	\verbatim
	This parameter specifies the fraction of the currently available
	number of function calls which are allocated to estimating the
	variance at each recursive step. The default value is 0.1.
	\endverbatim
    */
    double estimate_frac;

    /** 
	\brief How estimated variances for two sub-regions are combined

	From GSL documentation:
	\verbatim
	This parameter controls how the estimated variances for the two
	sub-regions of a bisection are combined when allocating points.
	With recursive sampling the overall variance should scale better
	than 1/N, since the values from the sub-regions will be obtained
	using a procedure which explicitly minimizes their variance.  To
	accommodate this behavior the MISER algorithm allows the total
	variance to depend on a scaling parameter \alpha,
	
	\Var(f) = {\sigma_a \over N_a^\alpha} + {\sigma_b \over N_b^\alpha}.
	
	The authors of the original paper describing MISER recommend the
	value \alpha = 2 as a good choice, obtained from numerical
	experiments, and this is used as the default value in this
	implementation.
	\endverbatim
    */
    double alpha;

    /** 
	\brief Minimum number of calls to estimate the variance
	(default 100)

	From GSL documentation:
	\verbatim
	This parameter specifies the minimum number of function calls
	required for each estimate of the variance. If the number of
	function calls allocated to the estimate using ESTIMATE_FRAC falls
	below MIN_CALLS then MIN_CALLS are used instead.  This ensures
	that each estimate maintains a reasonable level of accuracy.  The
	default value of MIN_CALLS is `16 * dim'.
	\endverbatim
    */
    size_t min_calls;

    /** 
	\brief Minimum number of calls required to proceed with bisection
	(default 4000)

	From GSL documentation:
	\verbatim
	This parameter specifies the minimum number of function calls
	required to proceed with a bisection step.  When a recursive step
	has fewer calls available than MIN_CALLS_PER_BISECTION it performs
	a plain Monte Carlo estimate of the current sub-region and
	terminates its branch of the recursion.  The default value of this
	parameter is `32 * min_calls'.
	\endverbatim
    */
    size_t min_calls_per_bisection;

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Desc
    size_t dim;
    /// Desc
    double *xmid;
    /// Desc
    double *sigma_l;
    /// Desc
    double *sigma_r;
    /// Desc
    double *fmax_l;
    /// Desc
    double *fmax_r;
    /// Desc
    double *fmin_l;
    /// Desc
    double *fmin_r;
    /// Desc
    double *fsum_l;
    /// Desc
    double *fsum_r;
    /// Desc
    double *fsum2_l;
    /// Desc
    double *fsum2_r;
    /// Desc
    size_t *hits_l;
    /// Desc
    size_t *hits_r;
    
    /// Desc
    virtual int estimate_corrmc(func_t &func, size_t ndim,
				const vec_t &xl, const vec_t &xu,
				param_t &pa, size_t calls, double &res,
				double &err, const double lxmid[],
				double lsigma_l[], double lsigma_r[]) {
      size_t i, n;
      
      double m = 0.0, q = 0.0;
      double vol = 1.0;

      for (i = 0; i < dim; i++) {
	vol *= xu[i] - xl[i];
	hits_l[i] = hits_r[i] = 0;
	fsum_l[i] = fsum_r[i] = 0.0;
	fsum2_l[i] = fsum2_r[i] = 0.0;
	lsigma_l[i] = lsigma_r[i] = -1;
      }

      for (n = 0; n < calls; n++) {
	double fval;

	unsigned int j = (n/2) % dim;
	unsigned int side = (n % 2);

	for (i = 0; i < dim; i++) {

	  // The equivalent of gsl_rng_uniform_pos()
	  double z;
	  do { z=this->def_rng.random(); } while (z==0);
	  
	  if (i != j) {
	    x[i] = xl[i] + z * (xu[i] - xl[i]);
	  } else {
	    if (side == 0) {
	      x[i] = lxmid[i] + z * (xu[i] - lxmid[i]);
	    } else {
	      x[i] = xl[i] + z * (lxmid[i] - xl[i]);
	    }
	  }
	}
	
	func(ndim,x,fval,pa);
	
	/* recurrence for mean and variance */
	{
	  double d = fval - m;
	  m += d / (n + 1.0);
	  q += d * d * (n / (n + 1.0));
	}

	/* compute the variances on each side of the bisection */
	for (i = 0; i < dim; i++) {
	  if (x[i] <= lxmid[i]) {
	    fsum_l[i] += fval;
	    fsum2_l[i] += fval * fval;
	    hits_l[i]++;
	  } else {
	    fsum_r[i] += fval;
	    fsum2_r[i] += fval * fval;
	    hits_r[i]++;
	  }
	}
      }

      for (i = 0; i < dim; i++)	{
	double fraction_l = (lxmid[i] - xl[i]) / (xu[i] - xl[i]);

	if (hits_l[i] > 0) {
	  fsum_l[i] /= hits_l[i];
	  lsigma_l[i] = sqrt (fsum2_l[i] - fsum_l[i] * fsum_l[i] / hits_l[i]);
	  lsigma_l[i] *= fraction_l * vol / hits_l[i];
	}
	
	if (hits_r[i] > 0) {
	  fsum_r[i] /= hits_r[i];
	  lsigma_r[i] = sqrt (fsum2_r[i] - fsum_r[i] * fsum_r[i] / hits_r[i]);
	  lsigma_r[i] *= (1 - fraction_l) * vol / hits_r[i];
	}
      }
      
      res = vol * m;
      
      if (calls < 2) {
	err = GSL_POSINF;
      } else {
	err = vol * sqrt (q / (calls * (calls - 1.0)));
      }
      
      return gsl_success;
    }

    /// Memory allocator
    alloc_t ao;

    /// The most recent integration point
    alloc_vec_t x;

#endif
    
    public:
    
    gsl_miser() {
      estimate_frac=0.1;
      alpha=2.0;
      dither=0.0;
      min_calls=100;
      min_calls_per_bisection=4000;
    }
    
    /// Allocate memory
    virtual int allocate(size_t ldim) {

      ao.allocate(x,ldim);
       
      xmid = (double *) malloc (ldim * sizeof (double));
      if (xmid == 0) {
	ao.free(x);
	set_err2_ret("Failed to allocate space for xmid",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      sigma_l = (double *) malloc (ldim * sizeof (double));
      if (sigma_l == 0) {
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for sigma_l",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      sigma_r = (double *) malloc (ldim * sizeof (double));
      if (sigma_r == 0) {
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for sigma_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fmax_l = (double *) malloc (ldim * sizeof (double));
      if (fmax_l == 0) {
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fmax_l",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fmax_r = (double *) malloc (ldim * sizeof (double));
      if (fmax_r == 0) {
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fmax_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fmin_l = (double *) malloc (ldim * sizeof (double));
      if (fmin_l == 0) {
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fmin_l",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fmin_r = (double *) malloc (ldim * sizeof (double));
      if (fmin_r == 0) {
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fmin_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fsum_l = (double *) malloc (ldim * sizeof (double));
      if (fsum_l == 0) {
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum_l",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fsum_r = (double *) malloc (ldim * sizeof (double));
      if (fsum_r == 0) {
	std::free(fsum_l);
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fsum2_l = (double *) malloc (ldim * sizeof (double));
      if (fsum2_l == 0) {
	std::free(fsum_r);
	std::free(fsum_l);
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum2_l",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      fsum2_r = (double *) malloc (ldim * sizeof (double));
      if (fsum2_r == 0) {
	std::free(fsum2_l);
	std::free(fsum_r);
	std::free(fsum_l);
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum2_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }


      hits_r = (size_t *) malloc (ldim * sizeof (size_t));
      if (hits_r == 0) {
	std::free(fsum2_r);
	std::free(fsum2_l);
	std::free(fsum_r);
	std::free(fsum_l);
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum2_r",
		     " in gsl_miser::allocate().",gsl_enomem);
      }

      hits_l = (size_t *) malloc (ldim * sizeof (size_t));
      if (hits_l == 0) {
	std::free(hits_r);
	std::free(fsum2_r);
	std::free(fsum2_l);
	std::free(fsum_r);
	std::free(fsum_l);
	std::free(fmin_r);
	std::free(fmin_l);
	std::free(fmax_r);
	std::free(fmax_l);
	std::free(sigma_r);
	std::free(sigma_l);
	std::free(xmid);
	ao.free(x);
	set_err2_ret("Failed to allocate space for fsum2_r",
		     " in gsl_miser::allocate().",gsl_enomem); 
      }

      dim = ldim;

      return 0;
    }

    /// Free allocated memory
    virtual int free() {
      std::free(hits_r);
      std::free(hits_l);
      std::free(fsum2_r);
      std::free(fsum2_l);
      std::free(fsum_r);
      std::free(fsum_l);
      std::free(fmin_r);
      std::free(fmin_l);
      std::free(fmax_r);
      std::free(fmax_l);
      std::free(sigma_r);
      std::free(sigma_l);
      std::free(xmid);
      ao.free(x);
      return 0;
    }
    
    /** \brief Integrate function \c func over the hypercube from
	\f$ x_i=\mathrm{xl}_i \f$ to \f$ x_i=\mathrm{xu}_i \f$ for
	\f$ 0<i< \f$ ndim-1
    */
    virtual int miser_minteg_err(func_t &func, size_t ndim, const vec_t &xl, 
				 const vec_t &xu, size_t calls, param_t &pa,
				 double &res, double &err) {
      
      size_t n, estimate_calls, calls_l, calls_r;
      size_t i;
      size_t i_bisect;
      int found_best;

      double res_est = 0, err_est = 0;
      double res_r = 0, err_r = 0, res_l = 0, err_l = 0;
      double xbi_l, xbi_m, xbi_r, s;

      double vol;
      double weight_l, weight_r;

      for (i = 0; i < dim; i++)	{
	if (xu[i] <= xl[i]) {
	  set_err2_ret("Upper limit must be greater than lower limit ",
		       "in gsl_miser::miser_minteg_err().",gsl_einval);
	}
	if (xu[i] - xl[i] > GSL_DBL_MAX) {
	  set_err2_ret("Range of integration is too large ",
		       "in gsl_miser::miser_minteg_err().",gsl_einval);
	}
      }

      if (alpha < 0) {
	set_err2_ret("Parameter 'alpha' must be non-negative ", 
		     "in gsl_miser::miser_minteg_err().",gsl_einval);
      }

      /* Compute volume */

      vol = 1;

      for (i = 0; i < dim; i++)	{
	vol *= xu[i] - xl[i];
      }

      if (calls < min_calls_per_bisection) {
	double m = 0.0, q = 0.0;

	if (calls < 2) {
	  set_err2_ret("Insufficient calls for subvolume ", 
		       "in gsl_miser::miser_minteg_err().",gsl_einval);
	}

	for (n = 0; n < calls; n++) {
	  /* Choose a random point in the integration region */

	  for (i = 0; i < dim; i++) {
	    
	    // The equivalent of gsl_rng_uniform_pos()
	    double rdn;
	    do { rdn=this->def_rng.random(); } while (rdn==0);

	    x[i] = xl[i] + rdn * (xu[i] - xl[i]);
	  }

	  {
	    double fval;
	    func(ndim,x,fval,pa);
	    
	    /* recurrence for mean and variance */

	    double d = fval - m;
	    m += d / (n + 1.0);
	    q += d * d * (n / (n + 1.0));
	  }
	}

	res = vol * m;

	err = vol * sqrt (q / (calls * (calls - 1.0)));

	return gsl_success;
      }
      
      // We explicitly make the typecast here
      estimate_calls=((size_t)GSL_MAX(min_calls,calls*(estimate_frac)));
      
      if (estimate_calls < 4 * dim) {
	set_err2_ret("Insufficient calls to sample all halfspaces ", 
		     "in gsl_miser::miser_minteg_err().",gsl_esanity);
      }

      /* Flip coins to bisect the integration region with some fuzz */

      for (i = 0; i < dim; i++) {
	s=(this->def_rng.random()-0.5) >= 0.0 ? dither : -dither;
	xmid[i] = (0.5 + s) * xl[i] + (0.5 - s) * xu[i];
      }

      /* 
	 The idea is to chose the direction to bisect based on which will
	 give the smallest total variance.  We could (and may do so later)
	 use MC to compute these variances.  But the NR guys simply estimate
	 the variances by finding the min and max function values
	 for each half-region for each bisection. 
      */
      
      estimate_corrmc(func,dim,xl,xu,pa,estimate_calls,
		      res_est,err_est,xmid,sigma_l,sigma_r);

      /* We have now used up some calls for the estimation */

      calls -= estimate_calls;

      /* Now find direction with the smallest total "variance" */

      {
	double best_var = GSL_DBL_MAX;
	double beta = 2.0 / (1.0 + alpha);
	found_best = 0;
	i_bisect = 0;
	weight_l = weight_r = 1.0;

	for (i = 0; i < dim; i++) {
	  if (sigma_l[i] >= 0 && sigma_r[i] >= 0) {
	    /* estimates are okay */
	    double var = pow (sigma_l[i], beta) + pow (sigma_r[i], beta);
	    
	    if (var <= best_var) {
	      found_best = 1;
	      best_var = var;
	      i_bisect = i;
	      weight_l = pow (sigma_l[i], beta);
	      weight_r = pow (sigma_r[i], beta);
	    }
	  } else {
	    if (sigma_l[i] < 0) {
	      set_err2_ret("No points in left-half space ", 
			   "in gsl_miser::miser_minteg_err().",gsl_esanity);
	    }
	    if (sigma_r[i] < 0) {
	      set_err2_ret("No points in right-half space ", 
			   "in gsl_miser::miser_minteg_err().",gsl_esanity);
	    }
	  }
	}
      }
      
      if (!found_best)	{
	/* All estimates were the same, so chose a direction at random */
	
	i_bisect = this->def_rng.random_int(dim);
      }

      xbi_l = xl[i_bisect];
      xbi_m = xmid[i_bisect];
      xbi_r = xu[i_bisect];

      /* Get the actual fractional sizes of the two "halves", and
	 distribute the remaining calls among them */

      {
	double fraction_l = fabs ((xbi_m - xbi_l) / (xbi_r - xbi_l));
	double fraction_r = 1 - fraction_l;

	double a = fraction_l * weight_l;
	double b = fraction_r * weight_r;

	calls_l = (size_t)(min_calls + (calls - 2 * min_calls) * a / (a + b));
	calls_r = (size_t)(min_calls + (calls - 2 * min_calls) * b / (a + b));
      }

      /* Compute the integral for the left hand side of the bisection */

      /* Due to the recursive nature of the algorithm we must allocate
	 some new memory for the integration limits for each recursive call */

      {
	int status;
	
	alloc_vec_t xu_tmp;
	ao.allocate(xu_tmp,dim);
	
	for (i = 0; i < dim; i++) {
	  xu_tmp[i] = xu[i];
	}
	
	xu_tmp[i_bisect] = xbi_m;
	
	status=miser_minteg_err(func,dim,xl,xu_tmp,calls_l,pa,res_l,err_l);

	ao.free(xu_tmp);

	if (status != gsl_success) {
	  return status;
	}
      }

      /* Compute the integral for the right hand side of the bisection */

      {
	int status;

	alloc_vec_t xl_tmp;
	ao.allocate(xl_tmp,dim);

	for (i = 0; i < dim; i++) {
	  xl_tmp[i] = xl[i];
	}

	xl_tmp[i_bisect] = xbi_m;

	status=miser_minteg_err(func,dim,xl_tmp,xu,calls_r,pa,res_r,err_r);
	
	ao.free(xl_tmp);

	if (status != gsl_success) {
	  return status;
	}
      }

      res=res_l+res_r;
      err=sqrt(err_l*err_l+err_r*err_r);
      
      return 0;
    }
    
    virtual ~gsl_miser() {}
    
    /// Integrate function \c func from x=a to x=b.
    virtual int minteg_err(func_t &func, size_t ndim, const vec_t &a, 
			   const vec_t &b, param_t &pa, double &res, 
			   double &err) {
      allocate(ndim);
      min_calls=16*ndim;
      min_calls_per_bisection=32*min_calls;
      int ret=miser_minteg_err(func,ndim,a,b,this->n_points,pa,res,err);
      free();
      return ret;
    }
    
    /** \brief Integrate function \c func over the hypercube from
	\f$ x_i=a_i \f$ to \f$ x_i=b_i \f$ for
	\f$ 0<i< \f$ ndim-1
    */
    virtual double minteg(func_t &func, size_t ndim, const vec_t &a, 
			  const vec_t &b, param_t &pa) {
      double res;
      minteg_err(func,ndim,a,b,pa,res,this->interror);
      return res;
    }

    /// Return string denoting type ("gsl_miser")
    virtual const char *type() { return "gsl_miser"; }
    
  };
  
#ifndef DOXYGENP
}
#endif

#endif

