/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_MONTE_H
#define O2SCL_GSL_MONTE_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/mcarlo_inte.h>
#include <o2scl/gsl_rnga.h>
#include <gsl/gsl_monte.h>
#include <gsl/gsl_monte_plain.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Multidimensional integration using plain Monte Carlo (GSL)

  */
  template<class param_t, class func_t, class rng_t=gsl_rnga, 
    class vec_t=ovector_view, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc> class gsl_monte : 
  public mcarlo_inte<param_t,func_t,rng_t,vec_t> {

    public:

    virtual ~gsl_monte() {}

    /// Integrate function \c func from x=a to x=b.
    virtual int minteg_err(func_t &func, size_t ndim, const vec_t &a, 
			   const vec_t &b, param_t &pa, double &res, 
			   double &err) 
    {

      double r;

      alloc_vec_t cc;
      alloc_t ao;
      ao.allocate(cc,ndim);
      
      double vol=1.0, m=0.0, q=0.0;
      for(size_t i=0;i<ndim;i++) vol*=b[i]-a[i];
      
      for(int n=0;n<this->n_points;n++) {
	for(size_t i=0;i<ndim;i++) {
	  do {
	    r=this->def_rng.random();
	  } while (r==0.0);
	  cc[i]=a[i]+r*(b[i]-a[i]);
	}
	double fval;
	func(ndim,cc,fval,pa);
	double d=fval-m;
	m+=d/(n+1.0);
	q+=d*d*(n/(n+1.0));
      }

      res=vol*m;

      if (this->n_points<2) {
	err=0.0;
      } else {
	err=vol*sqrt(q/(this->n_points*(this->n_points-1.0)));
      }
	  
      ao.free(cc);

      return 0;

    }
      
    /** \brief Integrate function \c func over the hypercube from
	\f$ x_i=a_i \f$ to \f$ x_i=b_i \f$ for
	\f$ 0<i< \f$ ndim-1
    */
    virtual double minteg(func_t &func, size_t ndim, const vec_t &a, 
			  const vec_t &b, param_t &pa) {
      double res;
      minteg_err(func,ndim,a,b,pa,res,this->interror);
      return res;
    }

    /// Return string denoting type ("gsl_monte")
    virtual const char *type() { return "gsl_monte"; }
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif

