/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NUCLEAR_DIST_H
#define O2SCL_NUCLEAR_DIST_H

#include <iostream>
#include <o2scl/nucleus.h>
#include <o2scl/nuclear_mass.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief A distribution of nuclei
      
      The virtual base class for a collection of objects of type \ref
      nucleus . See \ref simple_dist and \ref full_dist for
      implmentations of this base class.
   */
  class nuclear_dist {

#ifndef DOXYGENP

  protected:

    virtual nucleus *next(nucleus *np)=0;

#endif
    
  public:

    virtual ~nuclear_dist() {}
    
    /** 
	\brief An iterator for the nuclear distribution
	
	The standard usage of this iterator is something of the form:
	\code
	mnmsk_mass mth;
	simple_dist sd(5,6,10,12,&mth);
	for(nuclear_dist::iterator ndi=sd.begin();ndi!=sd.end();ndi++) {
	// do something here for each nucleus
	}
	\endcode
	which would create a list consisting of three isotopes (A=10,
	11, and 12) of boron and three isotopes carbon for a total
	of six nuclei.
    */
    class iterator {
      
#ifndef DOXYGEN_INTERNAL

    protected:

      // The iterator is really just a wrapper for the pointer
      nucleus *np;

      /// A pointer to the distribution
      nuclear_dist *ndp;

#endif
      
    public:

#ifndef DOXYGEN_INTERNAL

      /** \brief Create an iterator from the given distribution using the
	  nucleus specified in \c npp.
      */
      iterator(nuclear_dist *ndpp, nucleus *npp) {
	ndp=ndpp;
	np=npp;
      }

#endif

      /// Proceed to the next nucleus
      iterator operator++() {
	np=ndp->next(np);
	return iterator(ndp,np);
      }

      /// Proceed to the next nucleus
      iterator operator++(int unused) {
	nucleus *tmp=np;
	np=ndp->next(np);
	return iterator(ndp,tmp);
      }

      /// Dereference the iterator
      nucleus *operator->() const {
	return np;
      };
      
      friend int operator==(const nuclear_dist::iterator &i1,
			    const nuclear_dist::iterator &i2);

      friend int operator!=(const nuclear_dist::iterator &i1,
			    const nuclear_dist::iterator &i2);
    };

    /// The beginning of the distribution
    virtual iterator begin()=0;

    /// The end of the distribution
    virtual iterator end()=0;

    /// The number of nuclei in the distribution
    virtual size_t size()=0;
  };

  /// Compare two nuclei
  int operator==(const nuclear_dist::iterator &i1,
		 const nuclear_dist::iterator &i2);

  /// Compare two nuclei
  int operator!=(const nuclear_dist::iterator &i1,
		const nuclear_dist::iterator &i2);

  /** 
      \brief A simple nuclear distribution given a range in A and Z
      
      The iterator for this distribution begins with the nucleus with
      the lowest Z and A, and increases A before incrementing Z and
      beginning again with the lowest A for that value of Z. In other
      words, it proceeds through all the isotopes of an element first,
      and then proceeds to the next element.

      For example, to create a collection of isotopes of Carbon,
      Nitrogen and Oxygen using the most recent (2003) Atomic Mass
      Evaluation, and then output the nuclei in the collection
      \code
      ame_mass ame;
      simple_dist fd(6,8,2,30,&ame);
      for(nuclear_dist::iterator ndi=fd.begin();ndi!=fd.end();ndi++) {
        cout << ndi->Z << " " << ndi->A << " " << ndi->m << endl;
      }
      \endcode

      \todo This takes a nuclear_mass pointer as input. It should
      probably be a reference instead? (3/18/08 - Yeah, I think
      it should be a reference.)
      \todo Add error checking in constructors and set functions
  */
  class simple_dist : public nuclear_dist {
  public:

    simple_dist();

    /** \brief Create a distribution from ranges in A specified
	for each Z
	
	The length of the arrays minA and maxA should be exactly
	\f$\mathrm{maxZ}-\mathrm{minZ}+1\f$.
    */
    simple_dist(int minZ, int maxZ, int minA[], int maxA[],
		nuclear_mass *nm);
  
    /// Create a square distribution in A and Z
    simple_dist(int minZ, int maxZ, int minA, int maxA,
		nuclear_mass *nm);
  
    virtual ~simple_dist();

    /// The beginning of the distribution
    virtual iterator begin() {
      return iterator(this,list);
    };

    /// The end of the distribution
    virtual iterator end() {
      return iterator(this,list+list_size);
    };

    /// The number of nuclei in the distribution
    virtual size_t size() {
      return list_size;
    };

    /** \brief Set the distribution from ranges in A specified
	for each Z
	
	The length of the arrays minA and maxA should be exactly
	\f$\mathrm{maxZ}-\mathrm{minZ}+1\f$.
    */
    int set_dist(int minZ, int maxZ, int minA[], int maxA[],
		nuclear_mass *nm);
  
    /// Set a square distribution in A and Z
    int set_dist(int minZ, int maxZ, int minA, int maxA,
		nuclear_mass *nm);
  
#ifndef DOXYGENP

  protected:

    /// Specify how to move to the next nucleus
    virtual nucleus *next(nucleus *np) {
      return np+1;
    }

    /// The distribution of nuclei as an array
    nucleus *list;
    
    /// The minimum atomic number for a given Z
    int *min_A;
    
    /// The maximum atomic number for a given Z
    int *max_A;

    /// The minimum proton number
    int min_Z;
    
    /// The maximum proton number
    int max_Z;

    /// The size of \ref list array
    size_t list_size;

    /// True if memory has been allocated
    bool alloc;

#endif

  };

  /** 
      \brief Full distribution including all nuclei from a
      discrete mass formula

      For example, to create a collection of all nuclei from the most
      recent (2003) Atomic Mass Evaluation, and then output all the
      nuclei in the collection
      \code
      ame_mass ame;
      full_dist fd(&ame);
      for(nuclear_dist::iterator ndi=fd.begin();ndi!=fd.end();ndi++) {
        cout << ndi->Z << " " << ndi->A << " " << ndi->m << endl;
      }
      \endcode
      
  */
  class full_dist : public nuclear_dist {

  public:
    
    full_dist() {
      list_size=0;
    }

    virtual ~full_dist() {
      if (list_size>0) delete[] list;
    }
    
    /** \brief Create a distribution including all nuclei with atomic
	numbers less than \c maxA from the mass formula \c nm
    */
    full_dist(nuclear_mass *nm, int maxA=400, bool include_neutron=false);
    
    /** \brief Set the distribution to all nuclei with atomic
	numbers less than \c maxA from the mass formula \c nm

	The information for the previous distribution is cleared 
	before a new distribution is set. 
    */
    int set_dist(nuclear_mass *nm, int maxA=400, bool include_neutron=false);
  
    /// The beginning of the distribution
    virtual iterator begin() {
      return iterator(this,&list[0]);
    };

    /// The end of the distribution
    virtual iterator end() {
      return iterator(this,&list[list_size-1]);
    };

    /// The number of nuclei in the distribution
    virtual size_t size() {
      return list_size;
    };

#ifndef DOXYGENP

  protected:

    /// Specify how to move to the next nucleus
    virtual nucleus *next(nucleus *np) {
      return np+1;
    }

    /// The distribution of nuclei as an array
    nucleus *list;

    /// The size of \ref list array
    size_t list_size;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
