/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IV_SOLVE_H
#define O2SCL_ODE_IV_SOLVE_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/adapt_step.h>
#include <o2scl/gsl_astep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve an initial-value ODE problems given an 
      adaptive ODE stepper
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc, class mat_row_t=omatrix_row> 
    class ode_iv_solve {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// Derivative
    alloc_vec_t dydx;

    /// Memory allocator
    alloc_t ao;

    /// The adaptive stepper
    adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> *astp;
    
    /// Print out iteration information
    virtual int print_iter(double x, size_t nv, vec_t &y) {
      std::cout << type() << " x: " << x << " y: ";
      for(size_t i=0;i<nv;i++) std::cout << y[i] << " ";
      std::cout << std::endl;
      if (verbose>1) {
	char ch;
	std::cin >> ch;
      }
      return 0;
    }
  
#endif

    public:
      
    ode_iv_solve() {
      verbose=0;
      ntrial=1000;
      nsteps_out=10;
      astp=&gsl_astp;
      exit_on_fail=true;
    }
      
    virtual ~ode_iv_solve() {}

    /// Set the adaptive stepper to use
    int set_adapt_step(adapt_step<param_t,func_t,vec_t,
		       alloc_vec_t,alloc_t> &as) {
      astp=&as;
      return 0;
    }

    /** 
	\brief Solve the initial-value problem and output a table 

	Initially, \c xsol should be a vector of size \c nsol, and \c
	ysol should be a two-dimensional array (i.e. omatrix_view) of
	size \c [nsol][n]. On exit, \c nsol will will be the size of
	the solution table, less than or equal to the original value
	of \c nsol.

	If \ref verbose is greater than zero, The solution
	at each internal point will be written to \c std::cout. 
	If \ref verbose is greater than one, a character
	will be required after each point. 

	If the given value of \c h is small enough, the solution
	may generate more points than the space initially allocated
	and the full solution will not be generated. 
	
	\future Consider modifying so that this can handle tables
	which are too small by removing half the rows and doubling
	the stepsize.
      
    */
    template<class mat_t>
    int solve_table(double x0, double x1, double h, size_t n,
		    vec_t &ystart, size_t &nsol, vec_t &xsol, 
		    mat_t &ysol, param_t &pa, func_t &derivs) {

      int ret=0, nmax=nsol, i, first_ret=0;
      size_t j;
      
      xsol[0]=x0;
      for(j=0;j<n;j++) ysol[0][j]=ystart[j];
      if (verbose>0) {
	print_iter(xsol[0],n,ystart);
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
      }
      
      alloc_vec_t yerrl;
      ao.allocate(yerrl,n);
      ao.allocate(dydx,n);
      derivs(x0,n,ystart,dydx,pa);
      
      for(i=1;i<nmax && xsol[i-1]<x1;i++) {
	
	xsol[i]=xsol[i-1];
	for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
	  
	mat_row_t ar(ysol,i);

	ret=astp->astep_derivs(xsol[i],h,x1,n,ar,dydx,yerrl,pa,derivs);
	if (ret!=0) {
	  if (exit_on_fail) {
	    nsol=i+1;
	    add_err_ret
	      ("Adaptive stepper failed in ode_iv_solve::solve_table()",
	       ret);
	  } else if (first_ret!=0) {
	    first_ret=ret;
	  }
	}
    
	if (verbose>0) {
	  print_iter(xsol[i],n,ar);
	  if (verbose>1) {
	    char ch;
	    std::cin >> ch;
	  }
	}
      }

      ao.free(dydx);
      ao.free(yerrl);

      nsol=i;
      
      if (i==nmax && xsol[i-1]<x1) {
	set_err("Ran out of space in ode_iv_solve::solve_table().",
		gsl_etable);
	return gsl_etable;
      }
      return first_ret;
    }

    /** \brief Solve the initial-value problem from \c x0 
	to \c x1 over a grid 

	Initially, \c xsol should be an array of size \c nsol, and
	\c ysol should be a \c matrix of size \c [nsol][n]. This
	function never takes a step larger than the grid size.

	If \ref verbose is greater than zero, The solution
	at each grid point will be written to \c std::cout. 
	If \ref verbose is greater than one, a character
	will be required after each point. 
    */
    template<class mat_t>
    int solve_grid(double x0, double x1, double h, size_t n,
		   vec_t &ystart, size_t nsol, vec_t &xsol, mat_t &ysol,
		   param_t &pa, func_t &derivs) {

      double x=x0, xnext;
      int ret=0, first_ret=0;
      size_t j;

      xsol[0]=x0;
      for(j=0;j<n;j++) ysol[0][j]=ystart[j];
      if (verbose>0) print_iter(xsol[0],n,ystart);

      alloc_vec_t yerrl;
      ao.allocate(dydx,n);
      ao.allocate(yerrl,n);
      
      if (x0<x1) {
	
	derivs(x0,n,ystart,dydx,pa);

	for(size_t i=1;i<nsol && ret==0;i++) {
	    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
	    
	  xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	  mat_row_t ar(ysol,i);
	  while(xsol[i]<xnext && ret==0) {

	    ret=astp->astep_derivs(xsol[i],h,xnext,n,ar,dydx,yerrl,pa,derivs);
	    if (ret!=0) {
	      if (exit_on_fail) {
		add_err_ret
		  ("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		   ret);
	      } else if (first_ret!=0) {
		first_ret=ret;
	      }
	    }
	  }
	  if (verbose>0) print_iter(xsol[i],n,ar);
	    
	}

      } else {

	derivs(x0,n,ystart,dydx,pa);

	for(size_t i=1;i<nsol && ret==0;i++) {
	    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
	    
	  xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	  mat_row_t ar(ysol,i);
	  while(xsol[i]>xnext && ret==0) {
	    ret=astp->astep_derivs(xsol[i],h,xnext,n,ar,dydx,yerrl,pa,derivs);
	    if (ret!=0) {
	      if (exit_on_fail) {
		add_err_ret
		  ("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		   ret);
	      } else if (first_ret!=0) {
		first_ret=ret;
	      }
	    }
	  }
	  if (verbose>0) print_iter(xsol[i],n,ar);

	}

      }

      ao.free(dydx);
      ao.free(yerrl);

      return first_ret;
    }

    /** \brief Solve the initial-value problem from \c x0 
	to \c x1 over a grid storing derivatives

	Initially, \c xsol should be an array of size \c nsol, and
	\c ysol should be a \c omatrix of size \c [nsol][n]. This
	function never takes a step larger than the grid size.

	If \ref verbose is greater than zero, The solution
	at each grid point will be written to \c std::cout. 
	If \ref verbose is greater than one, a character
	will be required after each point. 

	\todo Add error information
    */
    template<class mat_t>
    int solve_grid_derivs(double x0, double x1, double h, size_t n,
			  vec_t &ystart, size_t nsol, vec_t &xsol, 
			  mat_t &ysol, mat_t &dydx_sol,
			  param_t &pa, func_t &derivs) {

      double x=x0, xnext;
      int ret=0, first_ret=0;
      size_t j;
      
      xsol[0]=x0;
      for(j=0;j<n;j++) ysol[0][j]=ystart[j];
      if (verbose>0) print_iter(xsol[0],n,ystart);

      alloc_vec_t yerrl;
      ao.allocate(yerrl,n);

      if (x0<x1) {
	
	mat_row_t mr0(dydx_sol,0);
	derivs(x0,n,ystart,mr0,pa);

	for(size_t i=1;i<nsol && ret==0;i++) {
	  
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) {
	    ysol[i][j]=ysol[i-1][j];
	    dydx_sol[i][j]=dydx_sol[i-1][j];
	  }
	    
	  xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	  mat_row_t ar(ysol,i);
	  mat_row_t ar2(dydx_sol,i);

	  while(xsol[i]<xnext && ret==0) {

	    ret=astp->astep_derivs(xsol[i],h,xnext,n,ar,ar2,yerrl,pa,derivs);
	    if (ret!=0) {
	      if (exit_on_fail) {
      ao.free(yerrl);
		add_err_ret
		  ("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		   ret);
	      } else if (first_ret!=0) {
		first_ret=ret;
	      }
	    }
	  }
	  if (verbose>0) print_iter(xsol[i],n,ar);
	    
	}

      } else {

	mat_row_t mr0(dydx_sol,0);
	derivs(x0,n,ystart,dydx,pa);

	for(size_t i=1;i<nsol && ret==0;i++) {
	    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) {
	    ysol[i][j]=ysol[i-1][j];
	    dydx_sol[i][j]=dydx_sol[i-1][j];
	  }
	    
	  xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	  mat_row_t ar(ysol,i);
	  mat_row_t ar2(dydx_sol,i);

	  while(xsol[i]>xnext && ret==0) {
	    ret=astp->astep_derivs(xsol[i],h,xnext,n,ar,ar2,yerrl,pa,derivs);
	    if (ret!=0) {
	      if (exit_on_fail) {
		ao.free(yerrl);
		add_err_ret
		  ("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		   ret);
	      } else if (first_ret!=0) {
		first_ret=ret;
	      }
	    }
	  }
	  if (verbose>0) print_iter(xsol[i],n,ar);

	}

      }

      ao.free(yerrl);

      return first_ret;
    }

    /** \brief Solve the initial-value problem 
	to get the final value 

	If \ref verbose is greater than zero, The solution at less
	than or approximately equal to \ref nsteps_out points will be
	written to \c std::cout. If \ref verbose is greater than one,
	a character will be required after each selected point.

	The solution fails if more than \ref ntrial steps are
	required.
    */
    int solve_final_value(double x0, double x1, double h, size_t n,
			  vec_t &ystart, vec_t &yend, 
			  param_t &pa, func_t &derivs) {

      ao.allocate(dydx,n);
      int ret=solve_final_value_derivs(x0,x1,h,n,ystart,yend,dydx,
				      dydx,pa,derivs);
      ao.free(dydx);
      return ret;
    }
      
    /** \brief Solve the initial-value problem 
	to get the final value and derivative

	If \ref verbose is greater than zero, The solution at less
	than or approximately equal to \ref nsteps_out points will be
	written to \c std::cout. If \ref verbose is greater than one,
	a character will be required after each selected point.

	The solution fails if more than \ref ntrial steps are
	required.

	\todo Add error information
    */
    int solve_final_value_derivs(double x0, double x1, double h, size_t n,
				 vec_t &ystart, vec_t &yend, 
				 vec_t &dydx_start, vec_t &dydx_end,
				 param_t &pa, func_t &derivs) {
      
      if ((x1>x0 && h<=0.0) || (x0>x1 && h>=0.0)) {
	set_err_ret("Interval ordering and sign of h don't match.",
		    gsl_einval);
      }

      double x=x0, xverb=0.0, dxverb=0.0;
      int ret=0, it=0, first_ret=0;
      if (verbose>0) {
	print_iter(x0,n,ystart);
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
	if (x1>x0) {
	  dxverb=(x1-x0)/((double)nsteps_out);
	  xverb=x0+dxverb;
	} else {
	  dxverb=(x0-x1)/((double)nsteps_out);
	  xverb=x0-dxverb;
	}
      }
	
      for(size_t i=0;i<n;i++) yend[i]=ystart[i];

      alloc_vec_t yerrl;
      ao.allocate(yerrl,n);
	
      if (x1>x0) {
	
	derivs(x,n,yend,dydx_start,pa);
	for(size_t i=0;i<n;i++) dydx_end[i]=dydx_start[i];
	
	while(x<x1 && ret==0) {
	  ret=astp->astep_derivs(x,h,x1,n,yend,dydx_end,yerrl,pa,derivs);
	  if (ret!=0) {
	    if (exit_on_fail) {
      ao.free(yerrl);
	      add_err_ret
		("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		 ret);
	    } else if (first_ret!=0) {
	      first_ret=ret;
	    }
	  }
	  if (verbose>0 && x>xverb) {
	    print_iter(x,n,yend);
	    if (verbose>1) {
	      char ch;
	      std::cin >> ch;
	    }
	    xverb+=dxverb;
	  }
	  it++;
	  if (it>ntrial) {
      ao.free(yerrl);
	    std::string s="Exceeded max number of iterations in ";
	    s+="ode_iv_solve::solve_final_value().";
	    set_err_ret(s.c_str(),gsl_emaxiter);
	  }
	}

      } else {

	derivs(x,n,yend,dydx_start,pa);
	for(size_t i=0;i<n;i++) dydx_end[i]=dydx_start[i];
	
	while(x>x1 && ret==0) {
	  ret=astp->astep_derivs(x,h,x1,n,yend,dydx_end,yerrl,pa,derivs);
	  if (ret!=0) {
	    if (exit_on_fail) {
      ao.free(yerrl);
	      add_err_ret
		("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		 ret);
	    } else if (first_ret!=0) {
	      first_ret=ret;
	    }
	  }
	  if (verbose>0 && x<xverb) {
	    print_iter(x,n,yend);
	    if (verbose>1) {
	      char ch;
	      std::cin >> ch;
	    }
	    xverb-=dxverb;
	  }
	  it++;
	  if (it>ntrial) {
      ao.free(yerrl);
	    std::string s="Exceeded max number of iterations in ";
	    s+="ode_iv_solve::solve_final_value().";
	    set_err_ret(s.c_str(),gsl_emaxiter);
	  }
	}
      }
      
      if (verbose>0) {
	print_iter(x,n,yend);
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
      }
	
      ao.free(yerrl);
      return first_ret;
    }
      
    /// Set output level
    int verbose;
      
    /** 
	\brief Number of output points if \ref verbose is greater 
	than zero (default 10)
    */
    size_t nsteps_out;

    /// Maximum number of steps for solve_final_value() (default 1000)
    int ntrial;

    /** 
	\brief If true, stop the solution if the adaptive stepper fails
	
	If this is false, then failures in the adaptive stepper are
	ignored.
     */
    bool exit_on_fail;

    /// The default adaptive stepper
    gsl_astep<param_t,func_t,vec_t,alloc_vec_t,alloc_t> gsl_astp;

    /// Return the type, \c "ode_iv_solve".
    virtual const char *type() { return "ode_iv_solve"; }
      
  };

#ifdef NEVER_DEFINED

  /** 
      \brief Solve an initial-value ODE problems given an 
      adaptive ODE stepper
  */
  template<class param_t,func_t=ode_funct<param_t> >
    class ode_iv_solve_x : public ode_iv_solve<param_t,func_t,
    ovector_view,ovector,ovector_alloc,omatrix_row> {
    
    public:
    
    int solve_x_derivs(double x0, double x1, double h, size_t n,
		       ovector_view &ystart, size_t &nsol, 
		       ovector &xsol, std::vector<ovector> &ysol,
		       std::vector<ovector> &dydx_sol,
		       param_t &pa, func_t &derivs) {
      
      // Temporary ovectors
      ovector y, dy;

      // Proceed with solution
      double x=x0, xnext;
      int ret=0, first_ret=0;
      size_t j;
      
      xsol.push_back(x);
      for(j=0;j<n;j++) y[j]=ystart[j];
      ysol.push_back(y);

      alloc_vec_t yerrl;
      ao.allocate(yerrl,n);
      
      if (x0<x1) {
	
	derivs(x,n,y,dy,pa);
	dydx_sol.push_back(dy);

	while (x<x1 && ret==0) {

	  ret=astp->astep_derivs(x,h,x1,n,y,dy,yerrl,pa,derivs);
	  if (ret!=0) {
	    if (exit_on_fail) {
      ao.free(yerrl);
	      add_err_ret
		("Adaptive stepper failed in ode_iv_solve::solve_grid()",
		 ret);
	    } else if (first_ret!=0) {
	      first_ret=ret;
	    }
	  }
	  if (verbose>0) print_iter(xsol[i],n,ar);

	  xsol.push_back(x);
	  ysol.push_back(y);
	  ysol.push_back(dy);

	    
	}

      }

      ao.free(yerrl);

      return first_ret;
    }

#endif


#ifndef DOXYGENP
}
#endif

#endif
