/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/timer.h>

using namespace std;
using namespace o2scl;

/// Return the number of seconds between set() and reset()
double timer_gettod::seconds_elapsed() {
  double ret=0.0;
  long diff=mark.tv_sec-zero.tv_sec;
  int udiff=mark.tv_usec-zero.tv_usec;
  if (udiff<0.0) {
    diff--;
    udiff+=1000000;
  }
  ret=((double)diff)+((double)udiff)/1.0e6;
  return ret;
}

/// Return the time between set() and reset()
int timer_gettod::time_elapsed(int &d, int &h, int &m, int &s, int &usec) {
  long diff=mark.tv_sec-zero.tv_sec;
  int udiff=mark.tv_usec-zero.tv_usec;
  if (udiff<0.0) {
    diff--;
    udiff+=1000000;
  }

  d=((int)(diff/86400));
  h=((int)((diff-d*86400)/3600));
  m=((int)((diff-d*86400-h*3600)/60));
  s=((int)(diff-d*86400-h*3600-m*60));
  usec=udiff;
    
  return 0;
}

/// Time remaining if 'n' out of 'tot' tasks have been completed
int timer_gettod::time_remaining(int n, int ntot, 
				  int &d, int &h, int &m, int &s, int &usec) {
  double diff=((double)(mark.tv_sec-zero.tv_sec));
  double udiff=((double)(mark.tv_usec-zero.tv_usec));
      
  // Fix if udiff < 0
  if (udiff<0.0) {
    diff-=1.0;
    udiff+=1.0e6;
  }

  // Do the multiplication
  diff*=((double)(ntot-n))/((double)n);
  udiff*=((double)(ntot-n))/((double)n);
      
  // Convert to an integral number of seconds
  int idiff=((int)diff);
  udiff+=1.0e6*(diff-((double)idiff));
      
  // Fix if udiff is more than 1 second
  while (udiff>1.0e6) {
    idiff++;
    udiff-=1.0e6;
  }

  // Separate into parts
  d=((int)(idiff/86400));
  h=((int)((idiff-d*86400)/3600));
  m=((int)((idiff-d*86400-h*3600)/60));
  s=((int)(idiff-d*86400-h*3600-m*60));
  usec=((int)udiff);
  return 0;
}

/// Time remaining if 'n' out of 'tot' tasks have been completed
std::string timer_gettod::time_remaining(int n, int ntot) { 
  int d, h, m, s, usec;
  time_remaining(n,ntot,d,h,m,s,usec);
  return interval_to_string(d,h,m,s,usec);
}

/// Convert a time interval to a string
std::string timer_gettod::interval_to_string(int d, int h, int m, 
					      int s, int usec) {
  std::string st;
  if (usec==0) {
    st=o2scl::itos(d)+" days, "+
      o2scl::itos(h)+" hours, "+
      o2scl::itos(m)+" minutes, and "+
      o2scl::itos(s)+" seconds";
  } else {
    double f=((double)s)+((double)usec)/1.0e6;
    if (d>0) {
      st+=o2scl::itos(d)+" days, ";
    }
    if (h>0 || d>0) {
      st+=o2scl::itos(h)+" hours, ";
      st+=o2scl::itos(m)+" minutes, and ";
    } else if (m>0) {
      st+=o2scl::itos(m)+" minutes and ";
    }
    st+=o2scl::dtos(f,0)+" seconds";
  }
  return st;
}

timer_clock::timer_clock() {
  time=clock();
}

double timer_clock::time_since() {
  return ((double)clock())/CLOCKS_PER_SEC;
}

void timer_clock::time_since(int &d, int &h, int &m, int &s, double &f) {
  
  double temp=((double)(clock()-time))/CLOCKS_PER_SEC;
  d=((int)temp)/86400;
  h=((int)(temp-d*86400.0))/3600;
  m=((int)(temp-d*86400.0-h*3600.0))/60;
  s=((int)(temp-d*86400.0-h*3600.0-m*60.0));
  f=temp-d*86400.0-h*3600.0-m*60.0-s;
  return;
}

void timer_clock::time_remaining(int n, int tot, int &d, int &h, 
				 int &m, int &s, double &f) {
  
  double temp=((double)(clock()-time))/CLOCKS_PER_SEC*
    ((double)(tot-n))/((double)n);
  d=((int)temp)/86400;
  h=((int)(temp-d*86400.0))/3600;
  m=((int)(temp-d*86400.0-h*3600.0))/60;
  s=((int)(temp-d*86400.0-h*3600.0-m*60.0));
  f=temp-d*86400.0-h*3600.0-m*60.0-s;
  return;
}

string timer_clock::interval_to_string(int d, int h, int m, int s, double f) {
  string st;
  if (f==0.0) {
    st= itos(d)+" days, "+
      itos(h)+" hours, "+
      itos(m)+" minutes, and "+
      itos(s)+" seconds";
  } else {
    f+=s;
    st= itos(d)+" days, "+
      itos(h)+" hours, "+
      itos(m)+" minutes, and "+
      dtos(f,0)+" seconds";
  }
  return st;
}

