/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_VEC_STATS_H
#define O2SCL_VEC_STATS_H

/** \file vec_stats.h
    \brief File containing statistics template functions
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <gsl/gsl_ieee_utils.h>
#include <gsl/gsl_sort.h>

#ifndef DOXYGENP
namespace o2scl
{
#endif

  /** 
      \brief Compute the mean of the first \c n elements of a vector

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
  */
  template<class vec_t>
    double vector_mean(const size_t n, vec_t &data) {
    
    long double mean=0;
    for(size_t i=0;i<n;i++) {
      mean+=(data[i]-mean)/(i+1);
    }
    return mean;
  }

  /// Variance
  template<class vec_t>
    double vector_variance_fmean(const size_t n, vec_t &data, 
				 double mean) {
    long double var=0;
    for(size_t i=0;i<n;i++) {
      long double delta=(data[i]-mean);
      var+=(delta*delta-var)/(i+1);
    }
    return var;
  }

  /// Standard deviation
  template<class vec_t>
    double vector_stddev_fmean(const size_t n, vec_t &data, 
			       double mean) {
    double sd=vector_variance_fmean<vec_t>(n,data,mean);
    return sqrt(sd);
  }

  /** \brief Compute the variance of the first \c n elements of a vector
      given the mean \c mean.

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
  */
  template<class vec_t>
    double vector_variance(const size_t n, vec_t &data, double mean) {
    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return var*n/(n-1);
  }

  /// Variance
  template<class vec_t>
    double vector_variance(const size_t n, vec_t &data) {
    
    double mean=vector_mean<vec_t>(n,data);
    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return var*n/(n-1);
  }


  /// Standard deviation
  template<class vec_t>
    double vector_stddev(const size_t n, vec_t &data) {
    
    double mean=vector_mean<vec_t>(n,data);
    double var=vector_variance_fmean<vec_t>(n,data,mean);
    return sqrt(var*n/(n-1));
  }

  /// Standard deviation
  template<class vec_t>
    double vector_stddev(const size_t n, vec_t &data, 
			 double mean) {
    double sd=vector_variance_fmean<vec_t>(n,data,mean);
    return sqrt(sd*n/(n-1));
  }
  
  /// Absolute deviation from the mean
  template<class vec_t>
    double vector_absdev(const size_t n, vec_t &data, 
			 double mean) {
    long double sum=0;
    for(size_t i=0;i<n;i++) {
      sum+=fabs(data[i]-mean);
    }
    return sum/n;
  }
  
  /// Absolute deviation from the mean
  template<class vec_t>
    double vector_absdev(const size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    return vector_absdev(n,data,mean);
  }

  /// Skewness
  template<class vec_t>
    double vector_skew(const size_t n, vec_t &data, double mean,
		       double stddev) {
    long double skew=0;
    for(size_t i=0;i<n;i++) {
      long double x=(data[i]-mean)/stddev;
      skew+=(x*x*x-skew)/(i+1);
    }
    return skew;
  }

  /// Skewness
  template<class vec_t>
    double vector_skew(const size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    double sd=vector_stddev<vec_t>(n,data,mean);
    return vector_skew(n,data,mean,sd);
  }

  /// Kurtosis
  template<class vec_t>
    double vector_kurtosis(const size_t n, vec_t &data, double mean,
			   double stddev) {
    long double avg=0;
    for(size_t i=0;i<n;i++) {
      long double x=(data[i]-mean)/stddev;
      avg+=(x*x*x*x-avg)/(i+1);
    }
    return avg-3.0;
  }

  /// Kurtosis
  template<class vec_t>
    double vector_kurtosis(const size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    double sd=vector_stddev<vec_t>(n,data,mean);
    return vector_kurtosis(n,data,mean,sd);
  }

  /// Lag1 autocorrelation
  template<class vec_t>
    double vector_lag1_autocorr(const size_t n, vec_t &data, double mean) {
    long double q=0;
    long double v=(data[0]-mean)*(data[0]-mean);
    for(size_t i=1;i<n;i++) {
      long double delta0=data[i-1]-mean;
      long double delta1=data[i]-mean;
      q+=(delta0*delta1-q)/(i+1);
      v+=(delta1*delta1-v)/(i+1);
    }
    return q/v;
  }

  /// Lag1 autocorrelation
  template<class vec_t>
    double vector_lag1_autocorr(const size_t n, vec_t &data) {
    double mean=vector_mean<vec_t>(n,data);
    return vector_lag1_autocorr(n,data,mean);
  }

  /// Covariance
  template<class vec_t>
    double vector_covariance(const size_t n, vec_t &data1, vec_t &data2,
			     double mean1, double mean2) {
    double covar=0;
    for(size_t i=0;i<n;i++) {
      double delta1=(data1[i]-mean1);
      double delta2=(data2[i]-mean2);
      covar+=(delta1*delta2-covar)/(i+1);
    }
    return covar;
  }

  /// Covariance
  template<class vec_t>
    double vector_covariance(const size_t n, vec_t &data1, vec_t &data2) {
    double covar=0;
    double mean1=vector_mean<vec_t>(n,data1);
    double mean2=vector_mean<vec_t>(n,data2);
    for(size_t i=0;i<n;i++) {
      long double delta1=(data1[i]-mean1);
      long double delta2=(data2[i]-mean2);
      covar+=(delta1*delta2-covar)/(i+1);
    }
    return covar;
  }
  
  /// Pearson's correlation
  template<class vec_t>
    double vector_correlation(const size_t n, vec_t &data1, vec_t &data2) {
    size_t i;

    double sum_xsq = 0.0;
    double sum_ysq = 0.0;
    double sum_cross = 0.0;
    double ratio;
    double delta_x, delta_y;
    double mean_x, mean_y;
    double r;

    /*
     * Compute:
     * sum_xsq = Sum [ (x_i - mu_x)^2 ],
     * sum_ysq = Sum [ (y_i - mu_y)^2 ] and
     * sum_cross = Sum [ (x_i - mu_x) * (y_i - mu_y) ]
     * using the above relation from Welford's paper
     */

    mean_x = data1[0];
    mean_y = data2[0];

    for (i = 1; i < n; ++i) {
      ratio = i / (i + 1.0);
      delta_x = data1[i] - mean_x;
      delta_y = data2[i] - mean_y;
      sum_xsq += delta_x * delta_x * ratio;
      sum_ysq += delta_y * delta_y * ratio;
      sum_cross += delta_x * delta_y * ratio;
      mean_x += delta_x / (i + 1.0);
      mean_y += delta_y / (i + 1.0);
    }
    
    r = sum_cross / (sqrt(sum_xsq) * sqrt(sum_ysq));
    
    return r;
  }

  /// Pooled variance
  template<class vec_t>
    double vector_pvariance(const size_t n1, vec_t &data1, 
			    const size_t n2, vec_t &data2) {
    double var1=vector_variance<vec_t>(n1,data1);
    double var2=vector_variance<vec_t>(n2,data2);
    return (((n1-1)*var1)+((n2-1)*var2))/(n1+n2-2);
  }

  /// Quantile
  template<class vec_t>
    double vector_quantile_sorted(const size_t n, vec_t &data, 
				  const double f) {

    double index=f*(n-1);
    size_t lhs=((size_t)index);
    double delta=index-lhs;
    if (n==0) return 0;
    if (lhs==n-1) return data[lhs];
    return (1-delta)*data[lhs]+delta*data[lhs+1];
  }

  /// Quantile
  template<class vec_t>
    double vector_median_sorted(const size_t n, vec_t &data) {

    if (n==0) return 0;

    size_t lhs=(n-1)/2;
    size_t rhs=n/2;

    if (lhs==rhs) return data[lhs];

    return (data[lhs]+data[rhs])/2.0;
  }

#ifndef DOXYGENP
}
#endif

#endif
