/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/eff_boson.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// eff_boson class

// Static variables
bool eff_boson::loaded=false;
double **eff_boson::Pmnb=NULL;
double eff_boson::stat_temper=0.0, eff_boson::fix_density=0.0;
double eff_boson::parma=0.4;
int eff_boson::sizem=4;
int eff_boson::sizen=4;

// Constructor, Destructor
eff_boson::eff_boson(double mass, double dof) : 
  part(mass,dof), boson(mass,dof) {
  co=0.0;
  name="eff_boson";
  density_mroot=&def_density_mroot;
  meth2_root=&def_meth2_root;
  psi_root=&def_psi_root;
  
  loadcoeff("boselat3");
}

eff_boson::~eff_boson() {}

// Load coefficients for finite-temperature approximations
int eff_boson::loadcoeff(string bfile) {
  
  collection co;
  
  double *dtemp;
  
  bfile=lib_settings.get_data_dir()+bfile;
  text_in_file tif(bfile);

  double_io_type double_io;
  string_io_type string_io;
  int_io_type int_io;

  if (co.fin(&tif)!=0) {
    set_err((((string)"File ")+bfile+" not found in loadcoeff().").c_str(),1);
    return 1;
  }
  parma=double_io.getd(co,"a");
  if (double_io.get(co,"Pmn",Pmnb,sizem,sizen)!=0) {
    set_err((((string)"Table 'Pmn' not found in file ")+
	    bfile+" in loadcoeff().").c_str(),3);
    return 3;
  }
  // We set sizem and sizen to coincide with the 
  // notation in the paper
  sizem--;
  sizen--;
  // Make sure that the space created for Pmnb
  // is not destroyed when the function exits
  if (co.disown("Pmn")!=0) {
    set_err("Failed to disown 'Pmn' in loadcoeff.",4);
    return 4;
  }

  loaded=true;

  return gsl_success;
}

int eff_boson::calc_mu(const double temper) {
  int nn, mm, nvar=1, ret=0;
  double xx[2], pren, preu, prep, sumn, sumu, sump;
  double gg, opg, nc, h, oph, psi;

  if (non_interacting) { nu=mu; ms=m; }

  // Massless eff_bosons
  if (ms==0.0) {
    massless_calc_mu(temper);
    return gsl_success;
  }
  if (loaded==false) {
    set_err_ret("Coefficients not set in nepn_mroot().",gsl_einval);
  }

  psi=(nu-ms)/temper;
  if (psi>=0.0) {
    h=0.0;
  } else {
    if (psi>-0.05) {
      xx[0]=sqrt(-2.0*parma*psi);
    } else if (psi>-1.0) {
      xx[0]=-sqrt(parma)*(-3.0+4.0*psi+4.0*log(2.0));
    } else {
      xx[0]=sqrt(parma)*exp(1.0-psi);
    }

    funct_mfptr<eff_boson,void *> mfs(this,&eff_boson::solve_fun);

    void *vpx=(void *)&psi;
    int psi_root_err=psi_root->solve(xx[0],vpx,mfs);
    if (psi_root_err!=0) {
      add_err("psi_root failed in nepn_mroot().",psi_root_err);
      // We continue execution anyway
    }
    h=xx[0];
  }

  oph=1.0+h;
  gg=temper/ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(ms,3.0);

  preu=pow(gg,2.5)/pow(oph,(double)sizem+1)/pow(opg,sizen-1.5);
  prep=preu;
  pren=pow(sqrt(parma)+h,2.0)*
    pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
    
  sumn=0.0;
  sumu=0.0;
  sump=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      if (h!=0.0 || mm>=2) {
	sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	  (h*(sizem+1.0-mm)-mm);
      }
      sumu+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn)*
	(1.5+nn+gg/opg*(1.5-sizen));
      sump+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn);
    }
  }
  if (h==0.0) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[1][nn]*pow(gg,(double)nn)*sizem;
    }
  }
    
  n=g/2.0*pren*sumn*nc;
  pr=g/2.0*prep*sump*nc*ms;
  ed=g/2.0*preu*sumu*nc*ms+n*ms;
  en=(ed+pr-mu*n)/temper;
  
  return ret;
}

int eff_boson::solve_fun(double x, double &y, void *&pa) {
  double psi, h, sqt;

  psi=*((double *)pa);
  h=x;
  sqt=sqrt(parma)+h;
  y=(h/sqt-log(sqt/sqrt(parma)))/psi-1.0;

  return gsl_success;
}

int eff_boson::calc_density(const double temper) {
  double h, oph, gg, opg, nc, preu, prep, pren;
  double sumn, sumu, sump, sqt, psi;
  int ret, mm, nn;
  ovector xx(2);

  if (non_interacting) { ms=m; nu=mu; }

  fix_density=n;
  stat_temper=temper;

  psi=(nu-ms)/temper;
  
  // If psi is too small, then we won't be able to solve for the 
  // right density, so we try to repair this here. It is possible
  // that -20 isn't large enough. This should be examined.
  if (psi<-20.0) psi=-20.0;

  mm_funct_mfptr<eff_boson,void *> mfd(this,&eff_boson::density_fun);

  if (psi>=0.0) {
    h=0.0;
  } else {
    if (psi>-0.05) {
      xx[0]=sqrt(-2.0*parma*psi);
    } else if (psi>-1.0) {
      xx[0]=-sqrt(parma)*(-3.0+4.0*psi+4.0*log(2.0));
    } else {
      xx[0]=sqrt(parma)*exp(1.0-psi);
    }
    void *vpx=(void *)this;
    ret=density_mroot->msolve(1,xx,vpx,mfd);
    if (ret!=0) {
      ovector yy(2);
      density_fun(1,xx,yy,vpx);
      cout << xx[0] << " " << yy[0] << " " << fix_density << endl;
      exit(-1);
      ret+=30;
      return ret;
    }
    h=xx[0];
  }
  if (!finite(h)) {
    set_err("h not finite in calc_density_mroot().",9);
    return 9;
  }
  sqt=sqrt(parma)+h;
  nu=(h/sqt-log(sqt/sqrt(parma)))*temper+ms;

  if (non_interacting) { mu=nu; }

  oph=1.0+h;
  gg=temper/ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(ms,3.0);

  preu=pow(gg,2.5)/pow(oph,(double)sizem+1)/pow(opg,sizen-1.5);
  prep=preu;
  pren=pow(sqrt(parma)+h,2.0)*
    pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
    
  sumn=0.0;
  sumu=0.0;
  sump=0.0;

  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      if (h!=0.0 || mm>=2) {
	sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	  (h*(sizem+1.0-mm)-mm);
      }
      sumu+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn)*
	(1.5+nn+gg/opg*(1.5-sizen));
      sump+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn);
    }
  }

  if (h==0.0) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[1][nn]*pow(gg,(double)nn)*sizem;
    }
  }

  n=g/2.0*pren*sumn*nc;
  pr=g/2.0*prep*sump*nc*ms;
  ed=g/2.0*preu*sumu*nc*ms+n*ms;
  en=(ed+pr-mu*n)/temper;

  return gsl_success;

}

int eff_boson::density_fun(size_t nv, const ovector_view &x, 
			   ovector_view &y, void *&pa) {
  double h,gg,opg,nc,oph,sumn,pren;
  int mm, nn;
  eff_boson *partx=(eff_boson *)pa;

  h=x[0];
  cout << h << endl;

  oph=1.0+h;
  gg=stat_temper/partx->ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(partx->ms,3.0);

  pren=pow(sqrt(parma)+h,2.0)*
    pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
  sumn=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	(h*(sizem+1.0-mm)-mm);
    }
  }

  y[0]=partx->g/2.0*pren*sumn*nc/fix_density-1.0;
  cout << h << " " << partx->g << " " << y[0] << endl;
  cout << pren << " " << sumn << " " << nc << " " << fix_density << endl;
  if (!finite(y[0])) exit(-1);

  return gsl_success;
}

int eff_boson::pair_mu(const double temper) {
  int ret;

  if (non_interacting) { nu=mu; ms=m; }
  ret=calc_mu(temper);

  if (ret==0) {
    eff_boson antip(ms,g);
    anti(antip);

    antip.density_mroot=density_mroot;
    antip.psi_root=psi_root;
    antip.meth2_root=meth2_root;

    if (non_interacting) { antip.nu=antip.mu; antip.ms=antip.m; }
    ret=antip.calc_mu(temper);
    n-=antip.n;
    pr+=antip.pr;
    ed+=antip.ed;
    en+=antip.en;
  }

  return ret;
}

int eff_boson::pair_density(const double temper) {
  double oph, opg, gg, h, pren, preu, prep, sumn;
  double sumu, sump, nc, sqt, psi;
  int mm, nn;
  ovector xx(2);

  if (non_interacting) { ms=m; nu=mu; }

  if (ms==0.0) return massless_calc_mu(temper);

  fix_density=n;
  stat_temper=temper;

  psi=(nu-ms)/temper;
  if (psi>-0.05) {
    xx[0]=sqrt(-2.0*parma*psi);
  } else if (psi>-1.0) {
    xx[0]=-sqrt(parma)*(-3.0+4.0*psi+4.0*log(2.0));
  } else {
    xx[0]=sqrt(parma)*exp(1.0-psi);
  }
  psi=(-nu-ms)/temper;

  mm_funct_mfptr<eff_boson,void *> mfd(this,&eff_boson::pair_density_fun);

  if (psi>-0.05) {
    xx[1]=sqrt(-2.0*parma*psi);
  } else if (psi>-1.0) {
    xx[1]=-sqrt(parma)*(-3.0+4.0*psi+4.0*log(2.0));
  } else {
    xx[1]=sqrt(parma)*exp(1.0-psi);
  }
  int density_mroot_err=0;
  //int density_mroot_err=density_mroot->msolve(2,xx,this,mfd);
  if (density_mroot_err!=0) {
    add_err("mroot failed in pair_density_mroot().",density_mroot_err);
    // We continue execution anyway
  }
  h=xx[1];
  sqt=sqrt(parma)+h;
  nu=(h/sqt-log(sqt/sqrt(parma)))*temper+ms;

  if (non_interacting) { mu=nu; }

  oph=1.0+h;
  gg=temper/ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(ms,3.0);

  preu=pow(gg,2.5)/pow(oph,(double)sizem+1)/pow(opg,sizen-1.5);
  prep=preu;
  pren=pow(sqrt(parma)+h,2.0)*
    pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
    
  sumn=0.0;
  sumu=0.0;
  sump=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	(h*(sizem+1.0-mm)-mm);
      sumu+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn)*
	(1.5+nn+gg/opg*(1.5-sizen));
      sump+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn);
    }
  }
  n=-g/2.0*pren*sumn*nc;
  pr=g/2.0*prep*sump*nc*ms;
  ed=g/2.0*preu*sumu*nc*ms+n*ms;

  h=xx[0];
  sqt=sqrt(parma)+h;
  nu=(h/sqt-log(sqt/sqrt(parma)))*temper+ms;

  if (non_interacting) { mu=nu; }

  oph=1.0+h;
  gg=temper/ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(ms,3.0);

  preu=pow(gg,2.5)/pow(oph,(double)sizem+1)/pow(opg,sizen-1.5);
  prep=preu;
  pren=pow(sqrt(parma)+h,2.0)*
    pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
    
  sumn=0.0;
  sumu=0.0;
  sump=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	(h*(sizem+1.0-mm)-mm);
      sumu+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn)*
	(1.5+nn+gg/opg*(1.5-sizen));
      sump+=Pmnb[mm][nn]*pow(h,(double)mm)*pow(gg,(double)nn);
    }
  }
  n+=g/2.0*pren*sumn*nc;
  pr+=g/2.0*prep*sump*nc*ms;
  ed+=g/2.0*preu*sumu*nc*ms+n*ms;
  en=(ed+pr-mu*n)/temper;

  return gsl_success;
}

int eff_boson::pair_density_fun(size_t nv, const ovector_view &x, 
				ovector_view &y, void *&pa) {
  double h,gg,opg,nc,oph,sumn,pren,sqt,psi;
  int mm, nn;

  eff_boson *partx=(eff_boson *)pa;

  if (x[0]<0.0 || x[1]<0.0) return 1;

  // Solve for particles:
  h=x[0];
  sqt=sqrt(parma)+h;
  if (sqt<0.0) return 1;
  psi=h/sqt-log(sqt/sqrt(parma));
  y[0]=psi;

  oph=1.0+h;
  gg=stat_temper/partx->ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(partx->ms,3.0);

  pren=sqt*sqt*pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
  sumn=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	(h*(sizem+1.0-mm)-mm);
    }
  }

  y[1]=partx->g/2.0*pren*sumn*nc;
  
  // Solve for anti-particles:
  h=x[1];
  sqt=sqrt(parma)+h;
  if (sqt<0.0) return 2;
  psi=h/sqt-log(sqt/sqrt(parma));
  y[0]+=psi;

  oph=1.0+h;
  gg=stat_temper/partx->ms;
  opg=1.0+gg;
  nc=1.0/pi2*pow(partx->ms,3.0);

  pren=sqt*sqt*pow(gg,1.5)/pow(oph,sizem+2.0)/pow(opg,sizen-1.5);
  sumn=0.0;
  for(mm=0;mm<=sizem;mm++) {
    for(nn=0;nn<=sizen;nn++) {
      sumn+=Pmnb[mm][nn]*pow(h,(double)mm-2.0)*pow(gg,(double)nn)*
	(h*(sizem+1.0-mm)-mm);
    }
  }
  
  y[1]-=partx->g/2.0*pren*sumn*nc;

  y[0]=y[0]/(2.0*partx->m/stat_temper)-1.0;
  y[1]=y[1]/fix_density-1.0;
  //  if (!finite(y[1]) || !finite(y[2])) return 1;

  return gsl_success;
}

template<> int io_tlate<eff_boson>::input
(cinput *co, in_file_format *ins, eff_boson *b) {
  std::string t;
  co->object_in("boson",ins,b,t);
  return 0;
}

template<> int io_tlate<eff_boson>::output
(coutput *co, out_file_format *outs, eff_boson *b) {
  std::string t;
  co->object_out("boson",outs,b,0,0,t);
  return 0;
}

template<> const char *io_tlate<eff_boson>::type() { return "eff_boson"; }

