/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_EFF_FERMION_H
#define O2SCL_EFF_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/mm_funct.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/misc.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Fermion class from fitting method
      
      Based on the fitting method of \ref Johns96 which is an update
      of the method from \ref Eggleton73 . This method is 
      approximate, but very fast. For a more accurate (but slower)
      method, use rel_fermion. 
      
      Given the chemical potential and the temperature the functions
      calc_mu() and pair_mu() work by solving the equation
      \f[
      \psi= 2 \sqrt{1+f/a}+\log\left(\frac{\sqrt{1+f/a}-1}
      {\sqrt{1+f/a}+1}\right)
      \f]
      for \f$ f \f$ given \f$ \psi=(\mu-m)/T \f$.
      If \f$ f/a<10^{-10} \f$, then the alternative expression
      \f[
      \psi= 2 \left(1+f/(2 a)\right)+\log\left[\frac{f/(2 a)}
      {\left(1+f/(2 a)\right)}\right]
      \f]
      is used. The pressure, energy density, and entropy, are
      determined as polynomials in \f$ f \f$ with a
      set of precomputed coefficients as done in \ref Johns96 .

      If \f$ \psi \f$ is too small (less than about -200), the 
      above procedure fails. To handle this, this class
      uses the classical result if \f$ \psi \f$ < \ref min_psi,
      where \ref min_psi defaults to -200.

      When the density and temperature is given instead
      (calc_density() and pair_density()), then there are two ways to
      proceed.
      - We can use the density to solve for \f$ f \f$
      - We can use the density to solve for the chemical potential

      Because the density is a complicated polynomial in \f$ f \f$,
      the former procedure does not work very well even though it might
      be less time consuming. The density is solved for the effective
      chemical potential instead. The initial guess is just taken from
      the present value of part::nu.

      The constructor uses the coefficients from the file \c fermilat3
      by default.

      It is important that if the \ref inte and \ref root objects are
      not thread safe, different instances should not use the same
      instance of a \ref inte or \ref root object simultaneously.

      \future Use bracketing to speed up one-dimensional root finding.
  */
  class eff_fermion : virtual public fermion {

  public:
    
    /// Create a fermion with mass \c mass and degeneracy \c dof 
    eff_fermion(double mass=0.0, double dof=0.0);
    
    virtual ~eff_fermion();
    
    /** \name Load coefficients for finite-temperature approximation

	\c ctype Should be one of the constants below:
	cf_fermilat3, cf_fermijel2, cf_fermijel3, or
	cf_fermijel3cons. 
	
	\todo Document these coefficients better
    */
    //@{
    static int load_coefficients(int ix);
    static const int cf_fermilat3=1;
    static const int cf_fermijel2=2;
    static const int cf_fermijel3=3;
    static const int cf_fermijel3cons=4;
    //@}

    /** 
	\brief Calculate thermodynamic 
	properties as function of chemical potential
	
	If the quantity \f$ (\mu-m)/T \f$ (or \f$ (\nu-m^{*})/T \f$ in
	the case of interacting particles) is less than -200, then
	this quietly sets the density, the scalar density, the energy
	density, the pressure and the entropy to zero and exits.

	\todo Should see if the function actually works if 
	\f$ (\mu-m)/T = -199 \f$ .
     */
    virtual int calc_mu(const double temper);

    /** 
	\brief Calculate thermodynamic 
	properties as function of density

	\warning This function needs a guess for the chemical potential,
	and will fail if that guess is not sufficiently accurate.
     */
    virtual int calc_density(const double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of density
     */
    virtual int pair_density(const double temper);

    /** \brief Set the solver for use in calculating \f$ \psi \f$ 
     */
    int set_psi_root(root<void *, funct<void *> > &rp) {
      psi_root=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density with meth2=true
    */
    int set_density_root(root<void *, funct<void *> > &rp) {
      density_root=&rp;
      return 0;
    }
    
    /** \brief If the temperature is less than \c tlimit then the
	zero-temperature functions are used (default 0).
    */
    double tlimit;
    
    /** \brief The default solver for \f$ \psi \f$
     */
    cern_mroot_root<void *, funct<void *> > def_psi_root;
    
    /** \brief The default solver for calc_density() and pair_density()
     */
    cern_mroot_root<void *, funct<void *> > def_density_root;
    
    /// Return string denoting type ("eff_fermion")
    virtual const char *type() { return "eff_fermion"; }

    /// The minimum value of \f$ \psi \f$ (default -200)
    double min_psi;

#ifndef DOXYGEN_INTERNAL

  protected:

    friend class io_tlate<eff_fermion>;

    /// The matrix of coefficients
    static double **Pmnf;
    /// The parameter \f$ a \f$
    static double parma;
    /// The array row size
    static int sizem;
    /// The array column size
    static int sizen;
    
    /// The solver for \f$ \psi \f$
    root<void *,funct<void *> > *psi_root;
    /// The other solver for calc_density()
    root<void *,funct<void *> > *density_root;
    
    /// The function which solves for \f$ f \f$ from \f$ \psi \f$.
    int solve_fun(double x, double &y, void *&pa);
    
    /// Fix density for calc_density()
    int density_fun(double x, double &y, void *&pa);
    
    /// Fix density for pair_density()
    int pair_density_fun(double x, double &y, void *&pa);

#endif

  };

  template<> int io_tlate<eff_fermion>::input
    (cinput *co, in_file_format *ins, eff_fermion *f);
  template<> int io_tlate<eff_fermion>::output
    (coutput *co, out_file_format *outs, eff_fermion *f);
  /*
    template<> int io_tlate<eff_fermion>::stat_input
    (cinput *co, in_file_format *ins, eff_fermion *f);
    template<> int io_tlate<eff_fermion>::stat_output
    (coutput *co, out_file_format *outs, eff_fermion *f);
    template<> bool io_tlate<eff_fermion>::has_static_data();
  */
  template<> const char *io_tlate<eff_fermion>::type();
  
  typedef io_tlate<eff_fermion> eff_fermion_io_type;

#ifndef DOXYGENP
}
#endif

#endif
