/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FERMION_H
#define O2SCL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <gsl/gsl_specfunc.h>
#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/part.h>
#include <o2scl/cern_mroot_root.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Fermion class

      This is a base class for the computation of fermionic
      thermodynamics. This class includes the computations of
      zero-temperature (possibly massive) and massless fermions at
      zero or finite temperature. The more general case of fermions
      with both finite mass and a finite temperature are taken care of
      by the functions calc_mu(), calc_density(), pair_mu(), and
      pair_density(). These functions are not implemented in this base
      class (see eff_fermion, rel_fermion, nonrel_fermion, sn_fermion,
      and sn_nr_fermion).

      The function massless_calc_density() uses a \ref root object to
      solve for the chemical potential as a function of the density.
      The default is an object of type cern_mroot_root. The function
      massless_pair_density() does not need to use the \ref root
      object because of the simplification afforded by the inclusion
      of antiparticles.

      \future Consider putting a parent version of calc_e and calc_p, or
      in part or fermion which automatically solves like
      eff_fermion::calc_density()?

      \htmlonly
      This Mathematica notebook contains the derivations of related
      series expansions and some algebra for the massless_pair()
      functions.
      <a href="../extras/fermion.nb">
      fermion.nb</a>, and
      <a href="../extras/fermion.pdf">
      fermion.pdf</a>.
      \endhtmlonly
      \latexonly
      This Mathematica notebook contains the derivations of related
      series expansions and some algebra for the massless\_pair()
      functions.
      \begin{verbatim}
      doc/o2scl/extras/fermion.nb
      doc/o2scl/extras/fermion.pdf
      \end{verbatim}
      \endlatexonly
  */
  class fermion : virtual public part {
  public:

    /// Fermi momentum
    double kf;
    /// Gap
    double del;

    /// Create a fermion with mass \c mass  and degeneracy \c dof.
    fermion(double mass=0, double dof=0);

    virtual ~fermion() {
    }

    /// \name Zero-temperature fermions 
    //@{
    /** 
	\brief Calculate the Fermi momentum from the density
	
	Uses the relation \f$ k_F = ( 6 \pi^2 n /g )^{1/3} \f$
    */
    int kffromden(); 

    /** 
	\brief Scalar number density at T=0 from \ref kf and \ref ms

	Calculates the integral 
	\f[
	n_s=\frac{g}{2 \pi^2} \int_0^{k_F} k^2 
	\frac{m^{*}}{\sqrt{k^2+m^{* 2}}} d k
	\f]
    */
    int sden(); 
    
    /** 
	\brief Energy density at T=0 from \ref kf and \ref ms

	Calculates the integral 
	\f[
	\varepsilon = \frac{g}{2 \pi^2} \int_0^{k_F} k^2 \
	sqrt{k^2+m^{* 2}} d k
	\f]
    */
    int eden(); 

    /** 
	\brief Pressure at T=0 from \ref kf and \ref ms

	Calculates the integral 
	\f[
	P=\frac{g}{6 \pi^2} \int_0^{k_F} \frac{k^4}{\sqrt{k^2+m^{* 2}}} d k
	\f]
    */
    int pres(); 

    /** 
	\brief Zero temperature fermions from nu and ms

	This function always returns \c gsl_success.
     */
    virtual int calc_mu_zerot();

    /** 
	\brief Zero temperature fermions from n and ms

	This function always returns \c gsl_success.
    */
    virtual int calc_density_zerot();
    //@}

    /** 
	\brief Calculate properties as function of chemical potential
     */
    virtual int calc_mu(const double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
    */
    virtual int pair_density(const double temper);
  
    /// \name Massless fermions
    //@{
    /// Finite temperature massless fermions
    virtual int massless_calc_mu(const double temper);

    /// Finite temperature massless fermions
    virtual int massless_calc_density(const double temper);

    /** 
	\brief Finite temperature massless fermions and antifermions 
    */
    int massless_pair_mu(const double temper);

    /** 
	\brief Finite temperature massless fermions and antifermions 

	In the cases \f$ n^3 >> T \f$ and \f$ T >> n^3 \f$ ,
	expansions are used instead of the exact formulas to avoid
	loss of precision.

	\todo Comment here about the precision of the expansions and
	allow the user to control how they are used if necessary.
    */
    int massless_pair_density(const double temper);
    //@}

    /** \brief Set the solver for use in massless_root_density() */ 
    int set_massless_root(root<void *,funct<void *> > &rp) {
      massless_root=&rp;
      return 0;
    }

    /** 
	\brief Degenerate expansion for specific heat
	
	This is a temporary location and is also unchecked.
    */
    double deg_specific_heat(double T) {
      double ret;
      if (non_interacting) {
	nu=mu;
	ms=m;
      }
      double sqd=nu*nu-ms*ms;
      ret=o2scl_const::pi2*T*nu/sqd*
	(1.0-o2scl_const::pi2*T*T*
	 (5.0*pow(ms,4.0)+4.0*ms*ms*nu*nu+14.0*pow(nu,4.0))/
	 15.0/nu/nu/sqd/sqd);
      return ret;
    }
      
    /** 
	\brief The default solver for massless_calc_mu()
	
	We default to cern_mroot_root here since we don't have
	a bracket or a derivative.
    */
    cern_mroot_root<void *,funct<void *> > def_massless_root;

    /// Return string denoting type ("fermion")
    virtual const char *type() { return "fermion"; }

#ifndef DOXYGENP

    friend class io_tlate<fermion>;

  protected:
    
    /// A pointer to the solver for massless fermions
    root<void *,funct<void *> > *massless_root;
    
    /// The function to solve for massless fermions
    int massless_fun(double x, double &y, void *&pa);

  private:

    fermion(const fermion &);
    fermion& operator=(const fermion&);

#endif

  };

  template<> int io_tlate<fermion>::input
    (cinput *co, in_file_format *ins, fermion *f);
  template<> int io_tlate<fermion>::output
    (coutput *co, out_file_format *outs, fermion *f);
  template<> const char *io_tlate<fermion>::type();
  
  typedef io_tlate<fermion> fermion_io_type;

#ifndef DOXYGENP
}
#endif

#endif
