/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/rel_boson.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// rel_boson class

rel_boson::rel_boson(double mass, double dof) : part(mass,dof),
						boson(mass,dof) {
  name="rel_boson";
  density_root=&def_density_root;
  nit=&def_nit;
  dit=&def_dit;
}

rel_boson::~rel_boson() {
}

int rel_boson::set_inte(inte<void *,funct<void *> > &l_nit, 
			inte<void *,funct<void *> > &l_dit) {
  nit=&l_nit;
  dit=&l_dit;
  return gsl_success;
}

int rel_boson::calc_mu(const double temper) {

  if (temper<=0.0) {
    set_err_ret("T <= 0 in rel_boson::calc_density",gsl_einval);
  }
  if (non_interacting==true) { nu=mu; ms=m; }

  double T=temper;
  void *vp=&T;

  funct_mfptr_noerr<rel_boson,void *> fd(this,&rel_boson::deg_density_fun);
  funct_mfptr_noerr<rel_boson,void *> fe(this,&rel_boson::deg_energy_fun);
  funct_mfptr_noerr<rel_boson,void *> fs(this,&rel_boson::deg_entropy_fun);

  n=dit->integ(fd,0.0,sqrt(pow(15.0*temper+nu,2.0)-ms*ms),vp);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  n*=g/2.0/pi2;
  ed=dit->integ(fe,0.0,sqrt(pow(15.0*temper+nu,2.0)-ms*ms),vp);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  ed*=g/2.0/pi2;
  en=dit->integ(fs,0.0,sqrt(pow(15.0*temper+nu,2.0)-ms*ms),vp);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  en*=g/2.0/pi2;

  pr=-ed+temper*en+mu*n;

  return gsl_success;
}

int rel_boson::nu_from_n(const double temper) {
  double nex;
  double T=temper;
  void *vp=&T;

  nex=nu/temper;
  funct_mfptr<rel_boson,void *> mf(this,&rel_boson::solve_fun);
  int ret=density_root->solve(nex,vp,mf);
  if (ret!=0) return ret;
  nu=nex*temper;
  
  return gsl_success;
}

int rel_boson::calc_density(const double temper) {

  if (temper<=0.0) {
    set_err_ret("T <= 0 in rel_boson::calc_density",gsl_einval);
  }
  if (non_interacting==true) { nu=mu; ms=m; }

  double T=temper;
  void *vp=&T;

  nu_from_n(temper);

  funct_mfptr_noerr<rel_boson,void *> fe(this,&rel_boson::deg_energy_fun);
  funct_mfptr_noerr<rel_boson,void *> fs(this,&rel_boson::deg_entropy_fun);

  ed=dit->integ(fe,0.0,sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vp);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  ed*=g/2.0/pi2;
  en=dit->integ(fs,0.0,sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vp);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  en*=g/2.0/pi2;

  pr=-ed+temper*en+mu*n;

  return gsl_success;

}

double rel_boson::deg_density_fun(double k, void *&pa) {
  double *temper=(double *)pa;

  double E=sqrt(k*k+ms*ms), ret;

  ret=k*k/(exp(E/(*temper)-nu/(*temper))-1.0);

  return ret;
}
  
double rel_boson::deg_energy_fun(double k, void *&pa) {
  double *temper=(double *)pa;

  double E=sqrt(k*k+ms*ms), ret;

  ret=k*k*E/(exp(E/(*temper)-nu/(*temper))-1.0);
  
  return ret;
}
  
double rel_boson::deg_entropy_fun(double k, void *&pa) {
  double *temper=(double *)pa;

  double E=sqrt(k*k+ms*ms), nx, ret;
  nx=1.0/(exp(E/(*temper)-nu/(*temper))-1.0);
  ret=-k*k*(nx*log(nx)-(1.0+nx)*log(1.0+nx));
  
  return ret;
}
  

double rel_boson::density_fun(double u, void *&pa) {
  double ret, y, mx;
  double *temper=(double *)pa;

  y=nu/(*temper);
  mx=ms/(*temper);
  
  ret=(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u+y)/(exp(y)-exp(mx+u));

  return ret;
}

double rel_boson::energy_fun(double u, void *&pa) {
  double ret, y, mx;
  double *temper=(double *)pa;

  y=nu/(*temper);
  mx=ms/(*temper);
  
  ret=(mx+u)*(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u+y)/(exp(y)-exp(mx+u));
  
  return ret;
}

double rel_boson::entropy_fun(double u, void *&pa) {
  double ret, y, mx, term1, term2;
  double *temper=(double *)pa;

  y=mu/(*temper);
  mx=ms/(*temper);

  term1=log(exp(y-mx-u)-1.0)/(exp(y-mx-u)-1.0);
  term2=log(1.0-exp(mx+u-y))/(1.0-exp(mx+u-y));
  ret=(mx+u)*exp(u)*sqrt(u*u+2.0*mx*u)*(term1+term2);
  
  return ret;
}

int rel_boson::solve_fun(double x, double &yy, void *&pa) {
  double nden;
  double *temper=(double *)pa;

  funct_mfptr_noerr<rel_boson,void *> fd(this,&rel_boson::deg_density_fun);
  
  nu=(*temper)*x;
  nden=dit->integ(fd,0.0,sqrt(pow(20.0*(*temper)+nu,2.0)-ms*ms),pa);
  inte_err=err_hnd->get_errno();
  if (inte_err!=0) return inte_err;
  nden*=g/2.0/pi2;
  yy=nden/n-1.0;

  return gsl_success;
}

int rel_boson::pair_mu(const double temper) {
  int ret;
  
  if (non_interacting) { nu=mu; ms=m; }
  ret=calc_mu(temper);
  
  if (ret==0) {
    rel_boson antip(ms,g);
    anti(antip);
    ret=antip.calc_mu(temper);
    n-=antip.n;
    pr+=antip.pr;
    ed+=antip.ed;
    en+=antip.en;
  }

  return ret;
}

