/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_BOSON_H
#define O2SCL_REL_BOSON_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/boson.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qag.h>
#include <o2scl/gsl_inte_qagiu.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic boson

      \todo Testing not completely finished.
  */
  class rel_boson : public boson {

  public:

    /// Create a boson with mass \c m and degeneracy \c g
    rel_boson(double m=0.0, double g=0.0);

    virtual ~rel_boson();

    /// The error value from mroot
    int mroot_err;

    /// The error value from inte
    int inte_err;
    
    /**
       \brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(const double temper);
    
    /** 
	\brief Calculate properties as function of density
    */
    virtual int calc_density(const double temper);
    
    /** \brief Calculate properties with antiparticles as function of
        chemical potential
    */
    virtual int pair_mu(const double temper);

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(const double temper);

    /// Set inte object
    int set_inte(inte<void *,funct<void *> > &l_nit, 
		 inte<void *,funct<void *> > &l_dit);

    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<void *,funct<void *> > &rp) {
      density_root=&rp;
      return 0;
    }

    /// The default solver for calc_density().
    cern_mroot_root<void *,funct<void *> > def_density_root;

    /// Default nondegenerate integrator
    gsl_inte_qagiu<void *,funct<void *> > def_nit;

    /// Default degenerate integrator
    gsl_inte_qag<void *,funct<void *> > def_dit;

    /// Return string denoting type ("rel_boson")
    virtual const char *type() { return "rel_boson"; }

  protected:

#ifndef DOXYGENP

    friend class io_tlate<rel_boson>;

    /// The non-degenerate integrator
    inte<void *,funct<void *> > *nit;
    /// The degenerate integrator
    inte<void *,funct<void *> > *dit;
    /// The solver for calc_density()
    root<void *,funct<void *> > *density_root;
    
    /// Non-degenerate density integral
    double density_fun(double u, void *&pa);
    /// Non-degenerate energy density integral
    double energy_fun(double u, void *&pa);
    /// Non-degenerate entropy integral
    double entropy_fun(double u, void *&pa);
    /// Degenerate density integral
    double deg_density_fun(double u, void *&pa);
    /// Degenerate energy density integral
    double deg_energy_fun(double u, void *&pa);
    /// Degenerate entropy integral
    double deg_entropy_fun(double u, void *&pa);
    /// Solve for the density in calc_density()
    int solve_fun(double x, double &yy, void *&pa);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
