/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/rel_fermion.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// rel_fermion class

rel_fermion::rel_fermion(double mass, double dof) : 
  part(mass,dof), fermion(mass,dof) {
  name="rel_fermion";
  deg_limit=2.0;
  nit=&def_nit;
  dit=&def_dit;
  density_root=&def_density_root;
}

rel_fermion::~rel_fermion() {
}

int rel_fermion::set_inte(inte<void *,funct<void *> > &l_nit, 
			  inte<void *,funct<void *> > &l_dit) {
  nit=&l_nit;
  dit=&l_dit;
  return gsl_success;
}

int rel_fermion::calc_mu(const double temper) {
  if (temper<=0.0) return fermion::calc_mu_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }
  
  double T=temper;
  void *vpx=(void *)&T;
  
  if ((nu-ms)/temper<deg_limit) {
    
    funct_mfptr_noerr<rel_fermion,void *> mfd(this,&rel_fermion::density_fun);
    funct_mfptr_noerr<rel_fermion,void *> mfe(this,&rel_fermion::energy_fun);
    funct_mfptr_noerr<rel_fermion,void *> mfs(this,&rel_fermion::entropy_fun);
    
    n=nit->integ(mfd,0.0,0.0,vpx);
    n*=g*pow(temper,3.0)/2.0/pi2;
    unc.n=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

    ed=nit->integ(mfe,0.0,0.0,vpx);
    ed*=g*pow(temper,4.0)/2.0/pi2;
    unc.ed=nit->get_error()*g*pow(temper,4.0)/2.0/pi2;
    
    en=nit->integ(mfs,0.0,0.0,vpx);
    en*=g*pow(temper,3.0)/2.0/pi2;
    unc.en=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

  } else {

    funct_mfptr_noerr<rel_fermion, void *> 
      mfd(this,&rel_fermion::deg_density_fun);
    funct_mfptr_noerr<rel_fermion, void *> 
      mfe(this,&rel_fermion::deg_energy_fun);
    funct_mfptr_noerr<rel_fermion, void *> 
      mfs(this,&rel_fermion::deg_entropy_fun);

    n=dit->integ(mfd,0.0,
		 sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vpx);
    n*=g/2.0/pi2;
    unc.n=dit->get_error()*g/2.0/pi2;
    
    ed=dit->integ(mfe,0.0,
		  sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vpx);
    ed*=g/2.0/pi2;
    unc.ed=dit->get_error()*g/2.0/pi2;
    
    en=dit->integ(mfs,0.0,
		  sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vpx);    
    en*=g/2.0/pi2;
    unc.en=dit->get_error()*g/2.0/pi2;
  }

  pr=-ed+temper*en+nu*n;
  unc.pr=sqrt(unc.ed*unc.ed+temper*unc.en*temper*unc.en+nu*unc.n*nu*unc.n);

  return gsl_success;
}

int rel_fermion::nu_from_n(const double temper) {
  double nex;
  double T=temper;

  if (guess_from_nu) {
    if (nu>ms) {
      nex=nu/temper;
    } else {
      nex=ms*1.01/temper;
    }
  } else {
    set_err("guess_from_nu==false not implemented in rel_fermion.",
	    gsl_eunimpl);
    if (nu>ms) {
      nex=nu/temper;
    } else {
      nex=ms*1.01/temper;
    }
  }
  
  funct_mfptr<rel_fermion, void *> mf(this,&rel_fermion::solve_fun);
  void *vpx=(void *)&T;
  int ret=density_root->solve(nex,vpx,mf);
  nu=nex*temper;
  
  return ret;
}

int rel_fermion::calc_density(const double temper) {
  int mret;
  double T=temper;

  void *vpx=(void *)&T;
  if (non_interacting==true) { nu=mu; ms=m; }
  
  mret=nu_from_n(temper);

  if (non_interacting) { mu=nu; }

  if ((nu-ms)/temper<deg_limit) {
    funct_mfptr_noerr<rel_fermion, void *> mfe(this,&rel_fermion::energy_fun);
    funct_mfptr_noerr<rel_fermion, void *> mfs(this,&rel_fermion::entropy_fun);

    ed=nit->integ(mfe,0.0,0.0,vpx);
    ed*=g*pow(temper,4.0)/2.0/pi2;
    unc.ed=nit->get_error()*g*pow(temper,4.0)/2.0/pi2;
    
    en=nit->integ(mfs,0.0,0.0,vpx);
    en*=g*pow(temper,3.0)/2.0/pi2;
    unc.en=nit->get_error()*g*pow(temper,3.0)/2.0/pi2;

  } else {
    funct_mfptr_noerr<rel_fermion, void *> 
      mfe(this,&rel_fermion::deg_energy_fun);
    funct_mfptr_noerr<rel_fermion, void *> 
      mfs(this,&rel_fermion::deg_entropy_fun);

    ed=dit->integ(mfe,0.0,
		  sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vpx);
    ed*=g/2.0/pi2;
    unc.ed=dit->get_error()*g/2.0/pi2;

    en=dit->integ(mfs,0.0,
		  sqrt(pow(20.0*temper+nu,2.0)-ms*ms),vpx);
    en*=g/2.0/pi2;
    unc.en=dit->get_error()*g/2.0/pi2;

  }

  pr=-ed+temper*en+mu*n;
  unc.pr=sqrt(unc.ed*unc.ed+temper*unc.en*temper*unc.en+mu*unc.n*mu*unc.n);
  
  return mret;

}

double rel_fermion::deg_density_fun(double k, void *&pa) {
  double *temper=(double *)pa;

  double E=sqrt(k*k+ms*ms), ret;
  
  ret=k*k/(1.0+exp((E-nu)/(*temper)));
  
  return ret;
}

double rel_fermion::deg_energy_fun(double k, void *&pa) {
  double *temper=(double *)pa;

  double E=sqrt(k*k+ms*ms), ret;

  ret=k*k*E/(1.0+exp((E-nu)/(*temper)));
  
  return ret;
}
  
double rel_fermion::deg_entropy_fun(double k, void *&pa) {
  double *temper=(double *)pa;
  double E=sqrt(k*k+ms*ms), nx, ret, nm1;

  // If the argument to the exponential is really small, then the
  // value of the integrand is just zero
  if (((E-nu)/(*temper))<-200.0) {
    ret=0.0;
    // Otherwise, if the argument to the exponential is still small,
    // then addition of 1 makes us lose precision like mad, so we use
    // an alternative:
  } else if (((E-nu)/(*temper))<-30.0) {
    nm1=-exp(E/(*temper)-nu/(*temper));
    ret=k*k*nm1*log(-nm1);
  } else {
    nx=1.0/(1.0+exp(E/(*temper)-nu/(*temper)));
    ret=-k*k*(nx*log(nx)+(1.0-nx)*log(1.0-nx));
  }

  return ret;
}
  
double rel_fermion::density_fun(double u, void *&pa) {
  double ret, y, mx;
  double *temper=(double *)pa;

  y=nu/(*temper);
  mx=ms/(*temper);
  
  if (y-mx-u>200.0) {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u);
  } else if (y>u+200.0 && mx>u+200.0) {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u)/(exp(mx+u-y)+1.0);
  } else {
    ret=(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u+y)/(exp(mx+u)+exp(y));
  }
  ret/=exp(u);

  if (!finite(ret)) {
    ret=0.0;
  }

  return ret;
}

double rel_fermion::energy_fun(double u, void *&pa) {
  double ret, y, mx;
  double *temper=(double *)pa;

  y=nu/(*temper);
  mx=ms/(*temper);
  
  if (y>u+200.0 && mx>u+200.0) {
    ret=(mx+u)*(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u)/(exp(mx+u-y)+1.0);
  } else {
    ret=(mx+u)*(mx+u)*sqrt(u*u+2.0*mx*u)*exp(u+y)/(exp(mx+u)+exp(y));
  }
  ret/=exp(u);
  
  if (!finite(ret)) {
    return 0.0;
  }

  return ret;
}

double rel_fermion::entropy_fun(double u, void *&pa) {
  double ret, y, mx, term1, term2;
  double *temper=(double *)pa;

  y=mu/(*temper);
  mx=ms/(*temper);

  term1=log(1.0+exp(y-mx-u))/(1.0+exp(y-mx-u));
  term2=log(1.0+exp(mx+u-y))/(1.0+exp(mx+u-y));
  ret=(mx+u)*exp(u)*sqrt(u*u+2.0*mx*u)*(term1+term2);
  ret/=exp(u);
  
  if (!finite(ret)) {
    return 0.0;
  }

  return ret;
}

int rel_fermion::solve_fun(double x, double &yy, void *&pa) {
  double nden;
  double *temper=(double *)pa;
  
  nu=(*temper)*x;

  // 6/6/03 - Should this be included? I think yes!
  if (non_interacting) mu=nu;

  if ((nu-ms)/(*temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion, void *> mfe(this,&rel_fermion::density_fun);

    nden=nit->integ(mfe,0.0,0.0,pa);
    nden*=g*pow((*temper),3.0)/2.0/pi2;
    unc.n=nit->get_error()* g*pow((*temper),3.0)/2.0/pi2;
    
    yy=(n-nden)/n;

  } else {
    
    funct_mfptr_noerr<rel_fermion, void *> 
      mfe(this,&rel_fermion::deg_density_fun);
    
    nden=dit->integ(mfe,0.0,
		    sqrt(pow(20.0*(*temper)+ nu,2.0)- ms* ms),pa);
    nden*=g/2.0/pi2;
    unc.n=err_hnd->get_errno()*g/2.0/pi2;

    yy=(n-nden)/n;
  }
  
  return gsl_success;
}

int rel_fermion::pair_mu(const double temper) {
  int ret=gsl_success;

  rel_fermion antip(ms,g);

  if (non_interacting) { nu=mu; ms=m; }
  anti(antip);

  ret=calc_mu(temper);
  if (ret==0) {
    ret=antip.calc_mu(temper);
    n-=antip.n;
    pr+=antip.pr;
    ed+=antip.ed;
    en+=antip.en;
    unc.n=sqrt(unc.n*unc.n+antip.unc.n*antip.unc.n);
    unc.ed=sqrt(unc.ed*unc.ed+antip.unc.ed*antip.unc.ed);
    unc.pr=sqrt(unc.pr*unc.pr+antip.unc.pr*antip.unc.pr);
    unc.en=sqrt(unc.en*unc.en+antip.unc.en*antip.unc.en);
  }

  return ret;

}

int rel_fermion::pair_density(const double temper) {
  double nex;
  int ret;

  if (temper<=0.0) return fermion::calc_density_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }

  double T=temper;
  void *vpx=(void *)&T;

  nex=nu/temper;
  funct_mfptr<rel_fermion, void *> mf(this,&rel_fermion::pair_fun);
  ret=density_root->solve(nex,vpx,mf);
  if (ret!=0) {
    add_err("Solve failed in pair_density().",ret);
  }
  nu=nex*temper;
  
  if (non_interacting==true) { mu=nu; }
  
  ret=pair_mu(temper);

  return ret;

}

int rel_fermion::pair_fun(double x, double &yy, void *&pa) {
  double nden;
  double *temper=(double *)pa;

  nu=(*temper)*x;

  // 6/6/03 - Should this be included? I think yes!
  if ( non_interacting) mu=nu;

  if ((nu-ms)/(*temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion, void *> mfe(this,&rel_fermion::density_fun);

    nden=nit->integ(mfe,0.0,0.0,pa);
    nden*=g*pow((*temper),3.0)/2.0/pi2;
    yy=nden;
  } else {
    funct_mfptr_noerr<rel_fermion, void *> 
      mfe(this,&rel_fermion::deg_density_fun);

    nden=dit->integ(mfe,0.0,
		    sqrt(pow(20.0*(*temper)+ nu,2.0)- ms* ms),pa);
    nden*=g/2.0/pi2;
    yy=nden;
  }

  nu=-(*temper)*x ;
  if ((nu-ms)/(*temper)< deg_limit) {
    funct_mfptr_noerr<rel_fermion, void *> mf(this,&rel_fermion::density_fun);

    nden=nit->integ(mf,0.0,0.0,pa);
    nden*=g*pow((*temper),3.0)/2.0/pi2;
    yy-=nden;
  } else {
    funct_mfptr_noerr<rel_fermion, void *> 
      mf(this,&rel_fermion::deg_density_fun);
    
    nden=dit->integ(mf,0.0,sqrt(pow(20.0*(*temper)+ nu,2.0)- ms* ms),pa);
    nden*=g/2.0/pi2;
    yy-=nden;
  }

  yy=yy/n-1.0;

  return gsl_success;
}

template<> int io_tlate<rel_fermion>::input
(cinput *co, in_file_format *ins, rel_fermion *f) {
  std::string t;
  co->object_in("fermion",ins,f,t);
  ins->bool_in(f->guess_from_nu,"guess_from_nu");
  return 0;
}

template<> int io_tlate<rel_fermion>::output
(coutput *co, out_file_format *outs, rel_fermion *f) {
  std::string t;
  co->object_out("fermion",outs,f,0,0,t);
  outs->bool_out(f->guess_from_nu,"guess_from_nu");
  return 0;
}

template<> const char *io_tlate<rel_fermion>::type()
{ return "rel_fermion"; }

template<> int io_tlate<rel_fermion>::stat_input(cinput *co, 
						 in_file_format *ins, 
						 rel_fermion *f) {
  /*
    string tmp;

    bool sw;
    ins->bool_in(sw,"inte");
    if (sw) {
    pointer_in(co,ins,((void **)(&f->nit)),tmp);
    pointer_in(co,ins,((void **)(&f->dit)),tmp);
    }

    ins->bool_in(sw,"density_root");
    if (sw) {
    pointer_in(co,ins,((void **)(&f->density_root)),tmp);
    }
  */

  return 0;
}

template<> int io_tlate<rel_fermion>::stat_output(coutput *co, 
						  out_file_format *outs,
						  rel_fermion *f) {

  /*
    if (f->nit!=&f->def_nit || f->dit!=&f->def_dit) {
    outs->bool_out(true,"inte");
    pointer_out(co,outs,f->nit,f->nit->type());
    pointer_out(co,outs,f->dit,f->dit->type());
    } else {
    outs->bool_out(false,"inte");
    }
  */

  /*
    if (f->density_root!=&f->def_density_root) {
    outs->bool_out(true,"density_root");
    pointer_out(co,outs,f->density_root,f->density_root->type());
    } else {
    outs->bool_out(false,"density_root");
    }
  */
    
  return 0;
}


