/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/sn_nr_fermion.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

//--------------------------------------------
// sn_nr_fermion class

sn_nr_fermion::sn_nr_fermion(double mass, double dof) : 
  part(mass,dof), deriv_fermion(mass,dof), unc(mass,dof) {
  
  name="sn_nr_fermion";
  flimit=20.0;
  guess_from_nu=true;

  inc_rest_mass=false;

  density_root=&def_density_root;
}

sn_nr_fermion::~sn_nr_fermion() {
}

int sn_nr_fermion::calc_mu(const double temper) {
  
  if (temper<=0.0) return fermion::calc_mu_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }
  
  double pfac2=g*pow(ms*temper/2.0/pi,1.5)/temper, y;
  
  if (inc_rest_mass) {
    y=(nu-m)/temper;
  } else {
    y=nu/temper;
  }

  double half=gsl_sf_fermi_dirac_half(y);
  double mhalf=gsl_sf_fermi_dirac_mhalf(y);
  double thalf=gsl_sf_fermi_dirac_3half(y);
  
  // Number density:
  n=pfac2*half*temper;
  
  // Energy density:
  ed=pfac2*temper*temper*1.5*thalf;

  if (inc_rest_mass) {
    
    // Finish energy density
    ed+=n*m;
    
    // entropy density
    en=((ed-n*m)/0.6-(nu-m)*n)/temper;
    
    // pressure
    pr=(ed-n*m)/1.5;
    
  } else {
    
    // entropy density
    en=(5.0*ed/3.0-nu*n)/temper;
    
    // pressure
    pr=2.0*ed/3.0;
    
  }
    
  dndT=pfac2*(1.5*half-y*mhalf);
  
  dndmu=pfac2*mhalf;
  
  dsdT=pfac2*(3.75*thalf-3.0*y*half+y*y*mhalf);
  
  dndm=1.5*pfac2*temper/ms*half;
  
  return gsl_success;
}

int sn_nr_fermion::nu_from_n(const double temper) {
  double nex;
  double T=temper;

  if (guess_from_nu) {
    nex=nu/temper;
  } else {
    set_err_ret("guess_from_nu==false not implemented in sn_nr_fermion.",
		gsl_eunimpl);
  }
  funct_mfptr<sn_nr_fermion,void *> mf(this,&sn_nr_fermion::solve_fun);
  void *vpT=&T;
  int ret=density_root->solve(nex,vpT,mf);
  nu=nex*temper;

  return ret;
}

int sn_nr_fermion::calc_density(const double temper) {
  int mret;
  double T=temper;
  
  if (non_interacting==true) { ms=m; nu=mu; }
  
  mret=nu_from_n(temper);
  
  if (non_interacting) { mu=nu; }
  
  mret=calc_mu(temper);
  
  return mret;
}

int sn_nr_fermion::solve_fun(double x, double &yy, void *&pa) {
  double nden, y;
  double temper=*((double *)pa);
  
  nu=temper*x;
  
  // 6/6/03 - I think this should this be included.
  if (non_interacting) mu=nu;

  if (inc_rest_mass) {
    y=(nu-m)/temper;
  } else {
    y=nu/temper;
  }

  nden=gsl_sf_fermi_dirac_half(y)*sqrt(pi)/2.0;
  nden*=g*pow(2.0*ms*temper,1.5)/4.0/pi2;

  yy=(n-nden)/n;
  
  return gsl_success;
}

int sn_nr_fermion::pair_mu(const double temper) {
  int ret;
  
  sn_nr_fermion antip(ms,g);

  if (non_interacting) { nu=mu; ms=m; }
  anti(antip);

  ret=calc_mu(temper);

  int ret2=antip.calc_mu(temper);
  n-=antip.n;
  pr+=antip.pr;
  ed+=antip.ed;
  en+=antip.en;
  if (ret2!=0 && ret==0) return ret2;
  
  return ret;
}

int sn_nr_fermion::pair_density(const double temper) {
  double nex;
  int ret;
  
  if (temper<=0.0) return fermion::calc_density_zerot();
  if (non_interacting==true) { nu=mu; ms=m; }
  
  double T=temper;

  nex=nu/temper;
  funct_mfptr<sn_nr_fermion,void *> mf(this,&sn_nr_fermion::pair_fun);
  void *vpT=&T;

  ret=density_root->solve(nex,vpT,mf);
  nu=nex*temper;

  if (non_interacting==true) { mu=nu; }
  
  ret=pair_mu(temper);

  return ret;

}

int sn_nr_fermion::pair_fun(double x, double &yy, void *&pa) {
  double nden, y;
  double temper=*((double *)pa);

  nu=temper*x;

  // 6/6/03 - Should this be included? I think yes!
  if (non_interacting) mu=nu;

  if (inc_rest_mass) {
    y=(nu-m)/temper;
  } else {
    y=nu/temper;
  }

  nden=gsl_sf_fermi_dirac_half(y)*sqrt(pi)/2.0;
  nden*=g*pow(2.0*ms*temper,1.5)/4.0/pi2;
  
  yy=nden;

  if (inc_rest_mass) {
    y=-(nu-m)/temper;
  } else {
    y=-nu/temper;
  }
  
  nden=gsl_sf_fermi_dirac_half(y)*sqrt(pi)/2.0;
  nden*=g*pow(2.0*ms*temper,1.5)/4.0/pi2;
  
  yy-=nden;
  
  yy=yy/n-1.0;

  return gsl_success;
}

template<> int io_tlate<sn_nr_fermion>::input(cinput *co, in_file_format *ins, 
					   sn_nr_fermion *f) {
  deriv_fermion *g=(deriv_fermion *)f;
  deriv_fermion *h=(deriv_fermion *)(&f->unc);
  std::string t;
  co->object_in("deriv_fermion",ins,g,t);
  co->object_in("deriv_fermion",ins,h,t);
  ins->double_in(f->flimit);
  ins->bool_in(f->guess_from_nu);
  return 0;
}

template<> int io_tlate<sn_nr_fermion>::output(coutput *co, 
					    out_file_format *outs, 
					    sn_nr_fermion *f) {
  deriv_fermion *g=(deriv_fermion *)f;
  deriv_fermion *h=(deriv_fermion *)(&f->unc);
  std::string t;
  co->object_out("deriv_fermion",outs,g,0,0,t);
  co->object_out("deriv_fermion",outs,h,0,0,t);
  outs->double_out(f->flimit);
  outs->bool_out(f->guess_from_nu);
  return 0;
}

template<> const char *io_tlate<sn_nr_fermion>::type() {
  return "sn_nr_fermion";
}

