/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_RNGA_H
#define O2SCL_GSL_RNGA_H

#include <gsl/gsl_rng.h>
#include <o2scl/rnga.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Random number generator (GSL)

      If \c seed is zero, or is not given, then the default seed
      specific to the particular random number generator is used.  No
      virtual functions are used in this class or its parent, \ref
      rnga. This should be as fast as the original GSL version.
  */

  class gsl_rnga : public rnga {
  public:

    /** \brief Initialize the random number generator with type \c gtype 
	and the default seed 
    */
    gsl_rnga(const gsl_rng_type *gtype=gsl_rng_mt19937);
    
    /// Initialize the random number generator with \c seed 
    gsl_rnga(unsigned long int seed, 
	     const gsl_rng_type *gtype=gsl_rng_mt19937);
    
    ~gsl_rnga();
    
    /// Return rng type
    const gsl_rng_type *get_type() { return rng; }
    
    /// Return a random number in \f$(0,1]\f$
    double random() {
      return (gr->type->get_double)(gr->state);
    }

    /// Return the maximum integer for random_int()
    unsigned long int get_max() {
      return gsl_rng_max(gr)-gsl_rng_min(gr);
    }

    /// Return random integer in \f$[0,\mathrm{max}-1]\f$.
    unsigned long int random_int(unsigned long int n=0);
    
    /** 
	\brief Return a pointer to the gsl_rng object (deprecated)
	
        Used in gsl_miser and gsl_anneal.
    */
    gsl_rng *get_gsl_rng();
  
    /// Set the seed
    void set_seed(unsigned long int s) { 
      seed=s;
      gsl_rng_set(gr,seed);
    }

    /// Set the seed
    void clock_seed() {
      seed=time(NULL);
      gsl_rng_set(gr,seed);
    }
  protected:

#ifndef DOXYGENP

    gsl_rnga(const gsl_rnga &);
    gsl_rnga& operator=(const gsl_rnga&);

#endif
#ifndef DOXYGEN_INTERNAL

    /// The GSL random number generator
    gsl_rng *gr;

    /// The GSL random number generator type
    const gsl_rng_type *rng;

#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif
