/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ROOT_H
#define O2SCL_ROOT_H

#include <iostream>
#include <cmath>
#include <o2scl/err_hnd.h>
#include <o2scl/collection.h>
#include <o2scl/funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief 1-dimensional solver base class

      \note This class does not actually do any solving, it is
      present to provide member data and various functions common
      to all the 1 dimensional solvers. 

      \future This does not have pure virtual functions, but I'd still
      like to prevent the user from directly instantiating a root
      object.
  */
  template<class param_t, class func_t, class dfunc_t=func_t> 
    class root {

    public:
      
    root() {
      ntrial=100;
      tolf=1.0e-8;
      verbose=0;
      tolx=1.0e-12;
      bracket_step=1.0e-4;
      last_ntrial=0;
    }

    virtual ~root() {}

    /** \brief The maximum value of the functions for success 
	(default \f$ 10^{-8} \f$ )
    */
    double tolf;
      
    /** \brief The minimum allowable stepsize 
	(default \f$ 10^{-12} \f$ )
    */
    double tolx;
    
    /// Output control (default 0)
    int verbose;
    
    /// Maximum number of iterations (default 100)
    int ntrial;

    /// Should be true if root_bkt() is overloaded
    bool over_bkt;
      
    /// Should be true if root_de() is overloaded
    bool over_de;
    
    /** 
	\brief The stepsize for automatic bracketing 
	(default \f$ 10^{-4} \f$)

	If this is exactly zero, it will be reset to
	\f$ 10^{-4} \f$ by solve().
    */
    double bracket_step;
 
    /// The number of iterations for in the most recent minimization
    int last_ntrial;

    /// Return the type, \c "root".
    virtual const char *type() { return "root"; }

    /** 
	\brief Print out iteration information.
         
	Depending on the value of the variable verbose, this prints
	out the iteration information. If verbose=0, then no
	information is printed, while if verbose>1, then after each
	iteration, the present values of \c x and \c y are
	output to std::cout along with the iteration number. If
	verbose>=2 then each iteration waits for a character before
	continuing
 
    */
    virtual int print_iter(double x, double y, int iter, double value=0.0,
			   double limit=0.0, std::string comment="") {
      if (verbose<=0) return gsl_success;
	
      char ch;
	
      std::cout << comment << " Iteration: " << iter << std::endl;
      if (x<0) std::cout << x << " ";
      else std::cout << " " << x << " ";
      if (y<0) std::cout << y << " ";
      else std::cout << " " << y << " ";
      if (value<0) std::cout << value << " ";
      else std::cout << " " << value << " ";
      if (limit<0) std::cout << limit << std::endl;
      else std::cout << " " << limit << std::endl;
      if (verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }
 
      return gsl_success;
    }
    
    /** 
	\brief Solve \c func using \c x as an initial guess
    */
    virtual int solve(double &x, param_t &pa, func_t &func) {

      if (over_bkt) {
	if (bracket_step==0.0) bracket_step=1.0e-4;

	double x2=0.0, dx, fx, fx2;
	int i=0;
	bool done=false;
	// Use function to try to bracket a root
	while(done==false && i<10) {
	  
	  func(x,fx,pa);
	  func(x*(1.0+bracket_step),fx2,pa);
	    
	  dx=(fx2-fx)/(bracket_step*x);
	  x2=x-2.0*fx/dx;
	  
	  func(x2,fx2,pa);
	  
	  if (fx*fx2<0.0) {
	    done=true;
	  } else {
	    x=(x2+x)/2.0;
	  }

	  i++;
	}
	if (done==false) {
	  set_err_ret("Failed to bracket function in root::solve().",
		      o2scl::gsl_emaxiter);
	}
	return solve_bkt(x,x2,pa,func);
      }

      return 0;
    }

    /** \brief Solve \c func in region \f$ x_1<x<x_2 \f$  
	returning \f$ x_1 \f$ .
    */
    virtual int solve_bkt(double &x1, double x2, param_t &pa, 
			  func_t &func) 
    {
      return solve(x1,pa,func);
    }

    /** \brief Solve \c func using \c x as an initial
	guess using derivatives \c df.
    */
    virtual int solve_de(double &x, param_t &pa, func_t &func, dfunc_t &df) {

      if (over_bkt) {
	double x2=0.0, dx, fx, fx2;
	int i=0;
	bool done=false;
	
	// Use derivative information to try to bracket a root
	while(done==false && i<10) {
	  
	  func(x,fx,pa);
	  df(x,dx,pa);
	  
	  x2=x-2.0*fx/dx;
	  
	  func(x2,fx2,pa);
	    
	  if (fx*fx2<0.0) {
	    done=true;
	  } else {
	    x=(x2+x)/2.0;
	  }
	  i++;
	}
	
	if (done==false) {
	  set_err_ret("Failed to bracket function in root::solve_de().",
		      o2scl::gsl_efailed);
	}
	
	return solve_bkt(x,x2,pa,func);
      } 
	
      return solve(x,pa,func);
    }

  };

#ifndef DOXYGENP
  template<> int io_tlate<root<void *,funct<void *> > >::input
    (cinput *co, in_file_format *ins, root<void *, funct<void *> > *ro);
  template<> int io_tlate<root<void *,funct<void *> > >::output
    (coutput *co, out_file_format *outs, root<void *, funct<void *> > *ro);
  template<> const char *io_tlate<root<void *,funct<void *> > >::type();
#endif
  
  typedef io_tlate<root<void *,funct<void *> > > root_io_type;
 
#ifndef DOXYGENP
}
#endif

#endif

